# Идентификация библиотечных функций

> Библиотечный код может составлять львиную долю программы. Ясное дело, ничего интересного этот код не содержит. Поэтому его анализ с полной уверенностью можно опустить. Незачем тратить на него наше драгоценное время. Однако как быть, если дизассемблер неправильно распознал имена функций? Что, придется изучать многотонный листинг самостоятельно? У хакеров есть проверенные методы решения подобных проблем — о них мы сегодня и поговорим.

Читая текст программы, написанный на языке высокого уровня, мы только в исключительных случаях изучаем реализацию стандартных библиотечных функций, таких, например, как printf. Да и зачем? Ее назначение хорошо известно, а если и есть какие неясности — всегда можно заглянуть в описание...

Анализ дизассемблерного листинга — дело другое. Имена функций, за редкими исключениями, в нем отсутствуют, и определить, printf это или что‑то другое, «на взгляд» невозможно. Приходится вникать в алгоритм... Легко сказать — вникать! Та же printf представляет собой сложный интерпретатор строки спецификаторов — с ходу в нем не разберешься! А ведь есть и более монструозные функции. Самое обидное — алгоритм их работы не имеет никакого отношения к анализу исследуемой программы. Тот же new может и выделять память из Windows-кучи, и реализовывать собственный менеджер, но нам‑то от этого что? Достаточно знать, что это именно new, то есть функция выделения памяти, а не free или, скажем, fopen.

Доля библиотечных функций в программе в среднем составляет от пятидесяти до девяноста процентов. Особенно она велика у программ, составленных в визуальных средах разработки, использующих автоматическую генерацию кода (например, Microsoft Visual Studio, Delphi). Причем библиотечные функции подчас намного сложнее и запутаннее тривиального кода самой программы. Обидно, львиная доля усилий на анализ тратится впустую... Как бы оптимизировать это?

## НА ПОМОЩЬ ВНОВЬ ПРИХОДИТ IDA

Уникальная способность IDA различать стандартные библиотечные функции множества компиляторов выгодно отличает ее от большинства других дизассемблеров, этого делать не умеющих. К сожалению, IDA (как и все, созданное человеком) далека от идеала: каким бы обширным ни был список поддерживаемых библиотек, конкретные версии конкретных поставщиков или моделей памяти могут отсутствовать. И даже из тех библиотек, что ей известны, распознаются не все функции (о причинах будет рассказано чуть позже).

Впрочем, нераспознанная функция — это полбеды, неправильно распознанная функция — много хуже, ибо это приводит к ошибкам (иногда трудноуловимым) в анализе исследуемой программы или ставит исследователя в глухой тупик. Например, вызывается fopen, и возвращенный ею результат спустя некоторое время передается free — с одной стороны: почему бы и нет? Ведь fopen возвращает указатель на структуру FILE, а free ее удаляет. А если free никакой не free, а, скажем, fseek? Пропустив операцию позиционирования, мы не сможем правильно восстановить структуру файла, с которым работает программа.

## ОПОЗНАНИЕ ФУНКЦИЙ

Распознать ошибки IDA будет легче, если представлять, как именно она выполняет распознавание. Многие почему‑то считают, что здесь задействован тривиальный подсчет CRC (контрольной суммы). Что ж, подсчет CRC — заманчивый алгоритм, но, увы, для решения данной задачи он непригоден. Основной камень преткновения — наличие непостоянных фрагментов, а именно перемещаемых элементов. И хотя при подсчете CRC перемещаемые элементы можно просто игнорировать (не забывая проделывать ту же операцию и в идентифицируемой функции), разработчик IDA пошел другим, более запутанным и витиеватым, но и более производительным путем.

Ключевая идея заключается в том, что незачем тратить время на вычисление CRC. Для предварительной идентификации функции вполне сойдет и тривиальное посимвольное сравнение, за вычетом перемещаемых элементов (они игнорируются и в сравнении не участвуют). Точнее говоря, не сравнение, а поиск заданной последовательности байтов в эталонной базе, организованной в виде двоичного дерева. Время двоичного поиска, как известно, пропорционально логарифму количества записей в базе. Здравый смысл подсказывает, что длина шаблона (иначе говоря, сигнатуры, то есть сравниваемой последовательности) должна быть достаточной для однозначной идентификации функции. Однако разработчик IDA по непонятным для авторов причинам решил ограничиться только первыми тридцатью двумя байтами, что (особенно если вычесть пролог, который у всех функций практически одинаков) довольно мало.

И верно! Многие функции попадают на один и тот же лист дерева, возникает коллизия — неоднозначность отождествления. Для разрешения ситуации у всех «коллизионных» функций подсчитывается CRC16 с тридцать второго байта до первого перемещаемого элемента и сравнивается с CRC16 эталонных функций. Чаще всего это срабатывает, но, если первый перемещаемый элемент окажется расположенным слишком близко к тридцать второму байту, последовательность подсчета контрольной суммы окажется слишком короткой, а то и вовсе равной нулю (может же быть тридцать второй байт перемещаемым элементом, почему бы и нет?). В случае повторной коллизии находим в функциях байт, в котором все они отличаются, и запоминаем его смещение в базе.

Все это (да простит авторов разработчик IDA!) напоминает следующий анекдот. Поймали туземцы немца, американца и украинца и говорят им: мол, или откупайтесь чем‑нибудь, или съедим. На откуп предлагается: миллион долларов (только не спрашивайте, зачем туземцам миллион долларов, — может, костер жечь), сто щелбанов или съесть мешок соли. Ну, американец достает сотовый, звонит кому‑то... Приплывает катер с миллионом долларов, и американца благополучно отпускают. Немец в это время героически съедает мешок соли, и его полумертвого спускают на воду. Украинец же ел соль, ел‑ел, две трети съел, не выдержал и говорит: а, ладно, черти, бейте щелбаны. Бьет вождь его, и только девяносто ударов отщелкал, тот не выдержал и говорит: да нате миллион, подавитесь! Так и с IDA, посимвольное сравнение не до конца, а только тридцати двух байтов, подсчет CRC не для всей функции, а сколько случай на душу положит, наконец, последний ключевой байт — и тот‑то «ключевой», да не совсем. Дело в том, что многие функции совпадают байт в байт, но совершенно различны по названию и назначению. Не веришь? Тогда как тебе понравится следующее:

```asm
read:                    write:
        sub  rsp, 28h           sub rsp, 28h
        call _read              call _write
        add  rsp, 28h           add rsp, 28h

    retn                     retn
```

Тут без анализа перемещаемых элементов никак не обойтись! Причем это не какой‑то специально надуманный пример — подобных функций очень много. В частности, библиотеки от Embarcadero (в прошлом от Borland) ими так и кишат. Поэтому в былые времена IDA часто

«спотыкалась» и впадала в грубые ошибки. Тем не менее сейчас IDA заметно возмужала и уже не страдает детскими болячками. Для примера скормим компилятору C++Builder такую функцию:

```cpp
void demo()
{
    printf("DERIVED\n");
}
```

Последняя версия IDA сейчас 7.4, между тем я использую IDA 7.2, и она чаще всего успешно распознает почти любые функции. В нашем случае результат выглядит следующим образом:

```asm
.text:0000000140001000 void demo(void) proc near
.text:0000000140001000 sub  rsp, 28h
.text:0000000140001004 lea  rcx, _Format ; "DERIVED\n"
.text:000000014000100B call printf
.text:0000000140001010 add  rsp, 28h
.text:0000000140001014 retn
.text:0000000140001014 void demo(void) endp
```

То есть дизассемблер правильно распознал имя функции. Но так бывает далеко не всегда и не со всеми библиотечными функциями. А когда проблемы возникают с ними, кодокопателю анализировать становится сложновато. Бывает, сидишь, тупо уставившись в листинг дизассемблера, и никак не можешь понять: что же этот фрагмент делает? И только потом обнаруживаешь — одна или несколько функций опознаны неправильно!

Для уменьшения количества ошибок IDA пытается по стартовому коду распознать компилятор, подгружая только библиотеку его сигнатур. Из этого следует, что «ослепить» IDA очень просто — достаточно слегка видоизменить стартовый код. Поскольку он, по обыкновению, поставляется вместе с компилятором в исходных текстах, сделать это будет нетрудно. Впрочем, хватит и изменения одного байта в начале startup-функции. И все, хакер склеит ласты!

К счастью, в IDA предусмотрена возможность ручной загрузки базы сигнатур (FILELoad UleFLIRT signature Ule), но попробуй‑ка вручную определить, сигнатуры какой именно версии библиотеки требуется загружать! Наугад — слишком долго... Хорошо, если удастся визуально опознать компилятор. Опытным исследователям это обычно удается, так как каждый из них (компилятор, а не исследователь) имеет свой уникальный «почерк». К тому же существует принципиальная возможность использования библиотек из поставки одного компилятора в программе, скомпилированной другим компилятором. Словом, будь готов к тому, что в один прекрасный момент ты можешь столкнуться с необходимостью самостоятельно опознавать библиотечные функции. Решение этой задачи состоит из трех этапов. Первое — определение самого факта «библиотечности» функции, второе — определение происхождения библиотеки и третье — идентификация функции по этой библиотеке.

Учитывая, что линкер обычно располагает функции в порядке перечисления obj-модулей и библиотек, а большинство программистов указывают сначала собственные obj-модули, а библиотеки — потом (кстати, так же поступают и компиляторы, самостоятельно вызывающие линкер после окончания работы), можно заключить: библиотечные функции размещаются в конце программы, а собственно ее код — в начале. Конечно, из этого правила есть исключения, но все же срабатывает оно достаточно часто.

Рассмотрим, к примеру, структуру консольной версии общеизвестной программы pkzip.exe последней на данный момент 14-й версии. Она существенно легче, чем ее оконная версия. Ее вывод в консоли имеет следующий вид.

| ![Консольный вывод приложения pkzipc.exe](./images/12/pkzipc_cmd.png)
|:--:|
| *Консольный вывод приложения pkzipc.exe* |

Во время загрузки приложения в IDA та делает предположение, что это ROM для SNES.

Забавно, но выберем AMD64, если, конечно, установлена 64-разрядная версия утилиты.

| ![Загрузка pkzipc.exe в IDA](./images/12/pkzipc_ida.png)
|:--:|
| *Загрузка pkzipc.exe в IDA* |

На диаграмме, построенной IDA Pro 7.2, видно, что все библиотечные функции находятся в сегменте кода среди обычных функций до начала сегмента данных.

| ![Диаграмма для pkzipc.exe](./images/12/pkzipc_ida_legend.png)
|:--:|
| *Диаграмма для pkzipc.exe* |

Самое интересное: startup-функция в подавляющем большинстве случаев расположена в самом начале региона библиотечных функций или находится в непосредственной близости от него. Найти же саму startup не проблема — она совпадает с точкой входа в файл!

Таким образом, можно с высокой долей уверенности утверждать, что все функции, расположенные «ниже» startup (то есть в более старших адресах), — библиотечные. Посмотри, распознала их IDA или переложила эту заботу на тебя? Возможны два варианта: вообще никакие функции не распознаны или не распознана только часть функций.

Если не распознана ни одна функция, скорее всего, IDA не сумела опознать компилятор или использовались неизвестные ей версии библиотек. Для примера подсунем IDA другой популярный архиватор — peazip.exe, который, как известно, написан на объектном паскале.

| ![Диаграмма для peazip.exe](./images/12/peazip_ida_legend.png)
|:--:|
| *Диаграмма для peazip.exe* |

На диаграмме видно, что IDA не опознала ни одной библиотечной функции! Вот так новость! Смотрим, какой компилятор определен дизассемблером: «Plan FLIRT signature: SEH for vc64 7-14». Картина проясняется: IDA неправильно определила компилятор! Но что она скажет, если мы предложим ей pkzipw.exe (for Windows)?

| ![Диаграмма для pkzipw.exe](./images/12/pkzipw_ida_legend.png)
|:--:|
| *Диаграмма для pkzipw.exe* |

Гораздо лучше! Мы знаем, что препарируемое приложение написано на C++ и откомпилировано в Visual Studio. Поэтому IDA выдает сносные результаты. Между тем техника распознавания компиляторов — разговор особый, а вот распознание версий библиотек — это то, чем мы сейчас и займемся.

## ОПРЕДЕЛЕНИЕ БИБЛИОТЕК И ИХ ВЕРСИЙ

Прежде всего, многие из них содержат копирайты с указанием имени производителя и версии библиотеки, просто поищи текстовые строки в бинарном файле. Если их нет, не беда — ищем любые другие текстовые строки (как правило, сообщения об ошибках) и простым контекстным поиском пытаемся найти их во всех библиотеках, до которых удастся дотянуться (хакер должен иметь широкий набор компиляторов и библиотек на своем жестком диске или SSD-накопителе — это уж кому как больше нравится).

Возможны варианты: никаких других текстовых строк вообще нет; строки есть, но они неуникальны — обнаруживаются во многих библиотеках; наконец, искомый фрагмент нигде не обнаружен. В первых двух случаях следует выделить из одной (нескольких) библиотечных функций характерную последовательность байтов, не содержащую перемещаемых элементов, и вновь попытаться отыскать ее во всех доступных библиотеках. Если же это не поможет, то, увы, искомой библиотеки у тебя в наличии нет и положение патовое.

Патовое, да не совсем! Конечно, автоматически восстановить имена функций уже не удастся, но надежда на быстрое выяснение назначения функций все же есть. Имена API-функций Windows, вызываемые из библиотек, позволяют идентифицировать по крайней мере категорию библиотеки (например, работа с файлами, памятью, графикой). Математические же функции, по обыкновению, богаты командами сопроцессора.

Дизассемблирование очень похоже на разгадывание кроссворда (хотя не факт, что хакеры любят разгадывать кроссворды) — неизвестные слова угадываются за счет известных. Применительно к данному случаю в некоторых контекстах название функции вытекает из ее использования. Например, запрашиваем у пользователя пароль, передаем его функции X вместе с эталонным паролем, если результат завершения нулевой — пишем «пароль ОК», и, соответственно, наоборот. Не подсказывает ли твоя интуиция, что функция X не что иное, как strcmp? Конечно, это простейший случай, но, столкнувшись с незнакомой подпрограммой, никогда не спеши впадать в отчаяние и приходить в ужас от ее «монструозности». Просмотри все вхождения, обращая внимания, кто вызывает ее, когда и сколько раз.

Статистический анализ проливает свет на очень многое (функции, как и буквы алфавита, встречаются каждая со своей частотой), а контекстная зависимость дает пищу для размышлений. Так, функция чтения из файла не может предшествовать функции открытия!

Другие зацепки: аргументы и константы. Ну, с аргументами все более или менее ясно. Если функция получает строку, то это, очевидно, функция из библиотеки работы со строками, а если вещественное значение — возможно, функция математической библиотеки. Количество и тип аргументов (если их учитывать) весьма сужают круг возможных кандидатов. С константами же еще проще — очень многие функции принимают в качестве аргумента флаг, допускающий ограниченное количество значений. За исключением битовых флагов, которые все похожи друг на друга, довольно часто встречаются уникальные значения, пускай не однозначно идентифицирующие функцию, но все равно сужающие круг «подозреваемых». Да и сами функции могут содержать характерные константы. Скажем, встретив стандартный полином для подсчета CRC, можно быть уверенным, что «подследственная» вычисляет контрольную сумму...

Мне могут возразить: мол, все это частности. Возможно. Но, опознав часть функций, назначения остальных можно вычислить «от противного» и уж по крайней мере понять, что это за библиотека такая и где ее искать.

## ЗАКЛЮЧЕНИЕ

Идентификацию алгоритмов (то есть назначения функции) очень сильно облегчает знание этих самих алгоритмов. В частности, код, выполняющий LZ-сжатие (распаковку), настолько характерен, что узнается с первого взгляда, достаточно знать этот механизм упаковки. Напротив, если не иметь о нем никакого представления, ох, и нелегко же будет анализировать программу! Зачем изобретать колесо, если можно взять уже готовое? Хоть и бытует мнение, что хакер в первую очередь хакер, а уж потом программист (да и зачем ему уметь программировать?), в жизни все наоборот — программист, не умеющий программировать, проживет — вокруг уйма библиотек, воткни — и заработает! Хакеру же знание основ необходимо, без этого далеко не уплывешь (разумеется, отломать серийный номер у софтины можно и без высшей математики).

Понятное дело, библиотеки как раз на то и создавались, чтобы избавить разработчиков от необходимости вникать в те предметные области, без которых им и так хорошо. Увы, у исследователей программ нет простых путей, приходится думать и руками, и головой, и даже... пятой точкой опоры вкупе со спинным мозгом, только так и дизассемблируются серьезные программы. Бывает, готовое решение приходит в поезде или во сне...

Анализ библиотечных функций — это сложнейший аспект дизассемблирования, и просто замечательно, когда есть возможность идентифицировать их имена по сигнатурам.
