# Соглашение о быстрых вызовах — fastcall

> На плат­форме x64 фак­тичес­ки сущес­тву­ет толь­ко одно сог­лашение вызовов фун­кций — fastcall. Судя по дизас­сем­блер­ным лис­тингам, ком­пилятор при генера­ции кода авто­мати­чес­ки меня­ет любое дру­гое про­писан­ное прог­раммис­том сог­лашение на него. В этой статье мы раз­берем­ся, что такое fastcall, рас­смот­рим исполь­зуемые в этом сог­лашении парамет­ры и дру­гие важ­ные понятия.

Как сле­дует из наз­вания, fastcall пред­полага­ет быс­трый вызов фун­кции. Дру­гими сло­вами, при его исполь­зовании парамет­ры переда­ются через регис­тры про­цес­сора, что отра­жает­ся на ско­рос­ти работы под­прог­рамм. Меж­ду тем во вре­мена x86 fastcall не был стан­дарти­зиро­ван, что соз­давало немало труд­ностей прог­раммис­ту.

В ори­гиналь­ном изда­нии кни­ги «Фун­дамен­таль­ные осно­вы хакерс­тва» Крис в свой­ствен­ной ему манере очень под­робно опи­сал механиз­мы переда­чи парамет­ров с помощью регис­тров для 32-бит­ных ком­пилято­ров. Он упо­мянул C/C++ и Turbo Pascal от раз­личных фирм‑раз­работ­чиков, таких как Microsoft, Borland, Watcom (эта ком­пания ныне ско­рее мер­тва, чем жива, одна­ко откры­тый про­ект их ком­пилято­ра мож­но най­ти на [GitHub](https://github.com/open-watcom/open-watcom-v2)). Крис под­робно опи­сал переда­чу как целых, так и вещес­твен­ных типов дан­ных: int, long, float, single, double, extended, real, etc.

С тех пор как бал пра­вит архи­тек­тура x86_64, обо всех сог­лашени­ях переда­чи парамет­ров про­цеду­рам и фун­кци­ям мож­но забыть, как о страш­ном сне. Теперь в этой сфе­ре все уста­кани­лось и ста­ло стан­дарти­зиро­вано, чего никак не мог­ли добить­ся во вре­мена x86.

Бла­года­ря воз­росше­му количес­тву регис­тров на плат­форме x64 для ком­пилято­ров C/C++ име­ется толь­ко одно сог­лашение вызова. Пер­вые четыре целочис­ленных парамет­ра или ука­зате­ля переда­ются в регис­трах `RCX`, `RDX`, `R8`, `R9`. Для C++ в подав­ляющем боль­шинс­тве вызовов методов пер­вый целочис­ленный параметр занима­ет ука­затель `this`. Пер­вые четыре парамет­ра, пред­став­ленные зна­чени­ями с пла­вающей запятой (вещес­твен­ные зна­чения), переда­ются в регис­трах `XMM0`, `XMM1`, `XMM2`, `XMM3` рас­ширения SSE. На 64-бит­ной плат­форме набор это­го рас­ширения был уве­личен с 8 до 16 регис­тров.

До­пол­нитель­ные парамет­ры, если они есть, переда­ются через стек. Вызыва­ющая фун­кция резер­виру­ет мес­то в сте­ке, а вызыва­емая может раз­местить там передан­ные в регис­трах перемен­ные.

| ![Регистры процессора архитектуры x86_64](./images/15/1b8b36d1067a6085d9b40060d1925349.png)
|:--:|
| *Регистры процессора архитектуры x86_64* |

## ИДЕНТИФИКАЦИЯ ПЕРЕДАЧИ И ПРИЕМА РЕГИСТРОВ

Пос­коль­ку вызыва­емая и вызыва­ющая фун­кции вынуж­дены при­дер­живать­ся общих сог­лашений при переда­че парамет­ров через регис­тры, ком­пилято­ру при­ходит­ся помещать парамет­ры в те регис­тры, в которых их ожи­дает вызыва­емая фун­кция, а не туда, куда ему «удоб­но». В резуль­тате перед каж­дой соот­ветс­тву­ющей сог­лашению fastcall фун­кци­ей появ­ляет­ся код, «тасу­ющий» содер­жимое регис­тров стро­го опре­делен­ным обра­зом — задан­ным стан­дартом для x64 (мы рас­смот­рели его в пре­дыду­щем раз­деле).

При ана­лизе кода вызыва­ющей фун­кции не всег­да мож­но рас­познать переда­чу парамет­ров через регис­тры (если толь­ко их ини­циали­зация не будет слиш­ком наг­лядна). Поэто­му при­ходит­ся обра­щать­ся непос­редс­твен­но к вызыва­емой фун­кции. Регис­тры, которые фун­кция сох­раня­ет в сте­ке сра­зу пос­ле получе­ния управле­ния, не переда­ют парамет­ры, и из спис­ка «кан­дидатов» их мож­но вычер­кнуть.

Есть ли сре­ди оставших­ся такие, содер­жимое которых исполь­зует­ся без явной ини­циали­зации? В пер­вом приб­лижении фун­кция при­нима­ет парамет­ры имен­но через эти регис­тры. При деталь­ном же рас­смот­рении проб­лемы всплы­вает нес­коль­ко ого­ворок. Во‑пер­вых, через регис­тры могут переда­вать­ся (и очень час­то переда­ются) неяв­ные парамет­ры фун­кции — ука­затель `this`, ука­зате­ли на вир­туаль­ные таб­лицы объ­екта и так далее. Во‑вто­рых, если кри­вору­кий прог­раммист, наде­ясь, что зна­чение перемен­ной пос­ле объ­явле­ния дол­жно быть рав­но нулю, забыва­ет об ини­циали­зации, а ком­пилятор помеща­ет перемен­ную в регистр, то при ана­лизе прог­раммы она может быть при­нята за переда­ваемый через регистр параметр фун­кции.

Са­мое инте­рес­ное, что этот регистр может по слу­чай­ному сте­чению обсто­ятель­ств явно ини­циали­зиро­вать­ся вызыва­ющей фун­кци­ей. Пред­ста­вим, что прог­раммист перед этим вызывал фун­кцию, воз­вра­щаемо­го зна­чения которой не исполь­зовал. Ком­пилятор помес­тил неини­циали­зиро­ван­ную перемен­ную в `RAX`. При­чем, если фун­кция при сво­ем нор­маль­ном завер­шении воз­вра­щает ноль (как час­то и быва­ет), все может работать... Что­бы выловить этот баг, иссле­дова­телю при­дет­ся про­ана­лизи­ровать алго­ритм и выяс­нить, дей­стви­тель­но ли в `RAX` находит­ся код успешно­го завер­шения фун­кции, или же име­ет мес­то наложе­ние перемен­ных? Впро­чем, если отки­нуть кли­ничес­кие слу­чаи, переда­ча аргу­мен­тов через регис­тры не силь­но усложня­ет ана­лиз, в чем мы сей­час и убе­дим­ся.

## ПРАКТИЧЕСКОЕ ИССЛЕДОВАНИЕ МЕХАНИЗМА ПЕРЕДАЧИ АРГУМЕНТОВ ЧЕРЕЗ РЕГИСТРЫ

Для зак­репле­ния все­го ска­зан­ного давай рас­смот­рим сле­дующий при­мер:

```cpp
#include <stdio.h>
#include <string>

// Функция MyFunc с различными типами аргументов для демонстрации механизма
// их передачи
int MyFunc(char a, int b, long int c, int d)
{
    return a + b + c + d;
}
int main()
{
    printf("%x\n", MyFunc(0x1, 0x2, 0x3, 0x4));
    return 0;
}
```

| ![Исходник fastcall_4_args](./images/15/159a83a6a3184750281ebfd2041f1235.png)
|:--:|
| *Исходник fastcall_4_args* |

| ![Вывод приложения — сумма чисел в шестнадцатеричном формате](./images/15/981680c43bf62f1330f68fc35f67af83.png)
|:--:|
| *Вывод приложения — сумма чисел в шестнадцатеричном формате* |

Пе­ред пос­тро­ением это­го при­мера отклю­чи опти­миза­цию. Резуль­тат его обра­бот­ки ком­пилято­ром Microsoft Visual C++ 2019 дол­жен выг­лядеть так:

```asm
main proc near
sub rsp, 28h
```

Все аргу­мен­ты помеща­ются в регис­тры. Судя по их зна­чени­ям, в обратном поряд­ке. Ради нез­начитель­ной опти­миза­ции ком­пилятор решил не задей­ство­вать регис­тры пол­ностью, а, зная типы дан­ных наперед, исполь­зовать толь­ко необ­ходимое прос­транс­тво. Таким обра­зом, вмес­то того, что­бы выделять регис­тры целиком: `R9`, `R8`, `RDX`, `RCX`, были отда­ны толь­ко полови­ны пер­вых трех (`R9D`, `R8D`, `EDX`) и лишь вось­мая часть пос­ледне­го — регистр `CL`.

```asm
mov r9d, 4 ; d
mov r8d, 3 ; c
mov edx, 2 ; b
mov cl, 1 ; a
```

В ито­ге IDA без нашей помощи вос­ста­нови­ла про­тотип вызыва­емой фун­кции. Но если бы у нее не получи­лось это сде­лать? Тог­да бы мы гадали: `Long`, как и `int`, занима­ет 32 бита, `char` — единс­твен­ный тип дан­ных, занима­ющий один байт.

```asm
call MyFunc(char,int,long,int)
mov edx, eax
```

Не­важ­но, ког­да ты будешь читать этот текст, воз­можно, во вре­мена Windows 17 и Intel Core i9, одна­ко сис­темные прог­раммис­ты могут изме­нить раз­меры базовых типов дан­ных в соот­ветс­твии с архи­тек­турой вычис­литель­ных сис­тем. Что­бы узнать их раз­мер кон­крет­но на тво­ей машине, мож­но вос­поль­зовать­ся такой незамыс­ловатой прог­раммой:

```cpp
#include <iostream>
const int byte = 8;
int main()
{
    std::cout << "int = " << sizeof(int) * byte << '\n';
    std::cout << "long = " << sizeof(long) * byte << '\n';
    std::cout << "bool = " << sizeof(bool) * byte << '\n';
    std::cout << "float = " << sizeof(float) * byte << '\n';
    std::cout << "double = " << sizeof(double) * byte << '\n';
    // и так далее нужные тебе типы данных
    return 0;
}
```

По­лучен­ный при выпол­нении фун­кции `MyFunc` резуль­тат (сум­ма парамет­ров) переда­ется фун­кции `printf` для вывода на кон­соль в шес­тнад­цатерич­ном виде:

```asm
lea  rcx, _Format ; "%x\n"
call printf
xor  eax, eax
add  rsp, 28h
retn
main endp
```

Ди­зас­сем­блер­ный лис­тинг фун­кции `MyFunc` выг­лядит сле­дующим обра­зом:

```asm
int MyFunc(char, int, long, int) proc near
arg_0  = byte ptr 8
arg_8  = dword ptr 10h
arg_10 = dword ptr 18h
arg_18 = dword ptr 20h
```

Сна­чала фун­кция раз­матыва­ет содер­жимое регис­тров по сте­ку. Обра­ти вни­мание, как она это дела­ет: в аргу­мен­тах заданы раз­меры. Они при­бав­ляют­ся к отри­цатель­ному зна­чению вер­шины сте­ка — `RSP`, и по получив­шемуся адре­су в сте­ке кла­дет­ся зна­чение из соот­ветс­тву­юще­го по раз­меру регис­тра.

```asm
mov [rsp+arg_18], r9d
mov [rsp+arg_10], r8d
mov [rsp+arg_8], edx
mov [rsp+arg_0], cl
movsx eax, [rsp+arg_0]
```

Пос­ле это­го в регис­тре `EAX` накап­лива­ется сум­ма всех шес­тнад­цатерич­ных зна­чений, передан­ных в фун­кцию `MyFunc`, и в него же воз­вра­щает­ся резуль­тат:

```asm
add eax, [rsp+arg_8]
add eax, [rsp+arg_10]
add eax, [rsp+arg_18]
retn
int MyFunc(char, int, long, int) endp
```

А теперь пос­мотрим, что сге­нери­ровал C++Builder 10.3. Сна­чала main:

```asm
main proc near
var_14 = dword ptr -14h
var_10 = qword ptr -10h
var_8  = dword ptr -8
var_4  = dword ptr -4
sub    rsp, 38h
```

Здесь мы видим, что пос­ле ини­циали­зации сте­ка ком­пилятор помеща­ет парамет­ры в регис­тры в пря­мом поряд­ке, то есть в том, в котором их передал прог­раммист в ори­гиналь­ной прог­рамме на язы­ке высоко­го уров­ня.

```asm
mov eax, 1
mov r8d, 2
mov r9d, 3
mov r10d, 4
```

Пос­ле это­го, по сути, мож­но вызывать сле­дующую фун­кцию, переда­вая парамет­ры в регис­трах. Имен­но это Visual C++ и делал. Тем не менее C++Builder нагоро­дил допол­нитель­ного кода: он заг­ружа­ет зна­чения регис­тров в стек, как бы обме­нивая их зна­чения, но в ито­ге все рав­но вызыва­ет `MyFunc` и переда­ет четыре парамет­ра в регис­трах. Как же это неоп­тималь­но!

```asm
mov  [rsp+38h+var_4], 0
mov  [rsp+38h+var_8], ecx
mov  [rsp+38h+var_10], rdx
mov  ecx, eax ; int
mov  edx, r8d ; __int64
mov  r8d, r9d
mov  r9d, r10d
call MyFunc(char,int,long,int)
lea  rcx, unk_44A000
mov  edx, eax
; Возвращаемое значение выводим на консоль
call printf
mov  [rsp+38h+var_4], 0
mov  [rsp+38h+var_14], eax
mov  eax, [rsp+38h+var_4]
; Обрати внимание: этот компилятор еще и сам чистит стек
add  rsp, 38h
retn
main endp
```

За­тем `MyFunc`:

```asm
MyFunc(char, int, long, int) proc near
var_14 = dword ptr -14h
var_10 = dword ptr -10h
var_C  = dword ptr -0Ch
var_8  = dword ptr -8
var_1  = byte ptr -1
```

И тут C++Builder нагоро­дил лиш­ний код. Пос­ле ини­циали­зации сте­ка зна­чение 8-бит­ного регис­тра `CL` копиру­ется в `AL`. Понят­но, что 8-бит­ным явля­ется пер­вый переда­ваемый параметр типа `char`. Но если мы заг­лянем в `main`, то обна­ружим, что отту­да не переда­ется 8-бит­ный параметр. Меж­ду тем переда­ется 32-бит­ный `ECX: mov ecx, eax`. И уже в `MyFunc` берет­ся толь­ко чет­вертая часть от `ECX` и помеща­ется в `AL`.

```asm
sub rsp, 18h
mov al, cl
; Помещает значения аргументов в стек с учетом размеров переменных
mov [rsp+18h+var_1], al
mov [rsp+18h+var_8], edx
mov [rsp+18h+var_C], r8d
mov [rsp+18h+var_10], r9d
```

Те­перь, узнав раз­меры парамет­ров, можем вывес­ти про­тотип (сов­ремен­ная IDA справ­ляет­ся с этим без нашей помощи, но нам ведь тоже это надо уметь):

```cpp
MyFunc(char a, int b, int c, int d)
```

Од­нако порядок сле­дова­ния трех пос­ледних аргу­мен­тов может быть иным, так­же сто­ит учи­тывать, что в Windows тип `Long` тоже 32-бит­ный и он тоже пред­назна­чен для обра­бот­ки целочис­ленных зна­чений. Сле­дова­тель­но, исполь­зует те же регис­тры про­цес­сора.

Да­лее к находя­щему­ся в регис­тре `ECX` зна­чению пос­тепен­но при­бав­ляет­ся каж­дое зна­чение аргу­мен­та, ранее помещен­ное в стек:

```asm
movsx ecx, [rsp+18h+var_1]
add   ecx, [rsp+18h+var_8]
add   ecx, [rsp+18h+var_C]
add   ecx, [rsp+18h+var_10]
mov   [rsp+18h+var_14], ecx
; В регистре EAX возвращаем результат
mov   eax, [rsp+18h+var_14]
; Обнуляем стек
add   rsp, 18h
retn
MyFunc(char, int, long, int) endp
```

К сло­ву, ком­пиляция выпол­нена из‑под све­жень­кой Windows 11, одна­ко это никак не пов­лияло на дизас­сем­блер­ный лис­тинг.

| ![Выполнение приложения командной строки в Windows 11](./images/15/cb2a51b87891b05f028537a9593bf472.png)
|:--:|
| *Выполнение приложения командной строки в Windows 11* |

Как мы видим, в переда­че парамет­ров через регис­тры ничего слож­ного нет. Мож­но, даже не при­бегая к помощи IDA, вос­ста­новить под­линный про­тотип вызыва­емой фун­кции. Одна­ко рас­смот­ренная выше ситу­ация иде­али­зиро­вана, и в реаль­ных прог­раммах переда­ча одних лишь непос­редс­твен­ных зна­чений встре­чает­ся ред­ко. Осво­ившись с быс­тры­ми вызова­ми, давай дизас­сем­бли­руем более труд­ный при­мер:

```cpp
#include <stdio.h>
#include <string.h>

int MyFunc(char a, int* b, int c)
{
    return a + b[0] + c;
}

int main()
{
    int a = 2;
    printf("%x\n", MyFunc(strlen("1"), &a, strlen("333")));
}
```

Ре­зуль­тат его ком­пиляции в Visual C++ 2019 дол­жен выг­лядеть так:

```asm
main proc near
; Объявляем переменные
b      = dword ptr -18h
var_10 = qword ptr -10h
; Инициализация стека
sub rsp, 38h
mov rax, cs:__security_cookie
xor rax, rsp
mov [rsp+38h+var_10], rax
; Присваиваем переменной b типа int значение 2
mov [rsp+38h+b], 2
```

Эге‑гей! Даже с отклю­чен­ной опти­миза­цией ком­пилятор не стал дваж­ды вызывать фун­кцию `strlen`, а сра­зу под­ста­вил вычис­ленные резуль­таты:

```asms
mov r8d, 3 ; c — очевидно, это третий параметр — strlen("333"), то есть число 3
lea rdx, [rsp+38h+b]; b — указатель на переменную b
mov cl, 1 ; a — первый параметр, тип char, число 1, результат strlen("1")
```

Па­рамет­ры фун­кции, как и положе­но, переда­ются в обратном поряд­ке. По име­ющей­ся информа­ции, даже без помощи IDA мы можем зап­росто вос­ста­новить про­тотип фун­кции: `MyFunc(char,int *,int)`.

```asm
call MyFunc(char,int *,int)
; Передаются три аргумента, а возвращается один — в регистре EAX
mov edx, eax
lea rcx, _Format ; "%x\n"
```

Пос­ле заряд­ки фор­матной стро­ки резуль­тат отправ­ляет­ся на печать. Затем регистр `EAX` обну­ляет­ся. Стек вос­ста­нав­лива­ется и деини­циали­зиру­ется.

```asm
call printf
xor  eax, eax
mov  rcx, [rsp+38h+var_10]
xor  rcx, rsp ; StackCookie
call __security_check_cookie
add  rsp, 38h
retn
```

Оз­накомим­ся с дизас­сем­блер­ным лис­тингом фун­кции `MyFunc`:

```asm
int MyFunc(char, int *, int) proc near ; CODE XREF: main+28↓p
arg_0  = byte ptr 8
arg_8  = qword ptr 10h
arg_10 = dword ptr 18h
```

Фун­кция при­нима­ет три аргу­мен­та в регис­трах по пра­вилам fastcall и раз­меща­ет их в сте­ке.

```asm
mov    [rsp+arg_10], r8d
mov    [rsp+arg_8], rdx
mov    [rsp+arg_0], cl
; Переменная var_0 расширяется до знакового целого (signed int)
movsx  eax, [rsp+arg_0]
mov    ecx, 4
; Умножением на 0 обнуляем RCX — странновато, но как вариант
imul   rcx, 0
; Значение переменной arg_8 помещаем в регистр RDX
mov    rdx, [rsp+arg_8]
```

Бе­рем зна­чение из ячей­ки по адре­су `[rdx+rcx]` и скла­дыва­ем с содер­жимым регис­тра `EAX`, в котором было сох­ранено зна­чение перемен­ной `arg_0`. Затем переза­писы­ваем этот регистр.

```asm
add eax, [rdx+rcx]
; Значение переменной arg_10 суммируем со значением в регистре EAX
add eax, [rsp+arg_10]
; В регистре EAX возвращаем результат
retn
int MyFunc(char, int *, int) endp
```

Прос­то? Прос­то! Тог­да рас­смот­рим резуль­тат твор­чес­тва C++Builder (обновлен­ного до 10.4.2 Sydney — это пос­ледняя на вре­мя написа­ния дан­ных строк вер­сия). Код дол­жен выг­лядеть так:

```asm
main proc near
; Объявление переменных
var_28 = dword ptr -28h
var_21 = byte ptr -21h
var_20 = dword ptr -20h
var_14 = dword ptr -14h
var_10 = qword ptr -10h
var_8  = dword ptr -8
var_4  = dword ptr -4
push rbp
sub rsp, 50h
```

От­кры­ваем кадр сте­ка, затем копиру­ем в регистр `RBP` ука­затель на вер­шину сте­ка, а в регистр `RAX` — ука­затель на перемен­ную `var_14`.

```asm
lea rbp, [rsp+50h]
lea rax, [rbp+var_14]
mov [rbp+var_4], 0 ; Переменную var_4 инициализируем нулем
mov [rbp+var_8], ecx
mov [rbp+var_10], rdx
mov [rbp+var_14], 2 ; Переменную var_14 инициализируем значением 2
```

Ес­ли уста­новить кур­сор мыши на мет­ку `unk_44D004`, появит­ся всплы­вающая под­сказ­ка, где будет ука­зано, что мет­ка рав­на `0x31`. При этом в ком­мента­рии ука­зана еди­ница. С чего бы это вдруг? Каким боком еди­ница свя­зана с `0х31`?

В десятич­ной сис­теме `0х31` будет 49. Вспо­мина­ем таб­лицу ASCII, где сим­вол под номером 49 и есть циф­ра 1. Ста­новит­ся понят­но, отку­да у оле­ня рога рас­тут. Фун­кции `srtlen` в качес­тве парамет­ра переда­ется ука­затель на стро­ку '1'. Ука­затель явля­ется целым зна­чени­ем, а, как мы пом­ним, по сог­лашению fastcall в x64 пер­вое целое зна­чение переда­ется в регис­тре `RCX`.

```asm
lea rcx, unk_44D004 ; 666
```

Вспо­мина­ем, что хра­нит­ся в `RAX`, — ука­затель на перемен­ную `var_14`. С этим понят­но. `RAX` — 64-бит­ный регистр, сле­дова­тель­но, адрес при­емни­ка тоже дол­жен быть 64-бит­ным, на что ука­зыва­ет клю­чевое сло­во `qword`. `Ptr` говорит ассем­бле­ру, что от перемен­ной `var_20` нуж­но не зна­чение, а адрес. Таким обра­зом, получа­ется «ука­затель на ука­затель на зна­чение 2».

```asm
mov qword ptr [rbp+var_20], rax
call strlen
```

По сог­лашению fastcall на 64-бит­ной плат­форме фун­кция воз­вра­щает резуль­тат в регис­тре `RAX`, а тут, оче­вид­но, резуль­тат нас­толь­ко мал, что для него хва­тило 8-бит­ного регис­тра `AL`. Естес­твен­но, дли­на стро­ки "1" сос­тавля­ет 1.

```asm
mov r8b, al
```

А здесь уже зна­комым обра­зом берет­ся ука­затель на стро­ку из сег­мента дан­ных, подоб­но тому как это было показа­но на при­мере стро­ки с мет­кой 666.

| ![Сегмент данных](./images/15/2330c44713c83024ed3f5cf31d2b6faa.png)
|:--:|
| *Сегмент данных* |

```asm
lea rcx, unk_44D006
```

Пе­ремен­ная `var_21` объ­явле­на раз­мером `byte`, соот­ветс­твен­но, она как раз под­ходит для 8-бит­ного зна­чения, хра­няще­гося в регис­тре `R8B`.

```asm
mov [rbp+var_21], r8b
call strlen
```

В дан­ном слу­чае фун­кция, похоже, вер­нула боль­шее зна­чение и прог­рамме понадо­бил­ся 32-бит­ный регистр `EAX`. Безус­ловно, дли­на стро­ки "333" рав­на 3... И на фига нам 32 бита? Ком­пилятор нем­ного перес­тарал­ся.

Ес­ли прис­мотреть­ся, далее мы уви­дим под­готов­ку регис­тров для переда­чи аргу­мен­тов. Как мы пом­ним, C++Builder испо­кон веков переда­ет аргу­мен­ты сле­ва нап­раво.

```asm
mov r9d, eax
mov r8b, [rbp+var_21]
```

Оче­вид­но, 8-бит­ная перемен­ная `var_21` на язы­ке высоко­го уров­ня будет иметь тип дан­ных `char`. Находя­щееся в ней зна­чение перед вызовом фун­кции `MyFunc` занима­ет мес­то в регис­тре `ECX`, из чего мож­но сде­лать вывод, что этот аргу­мент будет передан пер­вым.

```asm
movzx ecx, r8b
```

Да­лее в регистр `RDX` помеща­ется ука­затель на зна­чение 2. Так как `RDX` — вто­рой регистр в оче­реди переда­чи, то он будет содер­жать вто­рой аргу­мент.

```asm
mov rdx, qword ptr [rbp+var_20] ; int
```

За­тем по спис­ку для переда­чи исполь­зует­ся регистр `R8`, но от него нам нуж­на толь­ко полови­на — `R8D` (32 бита). На дан­ном шаге мы вла­деем всей информа­цией, что­бы вос­ста­новить про­тотип вызыва­емой кас­томной фун­кции. Поэто­му, не под­гля­дывая в дизас­сем­бли­рован­ный код IDA, выведем про­тотип сами: `MyFunc(char,int *,int)`.

```asm
mov r8d, r9d
call MyFunc(char,int *,int)
```

В `RCX` из сег­мента дан­ных копиру­ем стро­ку `"%х{символ начала новой строки}"`. Пос­ле чего переда­ем фун­кции printf два запол­ненных регис­тра: `RCX`, `EDX`, что­бы та извлек­ла аргу­мен­ты и вывела их на кон­соль.

```asm
lea    rcx, unk_44D000
mov    edx, eax
call   printf
; Остальное — закат функции известен и нас мало интересует
mov    [rbp+var_4], 0
mov    [rbp+var_28], eax
mov    eax, [rbp+var_4]
add    rsp, 50h
pop    rbp
retn
main endp
```

Меж­ду про­чим, не сто­ит слиш­ком полагать­ся на раз­меры перемен­ных: иног­да ком­пилятор может сге­нери­ровать такой код, что раз­меры не выручат. Разум­нее полагать­ся на здра­вый смысл. Кро­ме того, навер­няка ты обра­тил вни­мание, что целевой регистр для переда­чи аргу­мен­та выбира­ется заведо­мо боль­ше, чем нуж­но. Это дела­ется в уго­ду опти­миза­ции. Поэто­му при­ходит­ся рас­кру­чивать дизас­сем­блер­ный лис­тинг к самой ини­циали­зации перемен­ной, где ком­пилятор работа­ет с реаль­ными, а не опти­маль­ными раз­мерами. Ну и задач­ку нам под­кинул Embarcadero C++Builder! Ник­то не обе­щал, что будет лег­ко.

Ди­зас­сем­блер­ный лис­тинг фун­кции `main`, изго­тов­ленный ком­пилято­ром C++Builder, получил­ся в два раза объ­емнее, чем тот же, но получен­ный из‑под ком­пилято­ра Visual C++.

| ![Наглядное сравнение размеров функций main](./images/15/625ec346523e4df8970767b563e6a056.png)
|:--:|
| *Наглядное сравнение размеров функций main* |

| ![ASCII Table](./images/15/190856c13ee779020b918602cfe56fbc.png)
|:--:|
| *ASCII Table* |

Но это еще толь­ко пол­беды! Даль­ше нас ждет фун­кция `MyFunc`. Пер­вое, что бро­сает­ся в гла­за при взгля­де на код этой фун­кции, — для обра­щения к локаль­ным перемен­ным вмес­то вер­шины сте­ка (регис­тра `RSP`) исполь­зует­ся регистр `RBP`. Что это зна­чит? Офи­циаль­ная докумен­тация говорит нам, что `RBP` — ука­затель на кадр сте­ка, то есть на его текущую кон­фигура­цию. Поэто­му боль­шинс­тво ком­пилято­ров исполь­зуют регистр `RSP` для улуч­шения про­изво­дитель­нос­ти генери­руемо­го кода. Но, пос­коль­ку в нашем слу­чае опти­миза­ция отклю­чена, ком­пилятор не гонит­ся за ско­ростью и исполь­зует более мед­ленные решения.

```asm
MyFunc(char, int *, int) proc near ; CODE XREF: main+58↓p
; Объявление переменных
var_18 = dword ptr -18h ; 32
var_14 = dword ptr -14h ; 32
var_10 = qword ptr -10h ; 64
var_1  = byte ptr -1 ; 8
; Запоминаем кадр стека
push rbp
sub  rsp, 18h
lea  rbp, [rsp+10h]
; Открываем кадр стека.
; В 8-битном регистре CL был передан аргумент char
mov al, cl ; Копируем находящееся в нем значение в регистр AL
mov [rbp+8+var_1], al ; Затем копируем значение в переменную var_1
; Как мы помним, там находится число 1
```

В `RDX` переда­ется ука­затель на 2, его помеща­ем в 64-раз­рядную перемен­ную `var_10`.

```asm
mov [rbp+8+var_10], rdx
```

В 32-бит­ном регис­тре `R8D` переда­лось зна­чение 3. Оно копиру­ется в перемен­ную `var_14`.

```asm
mov [rbp+8+var_14], r8d
; Переменную размером в байт расширяем до знакового двойного слова
movsx ecx, [rbp+8+var_1]
; Ранее скопированный указатель вновь переносим в RDX
mov rdx, [rbp+8+var_10]
```

При­бав­ляем к содер­жимому `ECX` (к зна­чению 1) содер­жимое `RDX` (ука­затель на ука­затель на 2, про­ще говоря, прос­то 2).

```asm
add ecx, [rdx]
```

На сле­дующем шаге выпол­няет­ся оче­ред­ное сло­жение: к накоп­ленной в регис­тре `ECX` сум­ме (3) при­бав­ляет­ся зна­чение перемен­ной `var_14` (3). Сле­дующая инс­трук­ция помеща­ет сум­му в перемен­ную `var_18`, затем — в регистр `EAX`, и уже он воз­вра­щает­ся выз­вавшей фун­кции `main`, которая выведет его содер­жимое на кон­соль.

```asm
add ecx, [rbp+8+var_14]
mov [rbp+8+var_18], ecx
mov eax, [rbp+8+var_18]
add rsp, 18h
pop rbp
; Закрываем кадр стека
retn
MyFunc(char, int *, int) endp
```

## ПЕРЕДАЧА ВЕЩЕСТВЕННЫХ ЗНАЧЕНИЙ

В знак ува­жения к Кри­су нель­зя не упо­мянуть, о чем было ска­зано в этом раз­деле ори­гиналь­ного изда­ния. Дело было дав­нее, прог­раммис­ты выжива­ли, как мог­ли, и для работы с вещес­твен­ными зна­чени­ями исполь­зовали соп­роцес­сор, работав­ший на одном крис­талле с CPU. Он имел свой набор команд, а при выпол­нении прог­раммы для него соз­давал­ся отдель­ный стек. Как раз для переда­чи аргу­мен­тов, пред­став­ляющих вещес­твен­ные зна­чения: float, double, long double, исполь­зовал­ся стек соп­роцес­сора.

Чуть поз­днее в про­цес­соре Pentium III и всех пос­леду­ющих моделях кор­пораци­ей Intel был пред­став­лен набор инс­трук­ций SSE. Вмес­те с новыми инс­трук­циями рас­ширение вклю­чило в архи­тек­туру восемь новых 128-бит­ных регис­тров для 32-бит­ных про­цес­соров и впос­ледс­твии 16 128-бит­ных регис­тров для 64-бит­ных про­цес­соров.

У сов­ремен­ных прог­раммис­тов проб­лем нам­ного мень­ше, и кодоко­пате­лям мень­ше стра­дать, раз­гры­зая их тво­рения, потому что пер­вые четыре вещес­твен­ных зна­чения переда­ются в регис­трах `XMM0` — `XMM3`. Если аргу­мен­тов боль­ше, то оставши­еся переда­ются через стек. Прос­то и понят­но.

Приш­ло вре­мя разоб­рать неболь­шой при­мер переда­чи вещес­твен­ных зна­чений.

```cpp
#include <stdio.h>
float MyFunc(float a, double b)
{
    return a + b;
}

int main()
{
    printf("%f\n", MyFunc(6.66, 7.77));
}
```

Ре­зуль­тат ком­пиляции Microsoft Visual C++ дол­жен выг­лядеть так:

```asm
main proc near
; Инициализация стека
sub  rsp, 28h
```

Под­готав­лива­ем парамет­ры спра­ва налево (Visual C++ всег­да так дела­ет). Берем зна­чение для ска­ляра `b` из сег­мента дан­ных: 7.77 — зна­чение двой­ной точ­ности, помеща­ем его в регистр `XMM1` для пос­леду­ющей переда­чи в качес­тве парамет­ра фун­кции `MyFunc`.

Та­кое же дей­ствие про­вора­чива­ем со ска­ляром `a`, его зна­чение (оно, нап­ротив, оди­нар­ной точ­ности) записы­ваем в регистр `XMM0`. Под­готов­ленные зна­чения в качес­тве аргу­мен­тов переда­ются фун­кции `MyFunc` при ее вызове.

```asm
movsd xmm1, cs:__real@401f147ae147ae14 ; b
movss xmm0, cs:__real@40d51eb8 ; a
call  MyFunc(float,double)
```

`MyFunc` воз­вра­щает вещес­твен­ный резуль­тат в регис­тре `XMM0`, а сле­дующая инс­трук­ция из рас­ширения SSE кон­верти­рует зна­чение оди­нар­ной точ­ности в зна­чение двой­ной точ­ности и сох­раня­ет его в том же регис­тре.

```asm
cvtss2sd xmm0, xmm0
```

Да­лее про­исхо­дит пооче­ред­ное копиро­вание двух час­тей 128-бит­ного зна­чения в регистр `XMM1`.

```asm
movaps xmm1, xmm0
```

Сле­дующая инс­трук­ция «ужи­мает» зна­чение со 128 бит до 64, помещая его в соот­ветс­тву­ющий регистр.

```asm
movq rdx, xmm1
```

За­даем фор­мат вывода.

```asm
lea  rcx, _Format ; "%f\n"
call printf
xor  eax, eax
; Очистка стека
add  rsp, 28h
retn
main endp
```

Вспо­мога­тель­ная фун­кция `MyFunc`:

```asm
float MyFunc(float, double) proc near
; IDA обнаружила два аргумента
arg_0 = dword ptr 8
arg_8 = qword ptr 10h
; Размещаем их в памяти:
; второй — удвоенной точности
movsd [rsp+arg_8], xmm1
; первый — одинарной точности
movss [rsp+arg_0], xmm0
```

Эта фун­кция кон­верти­рует зна­чение оди­нар­ной точ­ности в зна­чение двой­ной точ­ности и сох­раня­ет его в регис­тре `XMM0`.

```asm
cvtss2sd xmm0, [rsp+arg_0]
```

Сум­миру­ем вещес­твен­ное зна­чение с низ­кой двой­ной точ­ностью из памяти со зна­чени­ем с низ­кой двой­ной точ­ностью из регис­тра `XMM0` и сох­раня­ем резуль­тат на его мес­те.

```asm
addsd xmm0, [rsp+arg_8]
```

Здесь про­исхо­дит обратное пре­обра­зова­ние — из зна­чения двой­ной точ­ности в зна­чение оди­нар­ной точ­ности:

```asm
cvtsd2ss xmm0, xmm0
; Возвращаем результат вызвавшей функции
retn
float MyFunc(float, double) endp
```

Те­перь взгля­нем на резуль­тат твор­чес­тва Embarcadero C++Builder.

```asm
; int __cdecl main(int argc, const char **argv, const char **envp)
public main
main proc near ; DATA XREF: __acrtused+29↑o
var_14 = dword ptr -14h
var_10 = qword ptr -10h
var_8  = dword ptr -8
var_4  = dword ptr -4
        ; C++Builder зачем-то добавил переменные
        push rbp
        sub rsp, 40h
        lea rbp, [rsp+40h]
        ; Готовит аргументы слева направо
        movss xmm0, cs:dword_44D000 ; float
        movsd xmm1, cs:qword_44D010 ; double
        ; Сохраняет значения в стеке
        mov [rbp+var_4], 0
        mov [rbp+var_8], ecx
        mov [rbp+var_10], rdx
        call MyFunc(float,double)
        ; После получения результата преобразует его к одинарной точности
        cvtss2sd xmm0, xmm0
        ; Форматная строка
        lea rcx, unk_44D020
        ; Копирование и подготовка числа к выводу
        movaps xmm1, xmm0
        movq rdx, xmm0
        call printf
        mov [rbp+var_4], 0
        mov [rbp+var_14], eax
        mov eax, [rbp+var_4]
        add rsp, 40h
        pop rbp
        retn
main    endp
; __int64 __fastcall MyFunc(float, double)
        public MyFunc(float, double)
MyFunc(float, double) proc near ; CODE XREF: main+28↓p
var_14 = dword ptr -14h
var_10 = qword ptr -10h
var_4 = dword ptr -4
```

Весь­ма стран­но здесь повела себя IDA, не рас­познав ни одно­го аргу­мен­та, но зато объ­явив три перемен­ные.

```asm
push rbp
sub rsp, 18h
lea rbp, [rsp+10h]
```

Та­ким же обра­зом, как в прош­лый раз, толь­ко в обратном поряд­ке (сле­ва нап­раво), прог­рамма ски­дыва­ет аргу­мен­ты из регис­тров в память.

```asm
movss    [rbp+8+var_4], xmm0
movsd    [rbp+8+var_10], xmm1
cvtss2sd xmm0, [rbp+8+var_4] ; Преобразовывает — с этим мы уже сталкивались
addsd    xmm0, [rbp+8+var_10] ; Складывает
cvtsd2ss xmm0, xmm0 ; Вновь преобразовывает
movss    [rbp+8+var_14], xmm0 ; Лишние телодвижения...
movss    xmm0, [rbp+8+var_14] ; ...
add      rsp, 18h
pop      rbp
retn
MyFunc(float, double) endp
```

Этот лис­тинг мало чем отли­чает­ся от пре­дыду­щего, раз­ве что наг­ромож­дени­ем избы­точ­ного кода, чем зна­менит ком­пилятор от Embarcadero.

Приш­ло вре­мя вспом­нить о при­ложе­нии DataSize, выводя­щем раз­меры типов дан­ных в битах ком­пилято­ра Visual C++ 2019 на плат­форме x64 в опе­раци­онной сис­теме Windows 10. На моем компь­юте­ре вывод выг­лядит так.

| ![Размеры типов данных языка C++](./images/15/4a07f57c0392a60e241a22009ff18d3f.png)
|:--:|
| *Размеры типов данных языка C++* |

Ти­пы язы­ков C/C++ вследс­твие их машино­ориен­тирован­ности сов­пада­ют с типами про­цес­сора. В дан­ном слу­чае нас инте­ресу­ют вещес­твен­ные типы: float (4 бай­та = 32/8), double (8 байт = 64/8).

Ти­пы дан­ных Delphi унас­ледова­ны от Turbo Pascal. В этой сис­теме основной вещес­твен­ный тип дан­ных — Real. Рань­ше он занимал 6 байт, что про­тиво­естес­твен­но для машины, поэто­му он каж­дый раз пре­обра­зовы­вал­ся в тип Extended, который занимал 8 байт (и был подобен типу double из C/C++). Эти пре­обра­зова­ния съеда­ли льви­ную долю про­изво­дитель­нос­ти.

Тем не менее сов­ремен­ный Delphi уже не стра­дает дет­ски­ми болез­нями:

* Real соот­ветс­тву­ет Double (8 байт);
* Real48 вве­ден для обратной сов­мести­мос­ти со ста­рым Real (6 байт);
* Single — Float (4 бай­та);
* Extended на 32-бит­ной Windows — 10 байт;
* Extended на 64-бит­ной Windows — 8 байт;
* Extended на 64-бит­ной Linux — 16 байт;
* Extended на 64-бит­ной macOS (Intel based) — 16 байт.

Вид­но, что раз­работ­чики заморо­чились и «широкий» тип дан­ных стал дей­стви­тель­но широким.

Ра­зоб­равшись с типами дан­ных в Delphi, рас­смот­рим при­мер переда­чи парамет­ра в при­ложе­нии, написан­ном с помощью этой сре­ды раз­работ­ки.

| ![Delphi 10.4 Community Edition](./images/15/d6671637402b5ad23c63b1f24e38ee71.png)
|:--:|
| *Delphi 10.4 Community Edition* |

```pas
program send_real_value_delphi;

{$APPTYPE CONSOLE} // Консольное приложение
{$R *.res}

uses
    System.SysUtils, Math; // Используем математическую библиотеку
procedure MyProc(a : Real);
begin
    WriteLn(FloatToStr(RoundTo(a,-2)));
end;

var
    a: Real;
    b: Real;
begin
    a := 6.66;
    b := 7.77;

    MyProc(a + b);
end.
```

А теперь откомпи­лиру­ем дебаж­ную вер­сию при­ложе­ния под x64 с отклю­чен­ной опти­миза­цией.

| ![Вывод программы](./images/15/4b0d61a689d37dcf29a39c7d6446ab36.png)
|:--:|
| *Вывод программы* |

Изу­чим дизас­сем­блер­ный лис­тинг.

| ![IDA View](./images/15/88a3f368f612d10beb126f118765eebf.png)
|:--:|
| *IDA View* |

IDA сра­зу же откры­вает­ся на фун­кции ини­циали­зации. Пос­ле ини­циали­зации сте­ка в регистр `RAX` помеща­ется флаг «кон­соль­ного при­ложе­ния». Сле­дующей строч­кой он перево­дит­ся в активное положе­ние.

```asm
_ZN22Send_real_value_delphi14initializationEv proc near
push rbp
sub  rsp, 20h
mov  rbp, rsp
mov  rax, cs:off_432480
mov  byte ptr [rax], 1
nop
lea rcx, qword_429460
```

За­тем в регистр `RCX` помеща­ется ука­затель на зна­чение `0x13` — оче­вид­но, для переда­чи сле­дующей фун­кции в качес­тве парамет­ра. Далее эта фун­кция и вызыва­ется в целях ини­циали­зации сис­темно­го модуля при­ложе­ния. Таким обра­зом нас­тра­ивает­ся спо­соб­ность при­нимать сис­темные события и отве­чать на них:

```asm
call  _ZN7Sysinit8_InitExeEPv
; В регистр RAX помещаем первое вещественное значение
mov   rax, 401AA3D70A3D70A4h
; Перемещаем его в локальную переменную a
mov   cs:_ZN22Send_real_value_delphi1aE, rax
; В регистр RAX помещаем второе вещественное значение
mov   rax, 401F147AE147AE14h
; Перемещаем его в локальную переменную b
mov   cs:_ZN22Send_real_value_delphi1bE, rax
; Значение удвоенной точности из переменной a копируем в регистр XMM0
movsd xmm0, cs:_ZN22Send_real_value_delphi1aE
```

К содер­жимому регис­тра `XMM0` при­бав­ляет­ся зна­чение перемен­ной двой­ной точ­ности `b`. Этот регистр будет передан вызыва­емой фун­кции `MyProc` в качес­тве парамет­ра по усло­вию fastcall.

```asm
addsd xmm0, cs:_ZN22Send_real_value_delphi1bE
; Происходит вызов функции MyProc с передачей параметра
call  _ZN22Send_real_value_delphi6MyProcEd
```

Сле­дующая фун­кция завер­шает про­цес­сы, осво­бож­дает память и тем самым завер­шает прог­рамму.

```asm
call _ZN6System6_Halt0Ev
```

Ес­ли при­ложе­ние завер­шилось успешно, выпол­няет­ся безус­ловный переход на мет­ку:

```asm
jmp short loc_429445
; ---------------------------------------------------------------
db  2 dup(90h)
; ---------------------------------------------------------------
```

При сис­темном сбое про­исхо­дит обра­щение к нижес­леду­ющей фун­кции, которая вытал­кива­ет ошиб­ку на вер­хний уро­вень, что­бы показать ее поль­зовате­лю.

```asm
call _ZN6System19_UnhandledExceptionEv
; ---------------------------------------------------------------
db   90h
; ---------------------------------------------------------------
loc_429445: ; Метка, где находится эпилог функции
lea  rsp, [rbp+20h]
pop  rbp
retn
; } // Starts at 4293E0
_ZN22Send_real_value_delphi14initializationEv endp
```

Рас­смот­рим дизас­сем­блер­ный лис­тинг вто­рой нашей фун­кции — `MyProc`. Она обе­щает быть инте­рес­ной: в ней исполь­зует­ся не толь­ко `WriteLn` для вывода тек­ста на кон­соль, но и фун­кция пре­обра­зова­ния вещес­твен­ного чис­ла в стро­ку — `FloatToStr`, а так­же фун­кция округле­ния до ука­зан­ного количес­тва цифр пос­ле десятич­ной запятой — `RoundTo`.

```asm
_ZN22Send_real_value_delphi6MyProcEd proc near
```

IDA вер­но рас­позна­ла один аргу­мент. Кро­ме того, здесь объ­явля­ются две перемен­ные.

```asm
var_s20 = qword ptr 20h
var_s28 = qword ptr 28h
arg_0   = qword ptr 40h
push    rbp
; Инициализируем стек
sub     rsp, 30h
mov     rbp, rsp
; Открываем кадр стека
mov     [rbp+var_s20], 0
; Помещаем аргумент из регистра в ячейку памяти
movsd   [rbp+arg_0], xmm0
nop
movsd   xmm0, [rbp+arg_0]
mov     dl, 0FEh ; 'þ'
```

Вы­зов фун­кции округле­ния вещес­твен­ного чис­ла до ука­зан­ного количес­тва цифр пос­ле десятич­ной запятой выг­лядит так:

```asm
call _ZN6System4Math7RoundToEea
```

Пос­ле выпол­нения фун­кции округле­ния резуль­тат воз­вра­щает­ся в регис­тре `XMM0`. Сле­дующая инс­трук­ция копиру­ет зна­чение в перемен­ную `var_s28`, отку­да оно перено­сит­ся в регистр `XMM1`.

```asm
movsd [rbp+var_s28], xmm0
lea   rcx, [rbp+var_s20] ; Теперь RCX указывает на 0
movsd xmm1, [rbp+var_s28]
; Подготовка форматных настроек
mov   r8, cs:off_432490
; Вызов FloatToStr
call  _ZN6System8Sysutils10FloatToStrEeRKNS0_15TFormatSettingsE
```

Воз­можное отоб­ражение исклю­читель­ной ситу­ации, если все пой­дет не так, как надо:

```asm
mov rcx, cs:off_4322A0
; RDX теперь указывает на 0
mov rdx, [rbp+var_s20]
```

Под­готав­лива­ем и выводим текст в кон­соль:

```asm
call _ZN6System14_Write0UStringERNS_8TTextRecENS_13UnicodeStringE
mov  rcx, rax
call _ZN6System8_WriteLnERNS_8TTextRecE ; WriteLn
call _ZN6System8__IOTestEv
nop
lea  rcx, [rbp+var_s20]
call _ZN6System8_UStrClrEPv
lea  rsp, [rbp+30h]
pop  rbp
retn
_ZN22Send_real_value_delphi6MyProcEd endp
```

Ка­залось бы, с типами дан­ных уже дав­но все понят­но, но даже здесь у Delphi есть скры­тый при­ем, на котором по недос­мотру мож­но зап­росто шею свер­нуть! Будь вни­мате­лен.

## СОГЛАШЕНИЕ О ВЫЗОВАХ ФУНКЦИЙ __THISCALL

В прог­раммах на C++ каж­дая фун­кция объ­екта (она называ­ется методом) неяв­но при­нима­ет аргу­мент `this` — ука­затель на экзем­пляр клас­са (дру­гими сло­вами, объ­екта), из которо­го вызыва­ется метод. Ког­да исполь­зует­ся сог­лашение `__this` на плат­форме x86, любые дру­гие аргу­мен­ты переда­ются через стек, в то же вре­мя ука­затель `this` — через регистр `ECX`.

На плат­форме x64, как мы зна­ем, есть толь­ко одно сог­лашение о вызовах фун­кций — fastcall. Отсю­да сле­дует, что все явные, а так­же неяв­ный аргу­мент `this` переда­ются в регис­трах. `This` всег­да переда­ется в самом пер­вом под­ходящем регис­тре `RCX`.

Рас­смот­рим сле­дующий прос­той при­мер. В нем объ­явлен класс, где опи­саны четыре метода, которые выводят в кон­соль получа­емые парамет­ры.

| ![Вывод приложения thiscall](./images/15/5132c6ce2298c527ea8fc74b0c64f284.png)
|:--:|
| *Вывод приложения thiscall* |

```cpp
#include <iostream>
class MyClass {
public:
    // Прототип demo_1 в действительности выглядит так: demo_1(this, int a);
    void demo_1(int a)
    {
        std::cout << "demo_1"
                  << ": " << a << std::endl;
    }
    // Прототип demo_2 в действительности выглядит так: demo_2(this, int a, int b);
    void demo_2(int a, int b)
    {
        std::cout << "demo_2"
                  << ": " << a << ", " << b << std::endl;
    }
    // Прототип demo_3 в действительности выглядит так: demo_3(this, int a, int b, int c);
    void demo_3(int a, int b, int c)
    {
        std::cout << "demo_3" << ": " << a
                  << ", " << b << ", " << c << std::endl;
    }
    // Прототип demo_4 в действительности выглядит так: demo_4(this, int a, int b, int c,
    // int d);
    void demo_4(int a, int b, int c, int d)
    {
        std::cout << "demo_4" << ": " << a
                  << ", " << b << ", " << c << ", " << d << std::endl;
    }
};

int main()
{
    MyClass* zzz = new MyClass();
    zzz->demo_1(0);
    zzz->demo_2(1, 2);
    zzz->demo_3(3, 4, 5);
    zzz->demo_4(6, 7, 8, 9);
    delete zzz;
    return 0;
}
```

Ре­зуль­тат ком­пиляции это­го при­мера ком­пилято­ром Microsoft Visual C++ 2019 дол­жен выг­лядеть так (показа­на лишь фун­кция `main`, все осталь­ное не пред­став­ляет на дан­ный момент никако­го инте­реса):

```asm
; int __cdecl main(int argc, const char **argv, const char **envp)
main proc near
d = dword ptr -48h
var_38 = qword ptr -38h
var_30 = qword ptr -30h
var_28 = qword ptr -28h
block  = qword ptr -20h
var_18 = qword ptr -18h
sub   rsp, 68h
; Выделяем 1 байт для экземпляра класса
mov   ecx, 1 ; size
; Вызываем конструктор
call  operator new(unsigned __int64)
mov   [rsp+68h+var_30], rax
; Переменная var_30 — указатель на объект
; но если же она равна нулю...
cmp   [rsp+68h+var_30], 0
; Делаем прыжок на метку
jz    short loc_140001257
; Продолжаем нормальное выполнение — в var_30 указатель на объект
mov   rax, [rsp+68h+var_30]
; В var_28 указатель на объект
mov   [rsp+68h+var_28], rax
; Безусловный переход
jmp   short loc_140001260
; ---------------------------------------------------------------
loc_140001257: ; CODE XREF: main+19↑j
mov   [rsp+68h+var_28], 0
loc_140001260: ; CODE XREF: main+25↑j
mov   rax, [rsp+68h+var_28]
; В var_38 указатель на объект
mov   [rsp+68h+var_38], rax
; Начинается подготовка регистров для передачи аргументов: справа налево
; Весьма интересный способ обнуления значения: с помощью инструкции xor
xor   edx, edx ; a
; В RCX помещается указатель на объект
mov   rcx, [rsp+68h+var_38] ; this
; Вызов метода
call  MyClass::demo_1(int)
; После возвращения сразу готовятся новые аргументы: два явных и один неявный
mov   r8d, 2 ; b
mov   edx, 1 ; a
mov   rcx, [rsp+68h+var_38] ; this
call  MyClass::demo_2(int,int)
; Три явных и один неявный — все помещаются в целочисленные регистры
mov   r9d, 5 ; c
mov   r8d, 4 ; b
mov   edx, 3 ; a
mov   rcx, [rsp+68h+var_38] ; this
call  MyClass::demo_3(int,int,int)
; Четыре явных аргумента и один неявный, не хватает регистров!
; Самый правый аргумент — d помещается в стек
mov   [rsp+68h+d], 9 ; d
mov   r9d, 8 ; c
mov   r8d, 7 ; b
mov   edx, 6 ; a
mov   rcx, [rsp+68h+var_38] ; this
call  MyClass::demo_4(int,int,int,int)
mov   rax, [rsp+68h+var_38]
mov   [rsp+68h+block], rax
mov   edx, 1 ; __formal
mov   rcx, [rsp+68h+block] ; block
; Вызов деструктора
call  operator delete(void *,unsigned __int64)
cmp   [rsp+68h+block], 0
jnz   short loc_1400012F5
mov   [rsp+68h+var_18], 0
jmp   short loc_140001308
; ---------------------------------------------------------------
loc_1400012F5: ; CODE XREF: main+B8↑j
mov   [rsp+68h+var_38], 8123h
mov   rax, [rsp+68h+var_38]
mov   [rsp+68h+var_18], rax
loc_140001308: ; CODE XREF: main+C3↑j
xor   eax, eax
add   rsp, 68h
retn
main endp
```

## ПАРАМЕТРЫ ПО УМОЛЧАНИЮ

Что­бы про­ще было вызывать фун­кции с «хорово­дом» парамет­ров, в язык C++ вве­ли воз­можность задать парамет­ры по умол­чанию. Отсю­да воз­ника­ет воп­рос: отли­чает­ся ли чем‑нибудь вызов фун­кций с парамет­рами по умол­чанию от обыч­ных фун­кций? И кто ини­циали­зиру­ет опу­щен­ные парамет­ры — вызыва­емая или вызыва­ющая фун­кция?

Так вот, при вызове фун­кций с парамет­рами по умол­чанию ком­пилятор самос­тоятель­но добав­ляет недос­тающие парамет­ры, а вызов такой фун­кции ничем не отли­чает­ся от вызова обыч­ных фун­кций.

До­кажем это на сле­дующем при­мере:

```cpp
#include <iostream>

void MyFunc(int a = 1, int b = 2, int c = 3)
{
    std::cout << a << ", " << b << ", " << c << std::endl;
}

int main()
{
    MyFunc();
}
```

Ре­зуль­тат его ком­пиляции в Visual C++ 2019 будет выг­лядеть приб­лизитель­но так (для эко­номии мес­та показа­на толь­ко вызыва­ющая фун­кция):

```asm
; int __cdecl main(int argc, const char **argv, const char **envp)
main proc near
sub rsp, 28h
mov r8d, 3 ; c
mov edx, 2 ; b
mov ecx, 1 ; a
```

Как вид­но, все парамет­ры, задан­ные по умол­чанию, были переда­ны фун­кции самим ком­пилято­ром.

```asm
call MyFunc(int,int,int)
    xor eax, eax
    add rsp, 28h
    retn
main endp
```

## ТЕХНИКА ИССЛЕДОВАНИЯ МЕХАНИЗМА ПЕРЕДАЧИ ПАРАМЕТРОВ НЕИЗВЕСТНЫМ КОМПИЛЯТОРОМ

Ком­пилято­ров сей­час огромное мно­гооб­разие, и пос­тоян­но появ­ляют­ся новые, так что при­вес­ти все­охва­тыва­ющую таб­лицу, которая бы рас­писыва­ла харак­тер каж­дого из них, невоз­можно. Как же быть, если тебе попада­ется прог­рамма, откомпи­лиро­ван­ная каким‑то неиз­вес­тным ком­пилято­ром?

Ес­ли ком­пилятор удас­тся опоз­нать (нап­ример, с помощью IDA или по тек­сто­вым стро­кам, содер­жащим­ся в фай­ле), оста­ется толь­ко раз­добыть его экзем­пляр и прог­нать через него серию тес­товых при­меров с переда­чей «подопыт­ной» фун­кции парамет­ров раз­лично­го типа. Нелиш­не изу­чить при­лага­емую к ком­пилято­ру докумен­тацию — воз­можно, там будут хотя бы крат­ко опи­саны все под­держи­ваемые им механиз­мы переда­чи парамет­ров.

Ху­же, ког­да ком­пилятор не опоз­нает­ся или дос­тать его копию нет никакой воз­можнос­ти. Тог­да при­дет­ся кро­пот­ливо и тща­тель­но иссле­довать вза­имо­дей­ствие вызыва­емой и вызыва­ющей фун­кций.
