# Ищем тестовые строки в чужой

Когда изучаешь дизассемблированный листинг программы в целях понять, как она работает, очень часто нужно отыскать в ней строки. В сегодняшней статье мы рассмотрим, какие типы строк существуют, из каких символов они могут состоять, какого размера бывают символы и чем отличается строка от простого набора байтов. Еще мы обсудим хранение, передачу и обработку строк разными компиляторами.

Казалось бы, что может быть сложного в идентификации строк? Если то, на что ссылается указатель, выглядит как строка, это и есть строка! Более того, в подавляющем большинстве случаев строки обнаруживаются и идентифицируются тривиальным просмотром дампа программы (при условии, конечно, что они не зашифрованы, но шифрование — тема отдель\- ного разговора). Так‑то оно так, да не все столь просто!



Задача номер один — автоматизированное выявление строк в программе: не пролистывать же мегабайтовые дампы вручную? Существует множество алгоритмов идентификации строк. Самый простой (но не самый надежный) основан на двух тезисах:



* строка состоит из ограниченного ассортимента символов. В грубом приближении — это цифры, буквы алфавита (включая пробелы), знаки препинания и служебные символы наподобие табуляции или возврата каретки;

* строка должна состоять по крайней мере из нескольких символов.




Условимся считать минимальную длину строки равной N байтам. Тогда для автоматичес\- кого выявления всех строк достаточно отыскать все последовательности из N и более

«строковых» символов. Весь вопрос в том, чему должна быть равна N и какие символы счи\- тать «строковыми».



Если N имеет малое значение, порядка трех‑четырех байтов, то мы получим очень большое количество ложных срабатываний. Напротив, когда N велико, порядка шести‑восьми байтов, число ложных срабатываний близко к нулю и ими можно пренебречь, но все короткие стро\- ки, например OK, YES, NO, окажутся не распознаны! Другая проблема: помимо знако‑циф\- ровых символов, в строках встречаются и элементы псевдографики (особенно часты они в консольных приложениях) и всякие там «мордашки», «стрелки», «карапузики» — словом, почти вся таблица ASCII. Чем же тогда строка отличается от случайной последовательности байтов? Частотный анализ здесь бессилен: ему для нормальной работы требуется как минимум сотня байтов текста, а мы говорим о строках из двух‑трех символов!



Зайдем с другого конца. Если в программе есть строка, значит, на нее кто‑нибудь да ссы\- лается. А раз так, можно поискать среди непосредственных значений указатель на рас\- познанную строку. И если он будет найден, шансы на то, что это действительно именно стро\- ка, а не случайная последовательность байтов, резко возрастают. Все просто, не так ли?



Просто, да не совсем! Рассмотрим следующий пример (writeln_d):



![image](19/Image_003.png)

program writeln_d; begin

Writeln('Hello, Sailor!'); end.



[Результат выполнения](https://xakep.ru/wp-content/uploads/2022/04/383034/HelloSailor.png) [writeln_d](https://xakep.ru/wp-content/uploads/2022/04/383034/HelloSailor.png)



Откомпилируем этот пример. Хотелось бы сказать, любым Pascal-компилятором, только любой нам не подойдет, поскольку нам нужен бинарный код под архитектуру x86-64. Это автоматически сужает круг подходящих компиляторов. Даже популярный Free Pascal все еще не умеет билдить программы для Windows x64. Но не убирай его далеко, он нам еще пригодится.



В таком случае нам придется воспользоваться Embarcadero Delphi 10.4. Настрой компилятор для построения 64-битных приложений и загрузи откомпилированный файл в дизассемблер:



![image](19/Image_004.jpg)

IDA определила, что перед вызовом функции

\_ZN6System14\_Write0UStringERNS\_8TTextRecENS\_13UnicodeStringE в регистр RDX загружа\- ется указатель на смещение aHelloSailor. Посмотрим, куда оно указывает (дважды щел\- кнем по нему):



.text:000000000040E6DC aHelloSailor: ; DATA XREF: _ZN9Writeln_d14initializationEv

.text:000000000040E6DC text "UTF-16LE", 'Hello, Sailor!',0

![image](19/Image_005.png)



Ага! Текст в кодировке UTF-16LE. Что такое UTF-16, думаю, всем понятно. Два конечных сим\- вола обозначают порядок байтов. В данном случае, поскольку приложение скомпилировано для архитектуры x86-64, в которой используется порядок байтов «от младшего к стар\- шему» — little endian, упомянутые символы говорят именно об этом. В противоположном слу\- чае, например на компьютере с процессором SPARC, кодировка имела бы название UTF-16BE от big endian.



Из этого следует, что Delphi кодирует каждый символ переменным количеством байтов: 2 или 4\. Посмотрим, как себя поведет Visual C++ 2019 с аналогичным кодом:

![image](19/Image_006.png)

#include &lt;stdio.h&gt; int main() {

printf("%s", "Hello, Sailor!");

}



Результат дизассемблирования:



![image](19/Image_007.png)

main proc near

sub rsp, 28h

lea rdx, aHelloSailor ; "Hello, Sailor!" lea rcx, _Format ; "%s"

call printf

xor eax, eax

add rsp, 28h retn

main endp



Поинтересуемся, что находится в сегменте данных только для чтения (rdata) по смещению

aHelloSailor:



.rdata:0000000140002240 aHelloSailor db 'Hello, Sailor!',0 ; DATA XREF: main+4

![image](19/Image_008.png)



Никаких дополнительных сведений о размере символов. Из этого можно сделать вывод, что используется стандартная 8-битная кодировка ASCII, в которой под каждый символ отводит\- ся только 1 байт.



Ядро Windows NT изначально использовало для работы со строковыми символами кодиров\- ку UTF-16, однако до Windows 10 в пользовательском режиме применялись две кодировки: UTF-16 и ASCII (нижние 128 символов для английского языка, верхняя половина — для рус\- ского). Начиная с Windows 10, в user mode используется только UTF-16. Между тем символы могут храниться и в ASCII, что мы видели в примере выше.



В C/C++ char является исходным типом символа и позволяет хранить любой символ нижней и верхней частей кодировки ASCII размером 8 бит. Хотя реализация типа wchar_t полностью лежит на совести разработчика компилятора, в Visual C++ он представляет собой полный аналог символа кодировки UTF-16LE, то есть позволяет хранить любой символ Юникода.



Для демонстрации двух основных символьных типов в Visual C++ напишем элементарный пример:

![image](19/Image_009.png)

#include &lt;iostream&gt; int main() {

std::cout << "size of 'char': " << sizeof(char) << "\\n"; std::cout << "size of 'wchar': " << sizeof(wchar_t) << "\\n";



char line1\[\] = "Hello, Sailor!"; wchar_t line2\[\] = L"Hello, Sailor!";



std::cout << "size of 'array of chars': " << sizeof(line1) << "\\n"; std::cout << "size of 'array of wchars': " << sizeof(line2) << "\\n";

}



Результат его выполнения представлен ниже.



![image](19/Image_010.jpg)

[Размеры символьных данных](https://xakep.ru/wp-content/uploads/2022/04/383034/sizes.png)



Думаю, все понятно без подробных пояснений: char — 1 байт, wchar_t — 2 байта. Строка "Hello, Sailor!" состоит из 14 символов, плюс конечный 0\. Это также отражено в выводе прог\- раммы.



INFO



В стандарте C++ есть типы символов: char8_t, char16_t, char32_t. Первый из них был добавлен с введением стан‐ дарта **C++20**, два других добавлены в **C++11**. Их размер отражается в их названиях: char16_t используется

для символов кодировки UTF-16, char32_t — для UTF-32. При этом char8_t не то же самое, что «унаследованный» char, хотя позволяет работать с символами последнего, главным образом он предназначен для литералов кодировки **UTF-8**.



Размеры символов важны для обнаружения границ строк при анализе дизассемблерных листингов программ.



Можно сделать вывод, что современные Visual C++ и Delphi оперируют одинаковыми типами строк, неважно какого размера, но оканчивающиеся символом 0\. Но так было не всегда. В качестве исторического экскурса откомпилируем пример writeln_d компилятором Free Pascal.



![image](19/Image_011.jpg)

[Среда](https://xakep.ru/wp-content/uploads/2022/04/383034/FreePascal.png) [Free Pascal](https://xakep.ru/wp-content/uploads/2022/04/383034/FreePascal.png)



Загрузим результат в IDA.



![image](19/Image_012.jpg)

[IDA](https://xakep.ru/wp-content/uploads/2022/04/383034/i386.png) [определила](https://xakep.ru/wp-content/uploads/2022/04/383034/i386.png)[,](https://xakep.ru/wp-content/uploads/2022/04/383034/i386.png) [что загружаемый исполняемый файл](https://xakep.ru/wp-content/uploads/2022/04/383034/i386.png) [32-](https://xakep.ru/wp-content/uploads/2022/04/383034/i386.png)[разрядный](https://xakep.ru/wp-content/uploads/2022/04/383034/i386.png)



_main proc near

![image](19/Image_013.png)

argc = dword ptr 8 argv = dword ptr 0Ch envp = dword ptr 10h

push ebp

mov ebp, esp

push ebx

call FPC_INITIALIZEUNITS

call fpc\_get\_output mov ebx, eax

mov ecx, offset _$WRITELN\_FP$\_Ld1 mov edx, ebx

mov eax, 0

call FPC\_WRITE\_TEXT_SHORTSTR

call FPC_IOCHECK

mov eax, ebx

call fpc\_writeln\_end call FPC_IOCHECK

call FPC\_DO\_EXIT

_main endp



Так‑так‑так… какой интересный код для нас приготовил Free Pascal! Сразу же бросается в глаза смещение



_$WRITELN\_FP$\_Ld1



адрес которого помещается в регистр ECX перед вызовом процедуры FPC\_WRITE\_TEXT_SHORTSTR, своим названием намекающей на вывод текста. Постой, ведь это же 32-разрядная программа, где передача параметров в регистрах скорее исключение, чем правило, и используется только при соглашении fastcall, в остальных же случаях параметры передаются через стек!



Заглянем‑ка в документацию по компилятору… Есть контакт! По умолчанию в коде механиз\- ма вызова процедур для процессоров i386 используется соглашение register. У нормальных людей оно называется fastcall. И, поскольку для платформы x86 оно не стандартизировано, в отличие от x64, для передачи параметров используются все свободные регистры! Поэтому в том, что используется регистр ECX, нет ничего сверхъестественного.



Чтобы окончательно убедиться в нашей догадке, посмотрим, как распоряжается передан\- ным параметром вызываемая функция FPC\_WRITE\_TEXT_SHORTSTR:



![image](19/Image_014.png)

FPC\_WRITE\_TEXT_SHORTSTR proc near

; CODE XREF: _main+1C↑p

; sub_403320+31↑p ... push ebx

push esi

push edi

mov ebx, eax

mov esi, edx

mov edi, ecx ; Копирование параметра в регистр EDI



Но тут много чего копируется, поэтому эта инструкция не доказательство. Смотрим дальше.



![image](19/Image_015.png)

mov edx, ds:FPC\_THREADVAR\_RELOCATE test edx, edx

jz short loc_40661C

mov eax, ds:U_$SYSTEM_$$_INOUTRES call edx ; FPC\_THREADVAR\_RELOCATE jmp short loc_406621



; \-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\- loc_40661C: ; CODE XREF: FPC\_WRITE\_TEXT_SHORTSTR+11↑j

mov eax, offset unk_40B154

loc_406621: ; CODE XREF: FPC\_WRITE\_TEXT_SHORTSTR+1A↑j

cmp word ptr \[eax\], 0 jnz loc_4066AC

mov eax, \[esi+4\]

cmp eax, 0D7B1h

jl short loc_40668C

sub eax, 0D7B1h

jz short loc_40666C

sub eax, 1

jnz short loc_40668C

mov esi, esi



Ага! Следующая инструкция копирует указатель, преобразуя его в 32-разрядное значение без учета знака (указатель не может быть отрицательным). Затем с помощью команды cmp сравниваются значения двух регистров: EAX и EBX. И если EAX больше или равен EBX, выпол\- няется переход на метку loc_40665C...



![image](19/Image_016.png)

movzx eax, byte ptr \[edi\] cmp eax, ebx

jge short loc_40665C movzx eax, byte ptr \[edi\] mov edx, ebx

sub edx, eax

mov eax, esi

call sub_4064F0

lea esi, \[esi+0\]

loc_40665C: ; CODE XREF: FPC\_WRITE\_TEXT_SHORTSTR+49↑j



...где происходит похожая на манипуляцию со строкой деятельность.



![image](19/Image_017.png)

movzx ecx, byte ptr \[edi\] lea edx, \[edi+1\]

mov eax, esi

call sub_406460

...



Теперь мы смогли убедиться в правильности нашего предположения! Вернемся к основному исследованию и посмотрим, что же скрывается под подозрительным смещением:



.rdata:00409004 _$WRITELN\_FP$\_Ld1 db 0Eh ; DATA XREF: _main+10↑o

.rdata:00409005 db 48h ; H

![image](19/Image_018.png)

.rdata:00409006 db 65h ; e

.rdata:00409007 db 6Ch ; l

.rdata:00409008 db 6Ch ; l

.rdata:00409009 db 6Fh ; o

.rdata:0040900A db 2Ch ; ,

.rdata:0040900B db 20h

.rdata:0040900C db 53h ; S

.rdata:0040900D db 61h ; a

.rdata:0040900E db 69h ; i

.rdata:0040900F db 6Ch ; l

.rdata:00409010 db 6Fh ; o

.rdata:00409011 db 72h ; r

.rdata:00409012 db 21h ; !

1.  data:00409013 db 0



    Согласись, не это мы ожидали увидеть. Однако последовательное расположение символов строки «в столбик» дела не меняет. Интересен другой момент: в начале строки стоит число, показывающее количество символов в строке, — 0xE (14 в десятичной системе).



    Оказывается, мало идентифицировать строку, требуется еще как минимум определить ее границы.



    ТИПЫ СТРОК

    Наиболее популярны следующие типы строк: С‑строки, которые завершаются нулем; DOS- строки, завершаются символом $ (такие строки используются не только в MS-DOS); Pascal- строки, которые предваряет одно-, двух\- или четырехбайтовое поле, содержащее длину стро\- ки. Рассмотрим каждый из этих типов подробнее.



    С-строки

    **С**‑**строки**, также именуемые ASCIIZ-строками (от Zero — ноль на конце) или нуль‑терминиро\- ванными, — весьма распространенный тип строк, широко использующийся в операционных системах семейств Windows и UNIX. Символ \\0 (не путать с 0) имеет специальное предназна\- чение и трактуется по‑особому, как признак завершения строки. Длина ASCIIZ-строк прак\- тически ничем не ограничена, ну разве что размером адресного пространства, выделенного процессу. Поэтому теоретически в Windows NT х64 максимальный размер ASCIIZ-строки лишь немногим менее 16 Тбайт.



    Фактическая длина ASCIIZ-строк лишь на байт длиннее исходной ASCII-строки. Несмотря на перечисленные выше достоинства, С‑строкам присущи и некоторые недостатки. Во‑пер\- вых, ASCIIZ-строка не может содержать нулевых байтов, поэтому она непригодна для обра\- ботки бинарных данных. Во‑вторых, операции копирования, сравнения и конкатенации С‑строк сопряжены со значительными накладными расходами — современным процес\- сорам невыгодно работать с отдельными байтами, им желательно иметь дело с четверными словами.



    Но, увы, длина ASCIIZ-строк наперед неизвестна, и ее приходится вычислять на лету, про\- веряя каждый байт на символ завершения. Правда, разработчики некоторых компиляторов идут на хитрость: они завершают строку семью нулями, что позволяет работать с двойными словами, а это на порядок быстрее. Почему семью, а не четырьмя, ведь в двойном слове байтов четыре? Да, верно, четыре, но подумай, что произойдет, если последний значимый символ строки придется на первый байт двойного слова? Верно, его конец заполнят три нулевых байта, но двойное слово из‑за вмешательства первого символа уже не будет равно нулю! Вот поэтому следующему двойному слову надо предоставить еще четыре нулевых байта, тогда оно гарантированно будет равно нулю. Впрочем, семь служебных байтов на каж\- дую строку — это уже перебор!



    DOS-строки

    В MS-DOS (и не только в ней) функция вывода строки воспринимает знак $ как символ завершения, поэтому в программистских кулуарах такие строки называют DOS-строками. Термин не совсем корректен: все остальные функции MS-DOS работают исключительно с ASCIIZ-строками!



    Причина выбора столь странного символа‑разделителя восходит к тем древнейшим вре\- менам, когда никакого графического интерфейса еще и в помине не существовало, а кон\- сольный терминал считался весьма продвинутой системой взаимодействия с пользовате\- лем. Клавиша Enter не могла служить завершителем строки, так как подчас приходилось вводить в программу несколько строк сразу.



    Комбинации Ctrl-Z или Alt-000 также не годились — на многих клавиатурах тех лет отсутство\- вали такие регистры! С другой стороны, компьютеры использовали главным образом для инженерных, а не бухгалтерских расчетов и символ «бакса» употребляли реже всего, вот и решили использовать его для сигнализации о завершении пользователем ввода и как сим\- вол — завершитель строки. Да, символ‑завершитель вводил пользователь, а не добавляла программа, как это происходит с ASCIIZ-строками. В настоящее время DOS-строки прак\- тически вышли из употребления.



    Pascal-строки

    Pascal-строки не имеют завершающего символа, вместо этого они предваряются специаль\- ным полем, содержащим длину этой строки. Достоинства этого подхода: возможность хра\- нения любых символов в строке (в том числе и нулевых байтов!) и высокая скорость обра\- ботки строковых переменных. Вместо постоянной проверки каждого байта на завершающий символ происходит лишь одно обращение к памяти — загрузка длины строки.



    Ну а раз длина строки известна, можно работать не с байтами, а с четверными словами —

    «родным» типом данных 64-разрядных процессоров. Весь вопрос в том, сколько байтов отвести под поле размера. Один? Что ж, это экономно, но тогда максимальная длина строки будет ограничена 255 символами, что во многих случаях оказывается явно недостаточно! Этот тип строк используют практически все старые Pascal-компиляторы (например, Borland Turbo Pascal, Free Pascal), поэтому‑то такие строки и называют Pascal-строками или, если более точно, короткими Pascal-строками.



    Delphi-строки

    Осознавая очевидную смехотворность ограничения длины Pascal-строк 255 символами, раз\- работчики Delphi расширили поле размера до двух байтов, увеличив тем самым максималь\- но возможную длину до 65 535 символов. Хотя такой тип строк поддерживают и другие ком\- пиляторы (тот же Free Pascal, к примеру), в силу сложившейся традиции их принято име\- новать Delphi-строками или Pascal-строками с двухбайтовым полем размера — двухбай\- товыми Pascal-строками.



    Ограничение в шестьдесят с гаком килобайтов и ограничением язык назвать не поворачива\- ется. Большинство строк имеют гораздо меньшую длину, а для обработки больших мас\- сивов данных (текстовых файлов, к примеру) есть куча (динамическая память) и ряд спе\- циализированных функций. Накладные же расходы (два служебных байта на каждую стро\- ковую переменную) не столь велики, чтобы их брать в расчет.



    Словом, Delphi-строки, которые сочетают в себе лучшие стороны С\- и Pascal-строк (прак\- тически неограниченную длину и высокую скорость обработки соответственно), представ\- ляются самым удобным и практичным типом.



    Wide-Pascal-строки

    «Широкие» Pascal-строки отводят на поле размера аж четыре байта, «ограничивая» мак\- симально возможную длину 4 294 967 295 символами или 4 Гбайт. Однако за эту роскошь приходится дорого платить, отдавая каждой строке четыре лишние байта, три из которых в большинстве случаев будут попросту пустовать. Накладные расходы на коротких строках становятся весьма велики, поэтому тип Wide-Pascal практически не используется.



    Комбинированные типы



    Некоторые компиляторы используют комбинированный C+Pascal-тип, что позволяет им, с одной стороны, достичь высокой скорости обработки строк и хранить в строках любые символы, а с другой — обеспечить совместимость с огромным количеством C-библиотек,

    «заточенных» под ASCIIZ-строки. Каждая комбинированная строка принудительно завер\- шается нулем, но этот ноль в саму строку не входит, и штатные библиотеки (операторы) язы\- ка работают с ней как с Pascal-строкой. При вызове же функций С‑библиотек компилятор передает им указатель не на истинное начало строки, а на первый символ строки.



    STL-строки

    Строки стандартной библиотеки языка C++ могут иметь произвольный размер, ограничен\- ный только размером виртуальной памяти, так как располагаются в куче. При этом в стеке создается только указатель специального типа размером в 40 байт независим от количес\- тва символов в строке. При работе с этими строками программист освобождается от всех забот, связанных с обработкой символов и слежением за памятью. Такие строки умеют пре\- образовываться в любой распространенный формат строк. По удобству использования это аналог Delphi-строк.



    ![image](19/Image_019.jpg)



    [Основные типы строк](https://xakep.ru/wp-content/uploads/2022/04/383034/strings.png)



    Не менее типа строк важен их размер. Как видно на картинке выше, размер есть сумма всех составляющих строку байтов. Также на картинке видно, что символ «вписывается» в байт, то есть размер любого символа равен байту.



    Но не все так однозначно. Восемь бит на символ принято отводить в кодировке ASCII. В наши дни пальма первенства принадлежит Юникоду. В соответствии с этим стандартом определено несколько способов преобразования кодов символов для хранения и передачи. Они получили названия UTF-8, UTF-16, UTF-32. Первый широко используется в Unix-подобных системах, в том числе macOS. В ней символы Юникода передаются и хранятся с исполь\- зованием переменного количества байтов: от 1 до 6 включительно.



    UTF-16 тоже использует переменное количество байтов, отличие состоит в том, что один символ кодируется либо 2, либо 4 байтами. Эта кодировка родная для Windows NT и всех последующих версий винды. Для кодирования Юникод‑символов в кодировке UTF- 32 для каждого символа неизменно используется 4 байта. Это влечет расточительный рас\- ход пространства. Зато работа с такими символами выполняется быстрее, так как они про\- индексированы.



    Определение типа строк



    По внешнему виду строки определить ее тип весьма затруднительно. Наличие завершающе\- го нуля в конце строки еще не повод считать ее ASCIIZ-строкой (Pascal-компиляторы в конец строк частенько дописывают один или несколько нулей для выравнивания данных по крат\- ным адресам), а совпадение предшествующего строке байта с ее длиной может действи\- тельно быть лишь случайным совпадением.



    Грубо говоря, тип строки определяется по роду компилятора (C или Pascal), а точно — по алгоритму обработки этой строки (то есть анализом манипулирующего с ней кода). Рас\- смотрим следующий пример, демонстрирующий идентификацию типа строк:



    #include &lt;stdio.h&gt;

    ![image](19/Image_020.png)

    #include &lt;string.h&gt;



    int main() {

    char s0\[\] = "Hello, World!"; char s1\[\] = "Hello, Sailor!"; if (strcmp(&s0\[0\], &s1\[0\]))

    printf("Woozl\\n"); else

    printf("OK\\n");



    return 0;

    }





    [Результат выполнения](https://xakep.ru/wp-content/uploads/2022/04/383034/stringTypes_vc.png) [stringTypes_vc](https://xakep.ru/wp-content/uploads/2022/04/383034/stringTypes_vc.png)



    Откомпилируем его в среде Visual Studio, загрузим результат в IDA Pro:



    ![image](19/Image_021.png)

    ; int cdecl main(int argc, const char \*\*argv, const char \*\*envp) main proc near ; CODE XREF: scrt\_common\_main_seh+107↓p

    ; DATA XREF: .pdata:0000000140004018↓o

    ; Три локальные переменные var_48 = byte ptr -48h var_38 = byte ptr -38h var_28 = qword ptr -28h

    ; unwind { // GSHandlerCheck

    ![image](19/Image_022.png)

    ; Сохраняем регистры push rsi

    push rdi

    ; Выделяем память для локальных переменных sub rsp, 58h

    mov rax, cs: security_cookie xor rax, rsp

    mov \[rsp+68h+var_28\], rax

    ; Помещаем в RAX указатель на область памяти (переменную) lea rax, \[rsp+68h+var_48\]

    ; В RCX помещаем указатель на строку

    lea rcx, aHelloWorld ; "Hello, World!"

    ; В RDI помещается адрес приемника (переменная var_48) mov rdi, rax

    ; В RSI помещается адрес строки-источника aHelloWorld mov rsi, rcx

    ; В ECX помещается количество повторений операции movsb mov ecx, 0Eh

    ; Запускаем цикл пересылки байтов из источника в приемник заданное количество раз rep movsb

    ; После каждой итерации значения в RSI и RDI увеличиваются на 1



    ; Помещаем в RAX указатель на область памяти (переменную) lea rax, \[rsp+68h+var_38\]

    ; В RCX помещаем указатель на строку

    lea rcx, aHelloSailor ; "Hello, Sailor!"

    ; В RDI помещается адрес приемника (переменная var_38) mov rdi, rax

    ; В RSI помещается адрес строки-источника aHelloSailor mov rsi, rcx

    ; В ECX помещается количество повторений операции movsb mov ecx, 0Fh

    ; Запускаем цикл пересылки байтов из источника в приемник заданное количество раз rep movsb

    ; После каждой итерации значения в RSI и RDI увеличиваются на 1



    Уже сейчас мы можем сделать вывод о том, с каким типом строк мы имеем дело. Очевидно, это однобайтовые C-строки.



    ![image](19/Image_023.png)

    ; Установка нулевого смещения... mov eax, 1

    imul rax, 0

    ; ... в строке aHelloSailor

    lea rax, \[rsp+rax+68h+var_38\]

    ; Установка нулевого смещения... mov ecx, 1

    imul rcx, 0

    ; ... в строке aHelloWorld

    lea rcx, \[rsp+rcx+68h+var_48\]

    ; Получаем разность смещений двух строк sub rax, rcx

    ; Блок кода сравнения строк loc_140001147: ; CODE XREF: main+74↓j

    ; Копируем байт (один символ) из строки aHelloWorld, на который указывает адрес в

    ; в регистр EDX, расширяя его до 32 разрядов movzx edx, byte ptr \[rcx\]



    Сравниваем единственный символ в регистре DL (EDX), помещенный туда на предыдущем шаге, с символом из строки aHelloSailor, на который указывает адрес RCX + RAX.



    cmp dl, \[rcx+rax\] ; Если первый операнд равен второму, то ZF = 1, а CF = 0

    ![image](19/Image_024.png)

    ![image](19/Image_025.png)

    ; Если же символы не одинаковые, то ZF = 0 jnz short loc_14000115A ; Если ZF == 1, прыжок не осуществляем

    inc rcx ; Увеличиваем RCX на 1 (переходим к следующему симво

    test dl, dl ; Если DL != 0, тогда ZF = 0



    Если ZF равен 0, делаем переход на метку выше, другими словами, зацикливаемся.



    ![image](19/Image_026.png)

    jnz short loc_140001147 ; Прыгаем обратно xor eax, eax ; Обнуление EAX

    jmp short loc_14000115F ; Безусловный прыжок



    ; ---------------------------------------------------------------------------

    loc_14000115A: ; CODE XREF: main+6D↑j



    Сюда осуществим переход, когда в строках попадутся разные символы. В результате выпол\- нения двух следующих инструкций регистр EAX будет иметь значение 1.



    ![image](19/Image_027.png)

    sbb eax, eax

    or eax, 1



    loc_14000115F: ; CODE XREF: main+78↑j

    ; Сюда происходит безусловный переход

    test eax, eax ; Если EAX == 1, то ZF = 0

    jz short loc_140001171 ; Если ZF == 0, то переход не делаем

    lea rcx, _Format ; "Woozl\\n" — загружаем указатель на строку в регис call printf ; Выводим на экран оповещение, что строки разные

    jmp short loc_14000117D ; Безусловный переход

    ; ---------------------------------------------------------------------------

    loc_140001171: ; CODE XREF: main+81↑j

    lea rcx, aOk ; "OK\\n" — загружаем указатель на строку в регистр call printf ; Выводим надпись на экран, оповещаем, что строки о

    loc_14000117D: ; CODE XREF: main+8F↑j

    xor eax, eax ; Обнуляем регистр mov rcx, \[rsp+68h+var_28\]

    xor rcx, rsp ; StackCookie call security\_check\_cookie

    ; Восстанавливаем стек

    add rsp, 58h

    ![image](19/Image_028.png)

    ; Восстанавливаем регистры pop rdi

    pop rsi retn

    ; } // starts at 1400010E0 main endp

    Наше предположение о типе строк не встретило противоречий. Даже наоборот, в коде срав\- нения строк мы нашли дополнительное подтверждение: в нем также происходит обработка символов размером в один байт, а первый символ строк ничем не отличается от остальных.



    Если анализ предыдущей программы на Visual C++ показался тебе простым, то предлагаю перейти к анализу программы, откомпилированной в Delphi. Вот увидишь, ее дизассемблер\- ный листинг покажет много нового.



    program stringTypes;



    ![image](19/Image_029.png)

    VAR



    s0, s1 : String;



    BEGIN

    s0 := 'Hello, Sailor!'; s1 := 'Hello, World!';

    IF s0 = s1 THEN Writeln('OK') ELSE Writeln('Woozl');

    END.



    [Результат выполнения](https://xakep.ru/wp-content/uploads/2022/04/383034/stringTypes-Run.png) [stringTypes](https://xakep.ru/wp-content/uploads/2022/04/383034/stringTypes-Run.png)



    Загрузим дебажный результат в IDA Pro. Дизассемблерный листинг откроется на главной функции программы _ZN11Stringtypes14initializationEv:



    public _ZN11Stringtypes14initializationEv



    ![image](19/Image_030.png)

    _ZN11Stringtypes14initializationEv proc near

    ; DATA XREF: HEADER:0000000000400128↑o

    ; .pdata:000000000042105C↓o

    ; unwind { // \_ZN6System23\_DelphiExceptionHandlerEPNS\_16TExceptionRecordEyPvS2\_ push rbp

    sub rsp, 20h mov rbp, rsp

    mov rax, cs:off_412038 mov byte ptr \[rax\], 1 nop

    lea rcx, qword_40E8A0

    call \_ZN7Sysinit8\_InitExeEPv nop

    ; В RCX загружаем указатель на строку-приемник lea rcx, _ZN11Stringtypes2s0E

    ; В RDX загружаем указатель на строку-источник lea rdx, aHelloSailor ; "Hello, Sailor!"

    ; Вызываем функцию, которой в параметрах передаются две строки, пока непонятно дл call \_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_

    ; В RCX загружаем указатель на строку-приемник lea rcx, _ZN11Stringtypes2s1E

    ; В RDX загружаем указатель на строку-источник lea rdx, aHelloWorld ; "Hello, World!"

    ; Вызываем функцию, которой в параметрах передаются две строки, пока непонятно дл call \_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_

    mov rcx, cs:_ZN11Stringtypes2s0E mov rdx, cs:_ZN11Stringtypes2s1E

    ; После загрузки двух результирующих строк в регистры RCX и RDX

    ; они передаются следующей функции, тоже неясно для чего call \_ZN6System10\_UStrEqualENS\_13UnicodeStringES0\_

    ; В регистре EAX функция вернула какой-то результат

    ; Если этот результат не равен 0, тогда флаг ZF устанавливается в 0 test eax, eax

    ; Если флаг ZF равен 0, тогда осуществляем переход по метке jnz short loc_40E7BF

    ; Продолжаем выполнение, если ZF == 1 mov rcx, cs:off_412030

    lea rdx, aOk ; "OK"

    ; Два параметра передаются следующей функции

    call \_ZN6System14\_Write0UStringERNS\_8TTextRecENS\_13UnicodeStringE mov rcx, rax

    call \_ZN6System8\_WriteLnERNS_8TTextRecE call _ZN6System8 IOTestEv

    jmp short loc_40E7DF ; Безусловный переход

    ; ---------------------------------------------------------------------------

    loc_40E7BF: ; CODE XREF: _ZN11Stringtypes14initializationEv+5B↑j

    ; Сюда происходит прыжок в случае, когда ZF == 0 mov rcx, cs:off_412030

    lea rdx, aWoozl ; "Woozl"

    ; Два параметра передаются уже знакомой функции, можно сделать вывод,

    ; что она выводит второй параметр

    call \_ZN6System14\_Write0UStringERNS\_8TTextRecENS\_13UnicodeStringE mov rcx, rax

    call \_ZN6System8\_WriteLnERNS_8TTextRecE call _ZN6System8 IOTestEv

    loc_40E7DF: ; CODE XREF: _ZN11Stringtypes14initializationEv+7D↑j

    ; Метка безусловного перехода nop

    call \_ZN6System6\_Halt0Ev

    ; ---------------------------------------------------------------------------

    jmp short loc_40E7EF

    ; ---------------------------------------------------------------------------

    db 2 dup(90h)

    ; ---------------------------------------------------------------------------

    call \_ZN6System19\_UnhandledExceptionEv

    ; ---------------------------------------------------------------------------

    db 90h

    ; ---------------------------------------------------------------------------

    loc_40E7EF: ; CODE XREF: _ZN11Stringtypes14initializationEv+A5↑j lea rsp, \[rbp+20h\]

    pop rbp retn

    ; } // starts at 40E740

    _ZN11Stringtypes14initializationEv endp



    Так много вопросов и так мало ответов! Присутствие отладочной информации нисколько не проясняет выполняемые функциями действия, поэтому можно было ее не загружать вов\- се. Поскольку мы сами написали код дизассемблируемой программы, мы, конечно, догады\- ваемся, что происходит на самом деле. Но такого во время проведения реального анализа не будет, поэтому засучим рукава и приступим к выяснению алгоритмов работы этих фун\- кций.



    ![image](19/Image_031.png)

    public \_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_



    \_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_ proc near

    ; CODE XREF: _ZN6System7TObject9UnitScopeEv+82↑p

    ; \_ZN6System19UTF8ToUnicodeStringERKNS\_11SmallStringILh255EEE+8F↓p ...

    ; Одна локальная переменная var_s28 = qword ptr 28h

    ; Открываем кадр стека push rbp

    ;EBP-8h=>D

    ; Сохраняем регистры push r13

    push rdi

    push rsi

    push rbx

    ; Выделяем 0х30 байт для локальных переменных sub rsp, 30h

    ; Копируем указатель на вершину стека mov rbp, rsp

    ; Вспомним, в RCX мы передали указатель на строку-приемник mov rbx, rcx

    ; а в RDX — указатель на строку-источник mov r13, rdx

    ; Указатель на строку-источник копируется в регистр RSI mov rsi, r13

    ; Если RSI == 1 (в нашем случае так и есть, потому что регистру присвоено значени test rsi, rsi

    ; Если ZF == 0, тогда переход не выполняется jz short loc_409954

    ; Проверяем элемент с индексом &lt;...\&gt; на равенство нулю

    ; Обрати внимание: размер значения — двойное слово cmp dword ptr \[r13-8\], 0

    ; Переход по метке выполняется, когда первый операнд больше второго или равен ему jge short loc_409946

    ; Получаем 32-разрядный элемент... mov edi, \[r13-4\]

    ; ...копируем его в ECX... mov ecx, edi

    ; ...передаем его следующей функции

    call \_ZN6System17\_NewUnicodeStringEi; Создаем и получаем Unicode-стр

    ; Копируем результат в регистр RSI mov rsi, rax

    ; Комплектуем три параметра для передачи функции _ZN6System4MoveEPKvPvx

    ; В R13 находится указатель на строку-источник, копируем его в RCX mov rcx, r13

    ; Копируем в RDX указатель на Unicode-строку mov rdx, rsi

    mov r8d, edi

    add r8d, r8d movsxd r8, r8d



    Анализ функции _ZN6System4MoveEPKvPvx показал, что она копирует символы стро\- ки‑источника в новую строку, возвращенную функцией \_ZN6System17\_NewUnicodeStringEi.



    ![image](19/Image_032.png)

    call _ZN6System4MoveEPKvPvx

    ; Указатель на новую Unicode-строку по-прежнему в RSI

    jmp short loc_409954 ; Безусловный переход на метку

    ; ---------------------------------------------------------------------------

    loc_409946: ; CODE XREF: \_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_+20↑j lea rax, \[rsi-10h\]

    mov ecx, 1



    Инструкция xadd выполняет три операции:



    1.  Меняет местами операнд‑приемник и операнд‑источник.

    2.  Вычисляет сумму операндов.

    3.  Помещает сумму на место операнда‑приемника (первый операнд);




Поэтому, чтобы между этими операциями ОС не попыталась выполнить над данными опе\- рандами какие‑либо действия, используется префикс LOCK. Он сообщает операционной сис\- теме о том, что три этих операции необходимо выполнить атомарно.



lock xadd \[rax+8\], ecx



![image](19/Image_033.png)

loc_409954: ; CODE XREF: \_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_+19↑j

; \_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_+44↑j

; Сюда был осуществлен безусловный переход

; В RAX копируется указатель на строку-приемник mov rax, \[rbx\]

; Указатель на строку-приемник копируется в локальную переменную mov \[rbp+var_s28\], rax

; В RBX копируется указатель на Unicode-строку mov \[rbx\], rsi

; Указатель на Unicode-строку копируется в RCX для передачи следующей функции lea rcx, \[rbp+var_s28\]



Функции \_ZN6System8\_UStrClrEPv в качестве параметра передается указатель на локаль\- ную переменную. Память из‑под этой переменной очищается путем последующего вызова процедуры \_ZN6System8\_FreeMemEPv.



call \_ZN6System8\_UStrClrEPv

; Очистка стека

![image](19/Image_034.png)

lea rsp, \[rbp+30h\]

; Восстановление регистров pop rbx

pop rsi

pop rdi

pop r13

; Закрываем кадр стека pop rbp

retn

\_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_ endp



Выходит, функция \_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_, принимая два парамет\- ра — указателя на области памяти: источника и приемника, создает на основе источника Unicode-строку, а возврат новой строки осуществляется в области памяти для строки‑при\- емника.



Теперь мы выяснили предназначение центрального кода процедуры

_ZN11Stringtypes14initializationEv:



![image](19/Image_035.png)

; В RCX загружаем указатель на строку-приемник lea rcx, _ZN11Stringtypes2s0E

; В RDX загружаем указатель на строку-источник lea rdx, aHelloSailor ; "Hello, Sailor!"

; Вызываем функцию, которой в параметрах передаются две строки,

; чтобы в параметре на строку-приемник получить результирующую Unicode-строку call \_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_

; В RCX загружаем указатель на строку-приемник lea rcx, _ZN11Stringtypes2s1E

; В RDX загружаем указатель на строку-источник lea rdx, aHelloWorld ; "Hello, World!"

; Вызываем функцию, которой в параметрах передаются две строки,

; чтобы в параметре на строку-приемник получить результирующую Unicode-строку call \_ZN6System8\_UStrAsgERNS\_13UnicodeStringES0\_

mov rcx, cs:_ZN11Stringtypes2s0E mov rdx, cs:_ZN11Stringtypes2s1E

; После загрузки двух результирующих строк в регистры RCX и RDX

; они передаются следующей функции, пока неясно для чего call \_ZN6System10\_UStrEqualENS_13UnicodeStringES0



Для восстановления алгоритма программы осталось проанализировать дизассемблерный код еще одной (последней) функции.



При этом вывод о типах строк мы можем сделать уже сейчас. Несмотря на то что эта прог\- рамма — результат работы компилятора Delphi, строки хранятся и обрабатываются подобно C-строкам. А если нечто выглядит как утка, плавает как утка и крякает как утка, то это, веро\- ятно, и есть утка. Впрочем, тут нет ничего удивительного, и выше мы уже говорили об этом: современные Visual C++ и Delphi оперируют одинаковыми типами строк, неважно какого раз\- мера, но оканчивающимися символом 0.



Изучим функцию \_ZN6System10\_UStrEqualENS\_13UnicodeStringES0\_, не забывая, что она получает два параметра — две строки (в нашем конкретном случае разные). Пройдя по перекрестной ссылке (для этого надо дважды нажать мышью на имени функции), мы попадаем в функцию‑посредник:



![image](19/Image_036.png)

\_ZN6System10\_UStrEqualENS\_13UnicodeStringES0\_ proc near

; CODE XREF: _ZN11Stringtypes14initializationEv+54↓p

; DATA XREF: .pdata:0000000000420B58↓o

sub rsp, 28h

call \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_ add rsp, 28h

retn

\_ZN6System10\_UStrEqualENS\_13UnicodeStringES0\_ endp



Где, очевидно, нас интересует вызываемая \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_:



\_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_ proc near

; CODE XREF: \_ZN6System10\_UStrEqualENS\_13UnicodeStringES0\_+4↓p

; DATA XREF: .pdata:0000000000420B4C↓o

![image](19/Image_037.png)

; Сохранение регистров в стеке push r14

push r13 push rdi push rsi push rbx

; Выделение памяти под локальные переменные sub rsp, 20h

mov rbx, rdx ; Второй параметр — в RBX mov rax, rcx ; Первый параметр — в RAX

mov rdx, rbx ; Второй параметр — обратно в RDX

; Сравнение параметров, в нашем случае они не равны, поэтому ZF=0 cmp rax, rdx

jnz short loc_409F5D ; Переход к метке, если ZF==0 xor eax, eax ; Обнуление EAX

jmp short loc_409FD1 ; Безусловный переход

; ---------------------------------------------------------------------------

loc_409F5D: ; CODE XREF: \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_+17↑j test rax, rax ; В нашем случае RAX != 0, поэтому ZF=0

jnz short loc_409F69 ; ZF == 0, делаем переход mov eax, \[rdx-4\] ; Копирование значения

neg eax ; Изменение знака значения jmp short loc_409FD1 ; Безусловный переход

; ---------------------------------------------------------------------------

loc_409F69: ; CODE XREF: \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_+20↑j test rdx, rdx ; В нашем случае RDX != 0, поэтому ZF=0

jnz short loc_409F73 ; ZF == 0, делаем переход

mov eax, \[rax-4\] ; Копирование значения по адресу ... в EAX jmp short loc_409FD1 ; Безусловный переход

; ---------------------------------------------------------------------------

loc_409F73: ; CODE XREF: \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_+2C↑j mov edi, \[rax-4\] ; Копируем значение по адресу ... в EDI

mov r13d, \[rdx-4\] ; Копируем значение по адресу ... в R13D mov esi, edi ; из EDI в ESI

cmp esi, r13d ; Сравнение значений из ESI и R13D...

; Если первый операнд меньше второго либо равен ему, сделать переход jle short loc_409F84

; Если при сравнении оказалось, что первый операнд больше \- заменить его вторым mov esi, r13d

loc_409F84: ; CODE XREF: \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_+3F↑j

; В RCX по-прежнему находится первый параметр, он передается следующей функции call \_ZN6System13\_UStrToPWCharENS_13UnicodeStringE



Приведенная выше функция в качестве параметра принимает Юникод‑строку в Delphi-фор\- мате, преобразует ее в строку PWChar и возвращает обратно. Одно из различий Delphi-строк и PWChar-строк состоит в том, что последние оканчиваются символом 0\. То есть приобре\- тают свойства C-строк. Что нам и нужно для сравнения строк, которое показано чуть ниже в соответствующем блоке кода.



Еще одно отличие этих типов строк заключается в том, что, по сути, PWChar представляет не строку, а только указатель на первый символ, подобно C-строкам. В отличие от Delphi- строк, полностью описывающих себя (подробности ищи выше в разделе «Типы строк»).



![image](19/Image_038.png)

; Результат функции помещается в регистр R14 mov r14, rax

; Передача второго параметра материнской функции, поскольку

; он до сих пор находится в регистре RBX, следующей функции mov rcx, rbx

; Действия, которые проводит эта функция, описаны выше call \_ZN6System13\_UStrToPWCharENS_13UnicodeStringE

; Результат выполнения функции копируется в RCX mov rcx, rax

test esi, esi ; Если ESI != 0, то ZF=0

jle short loc_409FCB ; Переход происходит, когда ZF==1



loc_409F9B: ; CODE XREF: \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_+89↓j

; Блок кода сравнения строк

movzx eax, word ptr \[r14\] ; копируется и расширяется до двойного слова

; Похоже, тут символ

movzx edx, word ptr \[rcx\] ; копируется и расширяется до двойного слова

; Похоже на символ

sub eax, edx ; Вычитание...

test eax, eax ; ...Проверка на 0: если EAX == 0, тогда ZF=1

; Если ZF==1, переход не делаем, в противном случае заканчиваем

; сравнение, идем на выход, строки не равны, результат возвращаем в EAX jnz short loc_409FD1

cmp esi, 1 ; Если ESI == 1, тогда ZF = 1 jz short loc_409FCB ; Переход, если ZF==1

movzx eax, word ptr \[r14+2\] ; копируется и расширяется до двойного слова movzx edx, word ptr \[rcx+2\] ; копируется и расширяется до двойного слова sub eax, edx ; Вычитание

test eax, eax ; Если EAX == 0, тогда ZF=1

; Если ZF==1, переход не делаем, в противном случае заканчиваем

; сравнение, идем на выход, строки не равны, результат возвращаем в EAX jnz short loc_409FD1

add r14, 4 ; Похоже на инкрементное увеличение в теле цикла... add rcx, 4 ; Получается, размер символа — 4 байта!

sub esi, 2 ; Уменьшаем значение в регистре ESI test esi, esi ; Если ESI > 0, тогда ZF = 0

; Если ZF == 0, то переходим назад, осуществляем еще одну итерацию цикла jg short loc_409F9B

; Когда ESI будет равно 0, пропустим переход и перейдем

; дальше, убедившись, что строки одинаковы

loc_409FCB: ; CODE XREF: \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_+59↑j

; \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_+6B↑j

nop

mov eax, edi ; Готовим результат для возврата sub eax, r13d

loc_409FD1: ; CODE XREF: \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_+1B↑j

; \_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_+27↑j ...

; Восстанавливаем стек add rsp, 20h

; Восстанавливаем регистры pop rbx

pop rsi

pop rdi

pop r13

pop r14 retn

\_ZN6System8\_UStrCmpENS\_13UnicodeStringES0\_ endp



Мы выяснили предназначение этой функции, она сравнивает переданные в параметрах строки и возвращает соответствующий результат, на основе которого главная функция при\- ложения решает, что вывести на экран. Ее анализ был не так прост, как могло показаться вначале!



TURBO-ИНИЦИАЛИЗАЦИЯ СТРОКОВЫХ ПЕРЕМЕННЫХ

Не всегда, однако, различить строки так просто. Чтобы убедиться в этом, достаточно откомпилировать предыдущий пример компилятором Embarcadero C++Builder и заглянуть в полученный файл любым подходящим дизассемблером, скажем IDA Pro. Главная функция, на которой откроется дизассемблер, выглядит так:



main proc near ; DATA XREF: acrtused+29↑o var_38 = dword ptr -38h

![image](19/Image_039.png)

var_34 = dword ptr -34h s1 = byte ptr -2Dh

s0 = byte ptr -1Eh argv = qword ptr -10h argc = dword ptr -8 var_4 = dword ptr -4

push rbp

sub rsp, 60h

lea rbp, \[rsp+60h\]

lea rax, \[rbp+s1\]

lea r8, \[rbp+s0\]

mov \[rbp+var_4\], 0

mov \[rbp+argc\], ecx

mov \[rbp+argv\], rdx

mov rdx, cs:qword_44D000

mov qword ptr \[rbp+s0\], rdx mov ecx, cs:dword_44D008

mov dword ptr \[rbp+s0+8\], ecx mov r9w, cs:word_44D00C

mov word ptr \[rbp+s0+0Ch\], r9w mov rdx, cs:qword_44D00E

mov qword ptr \[rbp+s1\], rdx mov ecx, cs:dword_44D016

mov dword ptr \[rbp+s1+8\], ecx mov r9w, cs:word_44D01A

mov word ptr \[rbp+s1+0Ch\], r9w mov r10b, cs:byte_44D01C

mov \[rbp+s1+0Eh\], r10b mov rcx, r8

mov rdx, rax

call strcmp

cmp eax, 0

jz short loc_40148E

lea rcx, aWoozl ; "Woozl\\n" call printf

...



Переходим по перекрестной ссылке, созданной IDA для aWoozl, в секцию данных только для чтения (.rodata). А там — никаких следов присутствия строк «Hello, Sailor!» и «Hello, World!». Зато обращает на себя внимание какая‑то странная череда слов разной длины:



.rodata:000000000044D000 qword_44D000 dq 57202C6F6C6C6548h ; DATA XREF: main+20↑r

![image](19/Image_040.png)

.rodata:000000000044D008 dword_44D008 dd 646C726Fh ; DATA XREF: main+2B↑r

.rodata:000000000044D00C word_44D00C dw 21h ; DATA XREF: main+34↑r

.rodata:000000000044D00E qword_44D00E dq 53202C6F6C6C6548h ; DATA XREF: main+41↑r

.rodata:000000000044D016 dword_44D016 dd 6F6C6961h ; DATA XREF: main+4C↑r

.rodata:000000000044D01A word_44D01A dw 2172h ; DATA XREF: main+55↑r

.rodata:000000000044D01C byte_44D01C db 0 ; DATA XREF: main+62↑r

.rodata:000000000044D01D aWoozl db 'Woozl',0Ah,0 ; DATA XREF: main+7D↑o

.rodata:000000000044D024 aOk db 'OK',0Ah,0 ; DATA XREF: main:loc_



Что бы это значило? Это не указатели — они никуда не указывают, это не переменные типа int — мы не объявляли таких в программе. Откроем в IDA вкладку Hex View (View → Open subviews → Hex dump) для перехода в hex-режим, и что мы видим? Вот они, наши строки, вот они, родимые:



000000000044D000 48 65 6C 6C 6F 2C 20 57 6F 72 6C 64 21 00 48 65 Hello, World!

![image](19/Image_041.png)

000000000044D010 6C 6C 6F 2C 20 53 61 69 6C 6F 72 21 00 57 6F 6F llo, Sailor!.

000000000044D020 7A 6C 0A 00 4F 4B 0A 00 00 00 00 00 00 00 00 00 zl..OK.......



Хм, почему же тогда IDA Pro посчитала их словами? Ответить на вопрос поможет анализ манипулирующего со строкой кода, но прежде, чем приступить к его исследованию, прев\- ратим эти слова разного размера в нормальные C-строки (клавиша U для преобразования слов разной длины в цепочку бестиповых байтов и клавиша A для преобразования их в строку). Вот как будет выглядеть сегмент данных после примененных операций:



.rodata:000000000044D000 aHelloWorld db 'Hello, World!',0 ; DATA XREF: main+20↑

![image](19/Image_042.png)

.rodata:000000000044D000 ; main+2B↑r ...

.rodata:000000000044D00E aHelloSailor db 'Hello, Sailor!',0 ; DATA XREF: main+41↑

.rodata:000000000044D00E ; main+4C↑r ...

.rodata:000000000044D01D aWoozl db 'Woozl',0Ah,0 ; DATA XREF: main+7D↑

.rodata:000000000044D024 aOk db 'OK',0Ah,0 ; DATA XREF: main:loc



Затем подведем курсор к первой перекрестной ссылке и нажмем клавишу X (Jump → Jump to xref operand…):



![image](19/Image_043.jpg)



[Переход по перекрестной ссылке](https://xakep.ru/wp-content/uploads/2022/04/383034/xref_op.png)



В появившемся окне выбираем любую ссылку (они все ведут примерно в одно и то же мес\- то) и жмем OK. Вновь очутившись в главной функции, обращаем внимание на ее преобра\- жение и приступаем к анализу:



![image](19/Image_044.png)

main proc near ; DATA XREF: acrtused+29↑o var_38 = dword ptr -38h

var_34 = dword ptr -34h s1 = byte ptr -2Dh

s0 = byte ptr -1Eh argv = qword ptr -10h argc = dword ptr -8 var_4 = dword ptr -4

; Откуда взялось столько локальных переменных?!



; Открываем кадр стека push rbp

; Резервируем память для локальных переменных sub rsp, 60h

; Указатель на дно стека lea rbp, \[rsp+60h\]

; В регистр RAX помещается указатель на переменную s1 lea rax, \[rbp+s1\]

; В регистр R8 помещается указатель на переменную s0 lea r8, \[rbp+s0\]

mov \[rbp+var_4\], 0

; Копируем в память два аргумента mov \[rbp+argc\], ecx

mov \[rbp+argv\], rdx



Загружаем в RDX... нет, не указатель на строку "Hello, World!", а восемь первых байтов этой строки! Теперь понятно, почему ошиблась IDA, и оригинальный код (до преобразования в строку) выглядел так: mov rdx, cs:qword_44D000.



Не правда ли, не очень наглядно? И если бы мы изучали не свою, а чужую программу, тогда этот трюк дизассемблера ввел бы нас в глубокое заблуждение!



![image](19/Image_045.png)

mov rdx, qword ptr cs:aHelloWorld ; "Hello, World!"

; Копируем восемь первых байтов строки в локальную переменную s0 mov qword ptr \[rbp+s0\], rdx

; Загружаем байты с восьмого по двенадцатый строки "Hello, World!" в ECX mov ecx, dword ptr cs:aHelloWorld+8 ; "orld!"

; Копируем их в локальную переменную s0. Но мы-то уже знаем,

; что это никакая не переменная s0, а строковый буфер mov dword ptr \[rbp+s0+8\], ecx

; Загружаем оставшийся двухбайтовый хвост строки в R9W mov r9w, word ptr cs:aHelloWorld+0Ch ; "!"

; Записываем его в конец строкового буфера s0 mov word ptr \[rbp+s0+0Ch\], r9w



Итак, строка по частям копируется в локальную переменную s0. Следовательно, как мы уже заметили, — это не просто переменная, а строковый буфер — массив байтов: char s0\[13\];.



Проделываем ту же самую операцию копирования со строкой "Hello, Sailor!":



mov rdx, qword ptr cs:aHelloSailor ; "Hello, Sailor!"

mov qword ptr \[rbp+s1\], rdx

![image](19/Image_046.png)

mov ecx, dword ptr cs:aHelloSailor+8 ; "ailor!" mov dword ptr \[rbp+s1+8\], ecx

mov r9w, word ptr cs:aHelloSailor+0Ch ; "r!" mov word ptr \[rbp+s1+0Ch\], r9w

mov r10b, byte ptr cs:aHelloSailor+0Eh ; "" mov \[rbp+s1+0Eh\], r10b



Копируем строку "Hello, Sailor!" в локальную переменную char s1\[14\].



Подготавливаем параметры для передачи функции strcmp:



; Вспоминаем, в регистре R8 находится указатель на переменную s0 —

![image](19/Image_047.png)

; первую строку: "Hello, World!" mov rcx, r8

; Тогда как в регистре RAX находится указатель на переменную s1 —

; вторую строку: "Hello, Sailor!" mov rdx, rax



Параметры на своих местах, вызываем функцию, прототип которой принял вид strcmp( "Hello, World!", "Hello, Sailor!");



call strcmp

![image](19/Image_048.png)



После проверки строк выше выясняется, что они не равны, поэтому инструкция cmp устанав\- ливает флаг ZF в 0...



![image](19/Image_049.png)

cmp eax, 0

; ...поскольку ZF == 0, переход по метке не делается jz short loc_40148E

; В таком случае в регистр RCX загружаем указатель на строку "Woozl\\n"... lea rcx, aWoozl ; "Woozl\\n"

; ... передаем его функции printf и выводим на консоль call printf

mov \[rbp+var_34\], eax

; Выполняем безусловный переход jmp short loc_40149D

; ---------------------------------------------------------------------------

loc_40148E: ; CODE XREF: main+7B↑j

; Сюда происходит переход, когда строки равны

; В таком случае загружаем в регистр RCX указатель на строку "OK\\n"… lea rcx, aOk ; "OK\\n"

; ...передаем его функции printf и выводим на терминал call printf

mov \[rbp+var_38\], eax loc_40149D: ; CODE XREF: main+8C↑j

; Сюда осуществляется безусловный переход

; Обнуляем переменную и регистр mov \[rbp+var_4\], 0

mov eax, \[rbp+var_4\]

; Восстанавливаем стек add rsp, 60h

; Закрываем кадр стека pop rbp

; Возвращаем ноль retn

main endp



Очень своеобразно в отношении строк ведет себя компилятор C++Builder. Такой подход он получил в наследство от своего давнего предка — компилятора Turbo C. Понятно, что такой способ манипуляции строками снижает производительность приложения по причине лиш\- них операций копирования данных между регистрами и памятью. Вдобавок он усложняет дизассемблирование программы, что можно записать ему в плюсы, однако трудности воз\- никнут лишь у совсем юных кодокопателей. И, как мы увидели, с помощью IDA их легко мож\- но решить.



ЗАКЛЮЧЕНИЕ

Символы, строки и различные кодировки по‑прежнему очень важны при исследовании при\- ложений. Созданные за последние 30 лет стандарты хранения, передачи и представления символов позволили прийти к некоему балансу и решению многих проблем прошлого. В то же время разные компиляторы решают вопросы с кодировками по‑своему, в связи с чем для хакеров ситуация не стала ясней. Напротив, из‑за многобайтного и переменного способа обработки символов задача заметно усложнилась. Но, как мы увидели, для всякой проб\- лемы есть решение! Внимательность, острый ум и хороший дизассемблер позволят раз\- грызть любую программу.



Кроме строк, в рассмотренных выше примерах нам встретились условные переходы. Их ана\- лиз не самое простое занятие, так как типов переходов довольно много, поэтому в сле\- дующей статье мы подробно обсудим идентификацию условных операторов if-then-else.



![image](19/Image_050.jpg)

Оцени статью:





[Крис Касперски](https://xakep.ru/author/kaspersky/)



Известный российский хакер. Легенда \]\[, ex-редактор ВЗЛОМа. Также известен под псевдонимами мыщъх, nezumi (яп. 鼠, мышь), n2k, elraton, souriz, tikus, muss, farah, jardon, KPNC.





[Юрий Язев](https://xakep.ru/author/yurembo/)



Широко известен под псевдонимом yurembo. Программист, разработчик видеоигр, независимый исследователь. Старый автор журнала «Хакер».



[www.t2d-dev.ru](http://www.t2d-dev.ru/)



![image](19/Image_051.jpg)

[Теги:](https://xakep.ru/tag/assembler/) [assembler](https://xakep.ru/tag/windows/) [Windows](https://xakep.ru/tag/vzlom/) [Взлом Выбор](https://xakep.ru/tag/vybor-redaktora/) [редактора](https://xakep.ru/tag/guides/) [Гайды Статьи](https://xakep.ru/tag/articles/)





Подписаться

![image](19/Image_052.png)



![image](19/Image_053.png)