# Определяем курс выполнения программ при реверсинге

Сегодня мы приступим к подробному изучению условного хода выполнения программ. Без него невозможно представить программирование как таковое. Мы начнем наше погружение с условных операторов if — then — else и разберемся, как компилятор интерпретирует подобные конструкции.

Раньше мы уже встречались с условным ходом выполнения программ, однако ограничивались короткими описаниями команд и поверхностным разбором выполняемых ими операций. Управляемый ход выполнения, без сомнения, самая важная веха развития программирования своего времени и затмевает многие последующие, даже такие, как структурное программирование или ООП. Только с высоты высокоуровневого языка кажется, что в условии if — then — else нет ничего любопытного и оно лишено какого-либо разнообразия. Но для компилятора это простор для самодеятельности! И в блzизких по духу ситуациях он может построить кардинально различающийся код.

В процессорной архитектуре x86 предусмотрен весьма «оригинальный» набор команд для перенаправления хода выполнения программного кода. «Оригинальный» не значит хороший или плохой, это значит, что он отличается от набора микропроцессоров других архитектур и производителей: ARM, MIPS, SPARC и так далее. Как ты знаешь, первый процессор этой серии создавался впопыхах для временной замены еще не готового iAPX-432, последний, в свою очередь, должен был уметь в многозадачность и управление памятью на аппаратном уровне. Чего в итоге не случилось. А x86 продолжил свое развитие. Поэтому сейчас, когда за годы эволюции процессоров серии x86 скопился набор инструкций, хакерам приходится разгребать весь этот хлам, дабы раскрутить порядок выполнения машинных инструкций.

## ИДЕНТИФИКАЦИЯ УСЛОВНЫХ ОПЕРАТОРОВ

Существует два вида алгоритмов — безусловные и условные. Порядок действий безусловного алгоритма всегда постоянен и не зависит от входных данных. Например, a = b + c. Порядок действий условных алгоритмов, напротив, зависит от данных, поступающих «на вход». Например:

> если c не равно нулю,
>
> то: a = b/c;
>
> иначе: вывести сообщение об ошибке

Обрати внимание на выделенные жирным шрифтом ключевые слова «если», «то» и «иначе», называемые операторами условия или условными операторами. Без них не обходится ни одна программа (вырожденные примеры наподобие «Hello, World!» не в счет). Условные операторы — сердце любого языка программирования. Поэтому чрезвычайно важно уметь их правильно определять.

В общем виде (не углубляясь в синтаксические подробности отдельных языков) оператор условия схематично изображается так:

> IF (условие) THEN { оператор_1; оператор_2; } ELSE { оператор_a; оператор_b; }

Задача компилятора — преобразовать эту конструкцию в последовательность машинных команд, выполняющих оператор_1, оператор_2, если условие истинно, и оператор_a, оператор_b — если оно ложно. Однако микропроцессоры серии 80x86 поддерживают весьма скромный набор условных команд, ограниченный фактически одними условными переходами. Программистам, знакомым лишь с IBM PC, такое ограничение не покажется чем-то неестественным, между тем существует масса процессоров, поддерживающих префикс условного выполнения инструкции. То есть вместо того, чтобы писать

```asm
TEST ECX,ECX
JNZ xxx
MOV EAX,0x666
```

там поступают так:

```asm
TEST ECX,ECX
IFZ MOV EAX,0x666
```

**IFZ** и есть префикс условного выполнения, разрешающий выполнение следующей команды только в том случае, если установлен флаг нуля. В этом смысле микропроцессоры 80x86 можно сравнить с ранними диалектами язы- ка Basic, не разрешающими использовать в условных выражениях никакой другой оператор, кроме GOTO. Сравни:

### Старый диалект Basic:

```basic
10 IF A=B THEN GOTO 30
20 GOTO 40
30 PRINT "A=B"
40 ... // Прочий код программы
```

### Новый диалект Basic:

```basic
IF A=B THEN PRINT "A=B"
```

Если ты когда-нибудь программировал на старых диалектах Basic, то, вероятно, помнишь, что гораздо выгоднее выполнять GOTO, если условие ложно, а в противном случае продолжать нормальное выполнение программы. Как видишь, вопреки расхожему мнению, навыки программирования на Basic отнюдь не бесполезны, особенно в дизассемблировании программ.

Большинство компиляторов (даже не оптимизирующих) инвертируют истинность условия, транслируя конструкцию

> IF (условие) THEN { оператор_1; оператор_2; }

в следующий псевдокод:

```basic
IF (NOT условие) THEN continue
оператор_1;
оператор_2;
continue:
…
```

Следовательно, для восстановления исходного текста программы нам придется вновь инвертировать условие и «подцепить» блок операторов { оператор_1; оператор_2; } к ключевому слову **THEN**. Иначе говоря, если откомпилированный код выглядит так:

```basic
10 IF A<>B THEN 30
20 PRINT "A=B"
30 ...// Прочий код программы
```

можно с уверенностью утверждать, что в исходном тексте присутствовали строки `IF A=B THEN PRINT "A=B"`. А если программист, наоборот, проверял переменные A и B на неравенство, то есть в коде присутствовала конструкция `IF A<>B THEN PRINT "A<>B"`? Все равно компилятор инвертирует истинность условия и сгенерирует следующий код:

```basic
10 IF A=B THEN 30
20 PRINT "A<>B"
30 ... // Прочий код программы
```

Конечно, встречаются и компиляторы, страдающие многословием. Их легко распознать по безусловному переходу, следующему сразу же после условного оператора:

```basic
IF (условие) THEN do
GOTO continue
do:
оператор_1;
оператор_2;
continue:
…
```

В таком случае инвертировать условие не нужно. Впрочем, если это сделать, ничего страшного не произойдет, разве что код программы станет менее понятным, да и то не всегда.

Рассмотрим теперь, как транслируется полная конструкция

> **IF** (условие) **THEN** { оператор_1; оператор_2; } **ELSE** { оператор_a; оператор_b; }

Одни компиляторы поступают так:

```basic
IF (условие) THEN do_it
// Ветка ELSE
оператор_a;
оператор_b;
GOTO continue
do_it:
// Ветка IF
оператор_1;
оператор_2;
continue:
```

А другие — так:

```basic
IF (NOT условие) THEN else
// Ветка IF
оператор_1;
оператор_2;
GOTO continue else:
// Ветка ELSE
оператор_a;
оператор_b
continue:
```

Разница между ними в том, что вторые инвертируют истинность условия, а первые — нет. Поэтому, не зная «нрава» компилятора, определить, как выглядел подлинный исходный текст программы, невозможно! Однако это не создает проблем, ибо условие всегда можно записать так, как это удобно. Допустим, не нравится тебе вот такая конструкция:

> IF (c <> 0) THEN a = b/c ELSE PRINT "Ошибка!"

Пиши ее так:

> IF (c == 0) THEN PRINT "Ошибка!" ELSE a = b/c

И никаких гвоздей!

## ТИПЫ УСЛОВИЙ

Условия делятся на простые (элементарные) и сложные (составные). Пример первых: `if (a==b) ...`, вторых: `if ((a==b) && (a!=0)) ...`. Очевидно, что любое сложное условие можно разложить на ряд простых. Вот с простых условий мы и начнем.

Существуют два основных типа элементарных условий: условия отношений («меньше», «равно», «больше», «меньше или равно», «не равно», «больше или равно», обозначаются как `<`, `==`, `>`, `<=`, `!=`, `>=`) и логические условия («И», «ИЛИ», «НЕ», в С-нотации обозначаются так: `&&`, `||`, `!`). Еще в отдельную категорию можно выделить битовые проверки. Их объединяют с логическими, однако несколько некорректно смешивать в одну кучу людей и коней, даже если они чем-то взаимосвязаны.

Если условие истинно, оно возвращает булево значение `TRUE`, если ложно — `FALSE`. Внутреннее (физическое) представление булевых переменных зависит от конкретной реализации и может быть любым. По общепринятому соглашению `FALSE` равно нулю, а `TRUE` **не** равно нулю. Часто (но не всегда) `TRUE` равно единице, но на это нельзя полагаться! Так, код `IF ((a>b)!=0) ...` абсолютно корректен, а вот `IF ((a>b)==1) ...` привязан к конкретной реализации и потому нежелателен.

Обрати внимание: `IF ((a>b)!=0) ...` проверяет на неравенство нулю отнюдь не значения самих переменных a и b, а именно результат их сравнения. Рассмотрим следующий пример:

> IF ((666==777)==0) printf("Woozl!");

Как ты думаешь, что отобразится на экране, если его запустить? Правильно — **Woozl**! Почему? Ведь ни 666, ни 777 не равно нулю! Да, но ведь `666 != 777`, следовательно, условие `(666==777)` ложно, то есть равно нулю. Кстати, если записать `IF ((a=b)==0) ...`, получится совсем иной результат — значение переменной b будет присвоено переменной a и потом проверено на равенство нулю.

Логические условия чаще всего используются для связывания двух или более элементарных условий отношения в составное. Например, `IF ((a==b) && (a!=0)) ...`. При трансляции программы компилятор всегда выполняет развертку составных условий в простые. В данном случае это происходит так: `IF a==b THEN IF a=0 THEN ...`. На втором этапе выполняется замена условных операторов на оператор `GOTO`:

```basic
IF a != b THEN continue
IF a == 0 THEN continue
... // Код условия
:continue
... // Прочий код
```

Порядок вычисления элементарных условий в сложном выражении зависит от прихотей компилятора, гарантируется лишь, что условия, «связанные» операцией логического «И», проверяются слева направо в порядке их объявления в программе. Причем если первое условие ложно, то следующее за ним вычислено не будет! Это дает возможность писать код наподобие следующего: `if ((filename) && (f=fopen(&filename[0],"rw")))...` — если указатель filename указывает на невыделенную область памяти (то есть, попросту говоря, содержит ноль — логическое FALSE), функция fopen не вызывается и ее краха не происходит. Такой способ вычислений получил название быстрых булевых операций (теперь-то ты знаешь, что подразумевается под «быстротой»).

Перейдем теперь к вопросу идентификации логических условий и анализу сложных выражений. Вернемся к уже облюбованному нами выражению `if (( a==b) && (a!=0)) ...` и вглядимся в результат его трансляции.

![image](./images/20/ifthen_1.png)

Легко видеть — он выдает себя серией условных переходов к одной и той же метке, причем, обрати внимание, выполняется проверка на неравенство каждого из элементарных условий, а сама метка расположена позади кода условия.

Идентификация логической операции «ИЛИ» намного сложнее в силу неоднозначности ее трансляции. Рассмотрим это на примере выражения `if((a==b) || (a!=0)) ...` Его можно разбить на элементарные операции так.

![image](./images/20/ifthen_2.png)

И так.

![image](./images/20/ifthen_3.png)

Первый вариант обладает весьма запоминающейся внешностью — серия проверок (без инверсии условия) на одну и ту же метку, расположенную перед кодом условия, а в конце этой серии — безусловный переход на метку, расположенную позади кода условия.

Однако оптимизирующие компиляторы выкидывают безусловный переход, инвертируя проверку последнего условия в цепочке и соответственно меняя адрес перехода. По неопытности эту конструкцию часто принимают за смесь `OR` и `AND`.

Кстати, о смешанных операциях. Рассмотрим результат трансляции следующего выражения: `if ((a==b) || (a==c) && (a!=0)) ...`:

```basic
IF a==b THEN check_null
IF a!=c THEN continue
check_null:
IF a==0 THEN continue
... // Код условия
continue:
... // Прочий код
```

Как из непроходимого леса элементарных условий получить одно удобочитаемое составное условие? Начинаем плясать от печки, то есть от первой операции сравнения. Смотри: если условие a==b окажется истинно, оно «выводит из игры» проверку условия a!=c. Такая конструкция характерна для операции OR, то есть достаточно выполнения хотя бы одного условия из двух для срабатывания кода.

Пишем в уме или карандашом: `if ((a==b) || ...)`, далее — если условие `(a!=c)` истинно, все дальнейшие проверки прекращаются и управление передается на метку, расположенную позади условного кода. Логично предположить, что мы имеем дело с последней операцией OR в цепочке сравнений — это ее почерк.

Значит, мы инвертируем условие выражения и продолжаем писать: `if ((a==b) || (a==c)...)`. Последний бастион — проверка условия `a==0`. Выполнить условный код, миновав условие, не удастся, следовательно, это не OR, а AND! A AND всегда инвертирует условие срабатывания, и поэтому оригинальный код должен был выглядеть так: `if ((a==b) || (a==c) && (a!=0))`. Ура! У нас получилось!

Впрочем, то, что мы рассмотрели, — это простейший пример. В реальной жизни оптимизирующие компиляторы такого понаворочают...

### Наглядное представление сложных условий в виде дерева

Конструкцию, состоящую из трех-четырех элементарных условий, можно проанализировать и в уме (если есть соответствующие навыки), но хитросплетения пяти и более условий образуют самый настоящий лабиринт — его с ходу не возьмешь. Неоднозначность трансляции сложных условий порождает неоднозначность интерпретации, что приводит к многовариантному анализу, причем с каждым шагом в голове приходится держать все больше и больше информации. Так недолго и крышей поехать или окончательно запутаться и получить неверный результат.

Выход в использовании двухуровневой системы ретрансляции. На первом этапе элементарные условия приводятся к некоторой промежуточной форме записи, наглядно и непротиворечиво отображающей взаимосвязь элементарных операций. Затем выполняется окончательная трансляция в любую подходящую нотацию (например, С, Basic или Pascal).

Единственная проблема — выбрать удачную промежуточную форму. Существует множество решений, но по соображениям экономии пространства мы рассмотрим только одно — деревья.

Изобразим каждое элементарное условие в виде узла с двумя ветвями, соответствующими состояниям: условие истинно и условие ложно. Для наглядности обозначим «ложь» равнобедренным треугольником, а «истину» — квадратом и условимся всегда располагать ложь на левой, а истину — на правой ветке. Получившуюся конструкцию назовем гнездом (nest).

| ![Схематическое представление гнезда](./images/20/sch_1.png)
|:--:|
| *Схематическое представление гнезда* |

Гнезда могут объединяться в деревья, соединяясь узлами с ветками другого узла. Причем каждый узел может соединяться только с одним гнездом, но всякое гнездо может соединяться с несколькими узлами. Непонятно? Не волнуйся, сейчас мы со всем этим разберемся.

Рассмотрим объединение двух элементарных условий логической операцией `AND` на примере выражения `((a==b) && (a!=0))`. Извлекаем первое слева условие `(a==b)`, «усаживаем» его в гнездо с двумя ветвями: левая соответствует случаю, когда `a!=b` (то есть условие `a==b` ложно), а правая, соответственно, наоборот. Затем то же самое делаем и со вторым условием `(a!=0)`.

У нас получаются два очень симпатичных гнездышка, остается лишь связать их меж собой операцией логического `AND`. Как известно, `AND` выполняет второе условие только в том случае, если истинно первое. Значит, гнездо `(a!=0)` следует прицепить к правой ветке гнезда `(a==b)`. Тогда правая ветка гнезда `(a!=0)` будет соответствовать истинности выражения `((a==b) && ( a!=0))`, а обе левые ветки — его ложности. Обозначим первую ситуацию меткой `do_it`, а вторую — `continue`. В результате дерево должно принять следующий вид.

| ![Графическое представление операции AND в виде двоичного дерева](./images/20/sch_2.png)
|:--:|
| *Графическое представление операции AND в виде двоичного дерева* |

Для наглядности отметим маршрут из вершины дерева к метке `do_it` жирной стрелкой. Как видишь, в пункт `do_it` можно попасть только одним путем. Вот так графически выглядит операция `AND`.

Теперь перейдем к операции логического OR. Рассмотрим конструкцию `((a==b) || (a!=0))`. Если условие `(a==b)` истинно, то и все выражение считается истинным. Следовательно, правая ветка гнезда `(a==b)` связана с меткой do_it. Если же условие `(a==b)` ложно, то выполняется проверка следующего условия. Значит, левая ветка гнезда `(a==b)` связана с гнездом `(a!=0)`.

Очевидно, если условие `(a!=0)` истинно, то истинно и все выражение `(( a==b) || (a!=0))`. Напротив, если условие `(a!=0)` ложно, то ложно и все выражение, так как проверка условия `(a!=0)` выполняется только в том случае, если условие `(a==b)` ложно. Отсюда мы заключаем, что левая ветка гнезда `(a!=0)` связана с меткой `continue`, а правая — с `do_it`. Обрати внимание, в пункт `do_it` можно попасть двумя различными путями! Вот так графически выглядит операция `OR`.

| ![Графическое представление операции OR в виде двоичного дерева](./images/20/sch_3.png)
|:--:|
| *Графическое представление операции OR в виде двоичного дерева* |

До сих пор мы отображали логические операции на деревья, но ведь деревья создавались как раз для противоположной цели — преобразования последовательности элементарных условий в интуитивно понятное представление. Давай займемся этим. Пусть в тексте программы встретится следующий код:

```basic
IF a==b THEN check_null
IF a!=c THEN continue
check_null:
IF a==0 THEN continue
... // Код условия continue:
... // Прочий код
```

Извлекаем условие `(a==b)` и сажаем его в «гнездо», смотрим: если оно ложно, то выполняется проверка `(a!=c)`, значит, гнездо `(a!=c)` связано с левой веткой гнезда `(a==b)`. Если же условие `(a==b)` истинно, то управление передается метке `check_null`, проверяющей истинность условия `(a==0)`. Следовательно, гнездо `(a==0)` связано с правой веткой гнезда `(a==b)`. В свою очередь, если условие `(a!=с)` истинно, управление получает метка continue, в противном случае — check_null. Значит, гнездо `(a==0)` связано одновременно и с правой веткой гнезда `(a==b)`, и с левой веткой гнезда `(a!=c)`.

Конечно, это проще рисовать, чем описывать! У тебя должно получиться дерево, очень похожее на следующее изображение.

| ![Графическое представление сложного выражения](./images/20/sch_4.png)
|:--:|
| *Графическое представление сложного выражения* |

Смотри: к гнезду `(a==0)` можно попасть двумя путями — либо через гнездо `(a==b)`, либо через цепочку двух гнезд `(a==b)` → `(a!=c)`. Следовательно, эти гнезда связаны операцией `OR`. Записываем: `if ((a==b) || !(a!=c)) ....`

Откуда взялся NOT? Так ведь гнездо `(a==0)` связано с левой веткой гнезда `(a!=c)`, то есть проверяется ложность его истинности! Кстати, «ложность истинности» очень хорошо звучит. Избавляемся от `NOT`, инвертируя условие: `if ((a==b) || (a==c)...)...` Далее, из гнезда `(a==0)` до пункта `do_it` можно добраться только одним путем, значит, оно связано операцией `AND`. Записываем: `if (((a==b) || (a==c)) && !(a==0))`... Теперь избавляемся от лишних скобок и операции NOT. В результате получается: `if (((a==b) || (a==c)) && (a!=0)) {/* Код условия */}`.

Не правда ли, все просто? Причем вовсе не обязательно строить деревья «вручную», при желании можно написать программу, берущую эту работу на себя.

### Исследование конкретных реализаций

Прежде чем приступать к отображению конструкции

> IF (сложное условие) THEN оператор_1;оператор_2; ELSE оператор_а;оператор_b

на машинный язык, вспомним, что, во-первых, агрегат `IF — THEN — ELSE` можно выразить через `IF — THEN`, во-вторых, `THEN` оператор_1;оператор_2 можно выразить через `THEN GOTO do_it`, в-третьих, любое сложное условие можно свести к последовательности элементарных условий отношения. Таким образом, на низком уровне мы будем иметь дело лишь с конструкциями вида

> IF (простое условие отношения) THEN GOTO do_it

А уже из них, как из кирпичиков, можно сложить что угодно.

Итак, что такое **условия отношения**, или, другими словами, результат

операции сравнения двух чисел? В микропроцессорах Intel 80x86 целочисленные значения сравниваются командой CMP, а с вещественными значениями ситуация сложилась не такая однозначная.

С момента появления первого процессора серии 8086 для работы с вещественными числами использовался сопроцессор. Он представлял собой обособленную микросхему, для которой на материнской плате предназначался отдельный разъем. Для сравнения вещественных значений с помощью сопроцессора использовалась одна из следующих инструкций: `FCOM`, `FCOMP`, `FCOMPP`, `FCOMI`, `FCOMIP`, `FUCOMI`, `FUCOMIP`.

С выходом процессора 80486 сопроцессор (или модуль операций с плавающей запятой) был объединен в одном корпусе с АЛУ, устройством управления памятью, кешем и прочим. Несмотря на это, он по-прежнему был отдельным устройством. С появлением Pentium MMX наметилась тенденция к глубокой интеграции этого модуля в основной процессор с использованием общих регистров. Для программиста, однако, никаких видимых изменений не произошло, в коде по-прежнему использовались команды сопроцессора.

Ветер перемен стал ощущаться вместе с выходом Pentium III. К его набору команд был добавлен SSE первой версии, выполнение команд которого реализовывало параллелизм на уровне данных — SIMD-параллелизм (англ. Single Instruction, Multiple Data, одна инструкция — множество данных). В результате SSE добавил 70 новых инструкций и восемь 128-битных регистров XMM0 — XMM7. Надо отметить, что процессорное расширение MMX тоже имело возможность реализовывать SIMD-параллелизм. Однако сопроцессор не мог работать параллельно с основным.

Когда вышла следующая марка процессора — Pentium IV, до второй версии был обновлен и SSE. Он получил еще 144 новые инструкции. Начиная с SSE2 стало возможным безболезненно отказаться от унаследованной технологии MMX и применить новый набор команд для работы с вещественными значениями. В этом наборе для сравнения чисел с плавающей запятой используются следующие команды: CMPPD, CMPSD, COMISD, UCOMISD.

Несмотря на появление нового поколения инструкций SSE для работы с вещественными значениями, инструкции прошлого поколения (именуемого x87 — от названия линии сопроцессоров) также поддерживаются современными микропроцессорами. Очевидно, для сохранения совместимости с прошлым рядом микропроцессоров и унаследованным программным обеспечением.

>**Развитие процессорного расширения SSE**
>
>* SSE — Intel Pentium III (начало 1999 года). SSE может выполнять 70 новых инструкций, которые используют 128-разрядные регистры ХММ0-ХММ7, MMX и даже обычные 32-битные регистры.
>* SSE2 — Intel Pentium IV (июнь 2000 года). SSE2-инструкции значительно усовершенствовали SSE первого поколения, добавив 144 новые инструкции.
>* SSE3 — Intel Pentium 4 на базе 90-нм (февраль 2004 года). Добавлены дополнительные 13 инструкций, которыерасширяют математические возможности DSP, и поток управления инструкциями.
>* SSSE3 — дополнительное обновление для SSE3, которое включает 16 новых инструкций. Позволяет работать как с 64-битными (MMX), так и с 128-битными (XMM) регистрами, вносить слова в инструкции и переставлять байты в них.
>* SSE4 — Intel Core 2 Duo (середина 2007 года). Добавлены 54 инструкции.
>* В SSE 4.1 (появилась в Intel Penryn) содержится урезанный набор из 47 инструкций, а в SSE 4.2 — полный из 54 (семь инструкций добавлены в архитектуре Nehalem). Инструкции работают только с 128-битными регистрами XMM0 — XMM15. Добавлены POPCNT и дополнительные целочисленные инструкции. Основным направлением при разработке было расширение мультимедийных возможностей: декодирование видео, ускорение физических и векторных расчетов для игр, ускорение обработки фото, музыки.
>* С середины 2008 года Intel занимается разработкой и развитием AVX — расширения системы команд для x86-совместимых микропроцессоров. Впоследствии вышли расширения AVX-2, AVX-512. Эти расширения добавили к архитектуре процессора шестнадцать 256-битных регистров YMM0 — YMM15. Обсуждение AVX выходит за рамки рассматриваемого материала

Не будем подробно останавливаться на этих инструкциях, вместо этого вкратце вспомним их, чтобы освежить в памяти.

* **CMP** — эта команда эквивалентна операции целочисленного вычитания SUB, за одним исключением — в отличие от SUB, CMP не изменяет операндов, а воздействует лишь на флаги основного процессора: флаг нуля, флаг переноса, флаг знака и флаг переполнения.

* **Флаг нуля** — `ZF` устанавливается в единицу, если результат вычитания равен нулю, то есть операнды равны друг другу.

* **Флаг переноса** — `CF` устанавливается в единицу, если при вычитании произошел заем из старшего бита уменьшаемого операнда, то есть уменьшаемое меньше вычитаемого.

* **Флаг знака** — `SF` равен старшему, знаковому биту результата вычислений, то есть результат вычислений — отрицательное число.

* **Флаг переполнения** — `OF` устанавливается в единицу, если результат вычислений «залез» в старший бит, приводя к потере знака числа.

    Редко используемые:

* **Флаг полупереноса** — `AF` устанавливается в единицу, если в результате предыдущей операции произошел перенос (или заем) из третьего бита в четвертый. Этот флаг используется автоматически командами двоично-десятичной коррекции (BCD).

* **Флаг четности** — `PF` устанавливается в единицу, если младший байт результата предыдущей команды содержит четное количество битов, равных 1\. Если количество единиц в младшем байте нечетное, то этот флаг устанавливается в 0.

>**Эволюция регистра флагов**
>
>* 16-битный регистр флагов, называемый FLAGS (в микропроцессорах с 8086 по 80286, ныне оставлены для обратной совместимости), содержит 16 флагов, из которых четыре зарезервированы.
>* 32-битный регистр флагов — EFLAGS (в микропроцессорах с 80386 по Pentium 4, далее поддержка по обратной совместимости), содер жит 32 бита в регистре флагов, среди которых добавлено шесть активных флагов и десять зарезервированных.
>* 64-раз рядный регистр флагов — RFLAGS (начиная с Pentium 4 на ядре Prescott) добавил 32 новых флага, каждый из которых зарезервирован для будущих нужд, поэтому их нельзя использовать. Вдруг в следующей модели процессора один из них будет отвечать за самоуничтожение компа или выход на околоземную орбиту! Шутка, но в ней есть доля правды.
> ![Формат регистра флагов FLAGS](./images/20/form_flags.png)

Для проверки состояния флагов существует множество команд условных переходов, выполняющихся в случае, если определенный флаг (набор флагов) установлен (или сброшен). Инструкции, использующиеся для анализа результата сравнения целых чисел, приведены в таблице 1.

| ![Таблица 1: соответствие операций отношения командам процессора](./images/20/table_1.png)
|:--:|
| *Таблица 1: соответствие операций отношения командам процессора* |

Обрати внимание, что флаг **OF** предназначен обнаруживать переполнение при работе со знаковыми операндами; когда вычисления ведутся с беззнаковыми числами, эта работа возлагается на флаг **CF**.

В общем случае конструкция **IF** (элементарное условие отношения) **THEN** **do_it** транслируется в следующие команды процессора:

```asm
CMP A,B
Jxx do_it
continue:
```

Между инструкциями `CMP` и `Jxx` могут находиться и другие команды, не изменяющие флагов процессора, например `MOV`, `LEA`.

### Сравнение вещетвенных чисел

| ![Таблица 2: команды сравнения вещественных значений](./images/20/table_2.png)
|:--:|
| *Таблица 2: команды сравнения вещественных значений* |

Команды сравнения вещественных чисел из набора SSE2, такие как COMISD (полный их набор см. на рисунке выше), равно как команды целочисленного сравнения, воздействуют на регистры основного процессора. Это отличает их от команд расширения х87, которые воздействуют на регистры сопроцессора.

Главная проблема заключается в том, что в природе не существует инструкций условного перехода, управляемых флагами сопроцессора! К тому же флаги сопроцессора непосредственно недоступны, и, чтобы прочитать их статус, необходимо было выгрузить регистр состояния сопроцессора SW

в память или регистр общего назначения основного процессора. Хуже всего — анализировать флаги придется «вручную»! Если при сравнении целых чисел можно и не задумываться, какими именно флагами управляется условный переход, достаточно написать:

```asm
CMP A,B
JGE do_it; (Jump [if] Great [or] Equal — прыжок, если A больше или равно B)
```

то с сопроцессором этот номер не пройдет! Правда, можно схитрить и скопировать флаги сопроцессора в регистр флагов основного процессора, а затем использовать родные инструкции условного перехода из серии Jxx.

Конечно, непосредственно скопировать флаги из сопроцессора в основной процессор было нельзя, и эту операцию приходилось выполнять в два этапа. Сначала флаги FPU выгрузить в память или регистр общего назначения, а уже оттуда заталкивать в регистр флагов CPU. Непосредственно модифицировать регистр флагов CPU умеет только одна команда — POPF. Остается только выяснить, каким флагам сопроцессора какие флаги основного процессора соответствуют. А это отдельная наука! Ей надо было заниматься всякий раз при сравнении вещественных чисел! Хорошо, что технологии не стоят на месте и теперь мы можем использовать команды из расширения SSE и забыть про сопроцессор, как про страшный сон!

Давай рассмотрим следующий крайне простой пример (compare_double), откомпилированный в Visual Studio 2022 в режиме Release и с отключенной оптимизацией:

```cpp
#include <iostream>

int main()
{
    double num1 = 0.0001;
    double num2 = 0.0002;

    if (num1 < num2)

        std::cout << "num1 меньше, чем num2" << "\\n";

    else
        std::cout << "num1 больше либо равно num2" << "\\n";
}
```

Ключевая активность здесь — это сравнение вещественных чисел двойной точности. Загрузим результирующий экзешник в IDA и посмотрим, какую инс- трукцию для сравнения выбрал наш компилятор:

```asm
; int cdecl main(int argc, const char **argv, const char **envp)
main proc near ; CODE XREF:

__scrt_common_main_seh+107↓p

ExceptionDir↓o
; Две локальные переменные
var_18 = qword ptr -18h
var_10 = qword ptr -10h
; Выделяем для них память
sub rsp, 38h
; DATA XREF: .pdata:
```

Копируем вещественное число двойной точности из сегмента данных только для чтения:

`.rdata:0000000140003330 real@3f1a36e2eb1c432d dq 0.0001 ; DATA XREF: main+4↑r` в регистр `XMM0`.

```asm
movsd xmm0, cs: real@3f1a36e2eb1c432d
; Из регистра копируем значение в локальную переменную var_10
movsd [rsp+38h+var_10], xmm0
```

Копируем вещественное число двойной точности из сегмента данных только для чтения:

`.rdata:0000000140003338 real@3f2a36e2eb1c432d dq 0.0002 ; DATA XREF: main+12↑r` в регистр `XMM0`.

```asm
movsd xmm0, cs: real@3f2a36e2eb1c432d
; Из регистра копируем значение в локальную переменную var_18
movsd [rsp+38h+var_18], xmm0
; Возвращаем второе значение в регистр XMM0
movsd xmm0, [rsp+38h+var_18]
```

Ага! Для сравнения двух вещественных значений компилятор выбрал инструкцию COMISD! Заглянем в таблицу 2, чтобы узнать, присутствует ли она там. Действительно, она тут как тут! Из этого следует, что данная инструкция вхо- дит в SSE2.

Обрати внимание: к моменту сравнения в XMM0 находится второе, большее число — 0.0002. А в переменной var_10 — первое, меньшее, — 0.0001. Когда первый операнд больше второго, инструкция COMISD устанавливает флаги следующим образом: ZF = 0; PF = 0; CF = 0.

`comisd xmm0, [rsp+38h+var_10]`

Инструкция JBE делает переход по метке, когда флаги CF либо ZF установлены в единицу. Иными словами, когда первый операнд меньше второго или равен ему. В нашем случае ни одно из перечисленных условий не соблюдено, поэтому переход на метку не выполняется.

`jbe short loc_140001052`

Изначально в этом дизассемблерном листинге строки в сегменте данных были представлены набором неопределенных байтов. Как придать им удобочитаемый вид, мы уже проходили.

```asm
; В регистр RDX грузим указатель на строку
lea rdx, aNum1MenEEmNum2 ; "num1 меньше, чем num2"
; В RCX помещаем указатель на символьный вывод
mov rcx, cs:std::basic_ostream<char,std::ch ar_traits<char>> std::cout ; _Ostr

; Вызываем функцию вывода строки на консоль, передавая ей параметры
call std::operator<<<std::char_traits<char>>(std::basic_ostream<char,std::char_traits<char>> &,char const *)

; Добавляем завершающий символ перехода на новую строку
lea rdx, asc_1400032D0 ; "\\n"
mov rcx, rax ; _Ostr

; Его тоже выводим
call std::operator<<<std::char_traits<char>>(std:: basic_ostream<char,std::char_traits<char>> &,char const *)

; Хо-хо! Безусловный переход
jmp short loc_140001074
;---------------------------------------------------------------------------
loc_140001052: ; CODE XREF: main+2C↑j
lea rdx, aNum1BolELiboRa ; "num1 больше либо равно num2"
mov rcx, cs:std::basic_ostream<char,std::char_traits<char>> std::cout ; _Ostr
call std::operator<<<std::char_traits<char>>(std:: basic_ostream<char,std::char_traits<char>> &,char const *)
lea rdx, asc_1400032F8 ; "\\n"
mov rcx, rax ; _Ostr
call std::operator<<<std::char_traits<char>>(std:: basic_ostream<char,std::char_traits<char>> &,char const *)
loc_140001074: ; CODE XREF: main+50↑j

; Завершаем выполнение приложения,
; возвращаем ноль
xor eax, eax

; Восстанавливаем стек
add rsp, 38h
retn
main endp
```

В результате, как и ожидалось, будет такой вывод.

![img](./images/20/console_1.png)

Любопытно увидеть, какой бинарник приведенного выше кода приготовит C++Builder. Итак, дизассемблерный листинг, построенный с отключенной оптимизацией программы, должен выглядеть примерно следующим образом:

```asm
; int cdecl main(int argc, const char **argv, const char **envp)
public main
main proc near ; DATA XREF: acrtused+29↑o
; Семь локальных переменных... Куда так много?! var_30 = qword ptr -30h
var_28 = qword ptr -28h
var_20 = qword ptr -20h
var_18 = qword ptr -18h
var_10 = qword ptr -10h
var_8 = dword ptr -8
var_4 = dword ptr -4
; Открываем кадр стека
push rbp
; Выделяем память в стеке для локальных переменных
sub rsp, 50h
; В регистр RBP заносим указатель на дно стека
lea rbp, [rsp+50h]
```

В регистры `XMM0` и `XMM1` из сегмента данных только для чтения загружаем значения вещественных констант:

```asm
.rodata:0000000000428000 qword_428000 dq 3F2A36E2EB1C432Dh ;DATA XREF: main+A↑r
.rodata:0000000000428008 qword_428008 dq 3F1A36E2EB1C432Dh ;DATA XREF: main+12↑r
```

Для представления вещественных (в нашем случае сверхмалых) значений C++Builder применил сверхбольшие числа.

```asm
movsd xmm0, cs:qword_428000
; А Visual C++ использовал только один регистр SSE
movsd xmm1, cs:qword_428008
; Инициализация локальных переменных
mov [rbp+var_4], 0
mov [rbp+var_8], ecx
mov [rbp+var_10], rdx
; В переменную var_18 помещается вещественное значение, ранее загруженное в регистр XMM1
movsd [rbp+var_18], xmm1
; В var_20 копируется вещественное значение из регистра XMM0
movsd [rbp+var_20], xmm0
; Числа из памяти возвращаются в регистры, меняясь начальными местами дислокации
movsd xmm0, [rbp+var_18]
movsd xmm1, [rbp+var_20]
```

Ага! C++Builder воспользовался немного другой инструкцией сравнения вещественных значений! Команда неупорядоченного сравнения UCOMISD отличается от команды COMISD только тем, что она возвращает ошибку, если операнд-источник равен значению SNaN. Тогда как COMISD выдает исключение, если операнд-источник равен SNaN либо QNaN.

Тем не менее с флагами процессоров она поступает точно так же, как команда упорядоченного сравнения.

`ucomisd xmm1, xmm0`

Инструкцию JBE мы уже проходили, останавливаться на ней не будем. В данном случае, как и в прошлый раз, переход на метку не выполняется.

```asm
jbe short loc_4013EA
lea rcx, xmmword_434B08
; Продолжаем выполнение, загружаем параметры в регистры...
lea rdx, aNum1MenEEmNum2 ; "num1 меньше, чем num2"
; ...и вызываем оператор вывода строки на экран
call std::operator<<<std::char_traits<char>>(std:: basic_ostream<char,std::char_traits<char>> &,char const*)
lea rdx, asc_428026 ; "\\n"
mov rcx, rax
call std::operator<<<std::char_traits<char>>(std:: basic_ostream<char,std::char_traits<char>> &,char const*)
mov [rbp+var_28], rax
; Безусловный переход
jmp short loc_401410
;---------------------------------------------------------------------------
loc_4013EA: ; CODE XREF: main+40↑j
lea rcx, xmmword_434B08
lea rdx, aNum1BolELiboRa ; "num1 больше либо равно num2"
call std::operator<<<std::char_traits<char>>(std::
basic_ostream<char,std::char_traits<char>> &,char const*) lea rdx, asc_428026 ; "\\n"
mov rcx, rax
call std::operator<<<std::char_traits<char>>(std:: basic_ostream<char,std::char_traits<char>> &,char const*)
mov [rbp+var_30], rax
loc_401410: ; CODE XREF: main+68↑j
; Завершаем выполнение приложения
; Загружаем ноль в переменную...
mov [rbp+var_4], 0
; ...затем перекладываем его в регистр
; Обрати внимание, сколько ненужных операций делает C++Builder
; Возвращаем ноль
mov eax, [rbp+var_4]
; Восстанавливаем память стека
add rsp, 50h
; Закрываем кадр стека
pop rbp retn
main endp
```

Было интересно увидеть, что даже такой крохотный примерчик двумя лидирующими компиляторами превращается в разный бинарный код!

## УСЛОВНЫЕ КОМАНДЫ БУЛЕВОЙ УСТАНОВКИ

Начиная с чипа 80386, язык микропроцессоров Intel обогатился командой условной установки бита — SETxx, которая устанавливает свой единственный операнд в единицу (булево TRUE), если в условии xx равенство, и сбрасывает его в ноль (булево FALSE), если условие xx ложно.

Команда SETxx широко используется оптимизирующими компиляторами, чтобы устранить ветвления, то есть избавиться от условных переходов, так как последние очищают конвейер процессора, чем серьезно снижают производительность программы.

| ![Таблица 3: условные команды булевой установки](./images/20/table_3.png)
|:--:|
| *Таблица 3: условные команды булевой установки* |

### Идентификация типов с помощью условий

Условные команды — ключ к идентификации типов. Поскольку результат сравнения знаковых и беззнаковых переменных анализируют разные группы инструкций, можно уверенно и однозначно отличить signed int (знаковые) от unsigned int (беззнаковых).

### Прочие условные команды

Микропроцессоры серии 80x86 поддерживают множество условных команд, которые в общем случае не отображаются на операции отношения, а потому и редко используются компиляторами (можно даже сказать — вообще не используются), но зато часто встречаются в ассемблерных вставках. Словом, они заслуживают хотя бы беглого упоминания.

### команды условного перехода

Помимо описанных в табл. 1, существуют другие условные переходы, среди них: JCXZ, JECXZ, JO, JNO, JP (он же JPE), JNP (он же JPO), JS и JNS. Из них только JCXZ и JECXZ имеют непосредственное отношение к операциям сравнения. Оптимизирующие компиляторы могут заменять конструкцию

```asm
CMP [E]CX, 0
JZ do_it
```

более коротким эквивалентом:

`J[E]CX do_it`

Однако чаще всего они (из-за ограниченности интеллекта и лени своих разработчиков) этого не делают.

Условные переходы JO и JNS используются в основном в математических библиотеках для обработки чисел большой разрядности (например, 1024битных целых).

Условные переходы JS и JNS, помимо основного своего предназначения, часто используются для быстрой проверки значения старшего бита.

Условные переходы JP и JNP вообще практически не используются, ну разве что в экзотичных ассемблерных вставках.

| ![Таблица 4: вспомогательные условные переходы](./images/20/table_4.png)
|:--:|
| *Таблица 4: вспомогательные условные переходы* |

### Команды условной пересылки

Начиная с Pentium, процессоры поддерживают команду условной пересылки CMOVxx, пересылающей значение из источника в приемник, если условие xx истинно. Это позволяет писать намного более эффективный код, не содержащий ветвлений и укладывающийся в меньшее число инструкций.

Рассмотрим конструкцию `IF a<b THEN a=b`. Сравни, как она транслируется с использованием условных переходов (1) и команды условной пересылки (2):

```asm
1)            2)
CMP A,B       CMP A,B
JAE continue: CMOVB A,B
MOV A,B
continue:
```

В табл. 5 дано краткое, но вполне достаточное для дизассемблирования программ описание команд семейства `CMOVxx`. За более подробными разъяснениями обращайся к с. 251 [справочного руководства](https://www.intel.com/content/www/us/en/architecture-and-technology/64-ia-32-architectures-software-developer-vol-2a-manual.html) от Intel.

| ![Таблица 5: основные команды условной пересылки](./images/20/table_5.png)
|:--:|
| *Таблица 5: основные команды условной пересылки* |

### Булевы сравнения

Логической лжи (FALSE) соответствует значение ноль, а логической истине (TRUE) — любое ненулевое значение. Таким образом, булевы отношения сводятся к операции сравнения значения переменной с нулем. Конструкция `IF (a) THEN do_it` транслируется в `IF (a!=0) THEN do_it`.

Практически все компиляторы заменяют инструкцию `CMP A, 0` более короткой командой `TEST A, A` или `OR A,A`. Во всех случаях, если `A==0`, устанавливается флаг нуля и наоборот.

Поэтому, встретив в дизассемблерном тексте конструкцию

```asm
TEST EAX, EAX
JZ do_it
```

можно с уверенностью утверждать, что мы имеем дело с булевым сравнением.

## ИТОГИ

В сегодняшней статье мы научились применять и выявлять ряд ассемблерных команд, постепенно добавляемых в язык сборки на протяжении многих десятилетий — за время непрекращающегося развития микропроцессоров серии x86. Таким образом мы не только рассмотрели состояние микропро- цессора настоящего времени, но и, вернувшись к его истокам, взглянули на его прародителя с сопроцессором и устаревшей системой команд.

Мы научились разбирать сложные условия на отдельные компоненты, тем самым облегчая себе жизнь при анализе программ, изучили различия бинарного кода, генерируемого разными компиляторами на основе одного и того же исходника, увидели, как с помощью условий можно определять типы чисел. Острой темой стали различные механизмы работы с целыми и вещественными значениями.

В следующей статье мы продолжим наш увлекательный заплыв в океане условных операторов, научимся идентифицировать тернарный оператор, рассмотрим массу сложных примеров, также относящихся к сегодняшней статье, и попробуем свои силы в оптимизации ветвлений наперекор компилятору.
