# Изучаем условные переходы в обработке современными компиляторами

Сегодня тебя ждет масса практических примеров — дизассемблированных листингов, на которых мы будем изучать представление условных переходов различными компиляторами. Таким образом, ты узнаешь, как ветвление используется в современных программах и как за ним следить при обратной разработке.

Сегодняшняя статья — продолжение предыдущей, если ее не прочитать, разобраться в коде будет трудновато. Что ж, продолжим наш заплыв в океан условий, ветвлений, отношений и компромиссов.

## ИДЕНТИФИКАЦИЯ ТЕРНАРНОГО ОПЕРАТОРА

Конструкция `a=(условие)?do_it:continue` языка C в общем случае транслируется так: `IF (условие) THEN a=do_it ELSE a=continue`, однако результат компиляции обеих конструкций, вопреки распространенному мнению, не всегда идентичен. Оператор `?:` значительно легче поддается оптимизации, чем ветвление `IF — THEN — ELSE`. Покажем это на следующем примере:

```c
#include <iostream>

int main() {
    int a = -2;
    int b = 2;
    a = (a > 0) ? 1 : -1; // Тернарный оператор

    if (b > 0) // Ветвление
        b = 1;
    else
        b = -1;

    std::cout << "a + b = " << a + b;
}
```

Если пропустить эту программу сквозь компилятор Microsoft Visual C++ 2022 с отключенной оптимизацией, на выходе получим такой код:

```asm
; int cdecl main(int argc, const char **argv, const char **envp) main proc near ; CODE XREF: scrt_common_main_seh+107↓p
; DATA XREF: .pdata:ExceptionDir↓o
; Четыре локальные переменные var_18 = dword ptr -18h

var_14 = dword ptr -14h
var_10 = dword ptr -10h
var_C = dword ptr -0Ch

; Резервируем место для локальных переменных sub rsp, 38h
; Инициализация локальных переменных:
; var_14 присваиваем -2, компилятор выразил отрицательное число
; сверхбольшим положительным
mov [rsp+38h+var_14], 0FFFFFFFEh

; var_18 присваиваем 2, здесь все прямолинейно
mov [rsp+38h+var_18], 2

; Сравниваем переменную var_14 с 0
cmp [rsp+38h+var_14], 0

; Переход по метке, если в результате предыдущего сравнения первый операнд меньше второго или равен ему
jle short loc_140001025 ; В нашем случае все сходится, делаем прыжок
mov [rsp+38h+var_10], 1 jmp short loc_14000102D

; --------------------------------------
loc_140001025: ; CODE XREF: main+19↑j

; Делаем переход сюда
mov [rsp+38h+var_10], 0FFFFFFFFh ; var_10 присваиваем -1
loc_14000102D: ; CODE XREF: main+23↑j mov eax, [rsp+38h+var_10]
mov [rsp+38h+var_14], eax ; var_14 = -1
; Сравниваем var_18 с 0, припоминаем или прокручиваем экран дизассемблера вверх,
; в var_18 помещено значение 2

cmp [rsp+38h+var_18], 0

; Переход по метке, если в результате предыдущего сравнения первый операнд меньше второго или равен ему

jle short loc_140001046 ; В данном случае карты не сходятся, поэтому продолжаем выполнение
mov [rsp+38h+var_18], 1 ; Переменной var_18 присваиваем значение 1
jmp short loc_14000104E ; Безусловный переход

; --------------------------------------

loc_140001046: ; CODE XREF: main+3A↑j mov [rsp+38h+var_18], 0FFFFFFFFh
loc_14000104E: ; CODE XREF: main+44↑j
; Прыгаем сюда
mov eax, [rsp+38h+var_18] ; EAX = 1 mov ecx, [rsp+38h+var_14] ; ECX = -1 add ecx, eax ; -1 + 1 mov eax, ecx
mov [rsp+38h+var_C], eax ; Переменной var_C присваиваем сумму — значение 0
; Готовим параметры
lea rdx, _Val ; "a + b = "
mov rcx, cs:std::basic_ostream<char,std::ch ar_traits<char>> std::cout ; _Ostr
; Вызываем оператор << для вывода строки
call std::operator<<<std::char_traits<char>>(std:: basic_ostream<char,std::char_traits<char>> &,char const *)
mov ecx, [rsp+38h+var_C] mov edx, ecx
mov rcx, rax
; Вызываем оператор << для вывода числа
call cs:std::basic_ostream<char,std::char_traits< char>>::operator<<(int)
xor eax, eax ; Возвращаем 0
add rsp, 38h ; Восстанавливаем стек retn
main endp
```

Во дела! Компилятор оба условных оператора скомпилил в одинаковый код! Теперь переведем компилятор в режим максимальной оптимизации с приоритетом по скорости — `/O2`, произведем компиляцию, откроем результат в дизассемблере и посмотрим на результат:

```asm
main proc near
sub rsp, 28h
mov rcx, cs:std::basic_ostream<char,std::ch ar_traits<char>> std::cout ; _Ostr
call std::operator<<<std::char_traits<char>>(std:: basic_ostream<char,std::char_traits<char>> &,char const *)
mov rcx, rax
xor edx, edx
call cs:std::basic_ostream<char,std::char_traits< char>>::operator<<(int)
xor eax, eax
add rsp, 28h retn
main endp

```

Что за чертовщина?! Здесь не осталось никаких условностей! Даже если включить режим оптимизации с приоритетом размера — /O1, ничего не изменится. Компилятор заоптимизировал все что можно!

А что нам продемонстрирует C++Builder? Сначала без оптимизации:

```asm
; int cdecl main(int argc, const char **argv, const char **envp) public main
main proc near ; DATA XREF: acrtused+29↑o var_20 = qword ptr -20h

var_18 = dword ptr -18h
var_14 = dword ptr -14h
var_10 = qword ptr -10h
var_8 = dword ptr -8
var_4 = dword ptr -4
; Открываем кадр стека
push rbp
; Выделяем память под локальные переменные sub rsp, 40h
; Запоминаем адрес дна стека
lea rbp, [rsp+40h]
mov eax, 0FFFFFFFFh ; EAX = -1
mov r8d, 1 ; R8D = 1
; Инициализация локальных переменных
mov [rbp+var_4], 0
mov [rbp+var_8], ecx mov [rbp+var_10], rdx
mov [rbp+var_14], 0FFFFFFFEh ; var_14 = -2 mov [rbp+var_18], 2 ; var_18 = 2 mov ecx, [rbp+var_14] ; ECX = -2
; Сравниваем ECX с 0
cmp ecx, 0
```

Ага! Используется команда условной пересылки (за подробностями обратись к предыдущей статье): G значит Greater. Другими словами, данные копируются, если первый операнд больше второго (в таком случае флаги ZF и SF равны 0). В нашем случае условие не выполняется, так как -2 < 0, поэтому значение из регистра R8D в регистр EAX не копируется.

```asm
cmovg eax, r8d
mov [rbp+var_14], eax ; var_14 = -1

; Сравнение var_18 с 0
cmp [rbp+var_18], 0
; Переход по метке, если в результате предыдущего сравнения первый операнд меньше второго или равен ему
jle short loc_4013ED ; В данном случае карты не сходятся (2 > 0), поэтому продолжаем выполнение
mov [rbp+var_18], 1 ; var_18 = 1
jmp short loc_4013F4 ; Безусловный прыжок
; --------------------------------------
loc_4013ED: ; CODE XREF: main+42↑j mov [rbp+var_18], 0FFFFFFFFh
loc_4013F4: ; CODE XREF: main+4B↑j
; Безусловный прыжок сюда
; Подготовка параметров для вызова оператора вывода
lea rcx, a008_free_list
lea rdx, aAB ; "a + b = "
; Вывод строки оператором <<
call _ZStlsISt11char_traitsIcEERSt13basic_ostreamIcT_ES5_PKc ; std:: operator<<<std::char_traits<char>>(std::basic_ostream<char,std:: char_traits<char>> &,char const*)
mov r8d, [rbp+var_14] ; R8D = -1
add r8d, [rbp+var_18] ; R8D = -1 + 1
; Готовим параметры для вызова оператора вывода mov rcx, rax
mov edx, r8d
; Вывод числа (суммы) оператором <<
call _ZNSolsEi ; std::ostream::operator<<(int)
; Возвращаем 0
mov [rbp+var_4], 0
mov [rbp+var_20], rax
mov eax, [rbp+var_4]
; Восстанавливаем стек
add rsp, 40h
; Закрываем кадр стека
pop rbp retn
main endp
```

Было интересно увидеть результат работы C++Builder, в особенности исполь- зование команды условной пересылки, которая позволила уменьшить количество условных переходов с двух до одного. За подробностями обра- тись к подготовленному Visual C++ неоптимизированному дизассемблерному листингу, чтобы увидеть оба условных перехода в действии.

Как же выглядит оптимизированный код от C++Builder?

```asm
; int cdecl main(int argc, const char **argv, const char **envp) public main
main proc near ; DATA XREF: acrtused+29↑o push rbp
sub rsp, 20h
lea rbp, [rsp+20h]
lea rcx, xmmword_461078
lea rdx, aAB ; "a + b = "
call std::operator<<<std::char_traits<char>>(std:: basic_ostream<char,std::char_traits<char>> &,char const*)
xor edx, edx
mov rcx, rax
call std::ostream::operator<<(int) xor eax, eax
add rsp, 20h
pop rbp retn
main endp
```

Выглядит он почти так же, как оптимизированный листинг от Visual C++, — все без остатка заоптимизировано! Разбирать тут нечего, все и так понятно. Идем дальше!

Между прочим, с точки зрения языка C/C++ тернарный оператор ?: выгодно отличается от ветвления тем, что может непосредственно использоваться в таких выражениях:

```c
#include <stdio.h>

int main() {
    int a = -1;
    printf("Hello, %s\n", (a > 0) ? "Sailor" : "World!");
}
```

Попробуй настолько же компактно реализовать это с помощью ветвлений! Но на самом деле это удобство лишь внешнее, а компилятор транслирует приведенный выше пример таким образом:

```c
#include <stdio.h>

int main() {
    int a = 1;
    char* p;
    static char s0[] = "Sailor";
    static char s1[] = "World";
    if (a > 0) p = s0; else p = s1;
    printf("Hello, %s\n", p);
}
```

Откомпилируй оба листинга и дизассемблируй полученные файлы. Они должны быть идентичны.

![image](./images/21/sample1.png)

Таким образом, при декомпиляции программ на С/С++ в общем случае невозможно сказать, использовалось ли в них ветвление или тернарный оператор. Однако все же есть зацепки, которые помогают восстановить истинный вид исходного текста в частных случаях.

Например, маловероятно, чтобы программист строил свой листинг, как показано в последнем примере. Зачем вводить статические переменные и сложным способом манипулировать с указателем, когда проще использовать тернарный оператор вместо ветвления?

Таким образом, если тернарный оператор гладко ложится в декомпилируемую программу, а ветвление не лезет в нее никаким боком, то очевидно, что в исходном тексте использовался именно тернарный оператор, а не ветвление. Однако надо быть настороже: мало ли что взбрело в ум программисту, пока он писал код!

## РАЗБОР ДОПОЛНИТЕЛЬНЫХ ПРИМЕРОВ

А теперь, чтобы лучше уяснить материал, рассмотренный в этой и предыдущей статьях, давай разберем несколько живых примеров, откомпилированных различными компиляторами.

### Целочисленные отношения

Начнем с исследования элементарных целочисленных отношений:

```c
#include <stdio.h>
int main()
{
    int a = 1;
    int b = 1;

    if (a < b)
        printf("a < b\n");
    if (a > b)
        printf("a > b\n");
    if (a == b)
        printf("a == b\n");
    if (a != b)
        printf("a != b\n");
    if (a >= b)
        printf("a >= b\n");
    if (a <= b)
        printf("a <= b\n");
}
```

Результат компиляции этого примера компилятором Microsoft Visual C++ 2022 должен выглядеть так:

```asm
; int cdecl main(int argc, const char **argv, const char **envp)
main proc near ; CODE XREF:scrt_common_main_seh+107↓p
; DATA XREF: .pdata:0000000140004018↓o

var_18 = dword ptr -18h
var_14 = dword ptr -14h
; Резервируем память под локальные переменные
sub rsp, 38h
; Инициализируем локальные переменные
mov [rsp+38h+var_14], 1
mov [rsp+38h+var_18], 1
mov eax, [rsp+38h+var_18] ; EAX = var_18
; Сравниваем значение переменной var_14 со значением переменной var_18
cmp [rsp+38h+var_14], eax
```

Если `var_14 >= var_18` (наш случай), то выполняется переход на метку `loc_14000110A`, иначе — печать строки. Обрати внимание, что оригинальный код выглядел так:

```c
if (a < b) printf("a < b\n");
```

То есть компилятор инвертировал условие!
Знаковая операция JGE говорит нам о том, что сравниваемые переменные var_14 и var_18 — знаковые.

```asm
jge short loc_14000110A
lea rcx, _Format ; "a < b\n"
call printf
loc_14000110A: ; CODE XREF: main+1C↑j
; Делаем сюда переход, когда var_14 и var_18 равны (наш случай)
mov eax, [rsp+38h+var_18]
; Снова сравниваем var_14 и var_18
cmp [rsp+38h+var_14], eax
```

Выполняем переход, если `var_14 <= var_18` (наш случай), в противном случае печатаем строку. Следовательно, строка печатается, когда `!(var_14 <= var_18)` или `var_14 > var_18`. Тогда исходный код программы должен выглядеть так:

```C
if (a > b) printf("a > b\n");
```

Рассмотрим следующий фрагмент кода.

```asm
jle short loc_140001120
lea rcx, aAB_0 ; "a > b\n"
call printf
loc_140001120: ; CODE XREF: main+32↑j
; Делаем сюда переход, когда var_14 и var_18 равны (наш случай)
mov eax, [rsp+38h+var_18]
; Снова сравниваем var_14 и var_18
cmp [rsp+38h+var_14], eax
```

Прыгаем, если `var_14 != var_18` (не наш случай). У нас же `var_14 == var_18`, поэтому пропускаем переход и печатаем строку. Оригинальный код программы выглядел следующим образом:

```c
if (a == b) printf("a == b\n");
```

Двигаемся дальше.

```asm
jnz short loc_140001136
lea rcx, aAB_1 ; "a == b\n"
call printf
loc_140001136: ; CODE XREF: main+48↑j
; Плавно продолжаем выполнение программы здесь
mov eax, [rsp+38h+var_18]
; Вновь сравниваем var_14 и var_18
cmp [rsp+38h+var_14], eax
```

Переходим, когда `var_14 == var_18` (наш случай), иначе печатаем строку. Оригинальный код программы выглядел так:

```c
if (a != b) printf("a != b\n");
```

Еще один фрагмент:

```asm
jz short loc_14000114C
lea rcx, aAB_2 ; "a != b\n"
call printf
loc_14000114C: ; CODE XREF: main+5E↑j
; Делаем прыжок сюда
mov eax, [rsp+38h+var_18]
; Сравниваем var_14 и var_18 в очередной раз
cmp [rsp+38h+var_14], eax
```

Переходим, когда `var_14 < var_18` (не наш случай), в противном случае продолжаем выполнение и печатаем строку, что мы и делаем. А оригинальный код имел следующий вид:

```c
if (a >= b) printf("a >= b\n");
```

Теперь рассмотрим такой код:

```asm
jl short loc_140001162
lea rcx, aAB_3 ; "a >= b\n"
call printf
loc_140001162: ; CODE XREF: main+74↑j
; Продолжаем последовательное выполнение в этой части программы
mov eax, [rsp+38h+var_18]
; Сравниваем var_14 и var_18 в последний раз!
cmp [rsp+38h+var_14], eax
```

Переходим на метку, если `var_14 > var_18` (не наш случай), иначе продолжаем выполнение и печатаем строку, что мы и делаем. А оригинальный код имел такой вид:

```c
if (a <= b) printf("a <= b\n");
```

```asm
jg short loc_140001178
lea rcx, aAB_4 ; "a <= b\n"
call printf
loc_140001178: ; CODE XREF: main+8A↑j
; Продолжаем последовательное выполнение в этой части программы
; Возвращаем 0
xor eax, eax
; Восстанавливаем стек
add rsp, 38h
; Закрываем кадр стека
retn
main endp
```

Обрати внимание: во всех случаях компилятор инвертировал условия! Обсуждая это выше, мы отметили, что такие операции он делает в целях оптимизации.

| ![ГРезультат выполнения программы int_compare_vc](./images/21/int_compare_vc_result.png)
|:--:|
| *Результат выполнения программы int_compare_vc* |

### Вещественные отношения Microsoft Visual C++

А теперь сменим тип сравниваемых переменных с int на float и посмотрим, как это повлияет на сгенерированный код. Результат компиляции Microsoft Visual C++ должен выглядеть так:

```asm
main proc near
var_18 = dword ptr -18h
var_14 = dword ptr -14h
sub rsp, 38h
```

Для инициализации регистровых и локальных переменных на этот раз используется инструкция MOVSS, она предназначена для копирования чисел с плавающей запятой одинарной точности.

```asm
movss xmm0, cs: real@3dcccccd
movss [rsp+38h+var_14], xmm0
movss xmm0, cs: real@3dcccccd
movss [rsp+38h+var_18], xmm0
movss xmm0, [rsp+38h+var_18]
```

На этот раз для сравнения чисел также используется другая инструкция — из процессорного расширения SSE2, предназначенная для работы с вещественными числами. Эти команды мы уже проходили: чтобы освежить их в памяти, загляни в таблицу 2 в предыдущей статье.

```asm
comiss xmm0, [rsp+38h+var_14]
```

Зато за принятие решения о переходе на метку отвечают старые добрые команды условного перехода, которые опираются на биты регистра флагов основного процессора. Надо понимать, что в современных многоядерных процессорах каждое ядро содержит такой механизм.

Инструкция `JBE` выполняет переход на метку `loc_140001119`, когда первый операнд предыдущего сравнения меньше второго или равен ему (наш случай), иначе строка выводится на консоль. Заметь, в данном случае условие не инвертировано, но к нему добавлено условие равенства.

```asm
jbe short loc_140001119
lea rcx, _Format ; "a < b\n"
call printf
loc_140001119: ; CODE XREF: main+2B↑j
movss xmm0, [rsp+38h+var_14]
comiss xmm0, [rsp+38h+var_18]
; Здесь все то же самое, что в предыдущем блоке кода
jbe short loc_140001132
lea rcx, aAB_0 ; "a > b\n"
call printf
loc_140001132: ; CODE XREF: main+44↑j
movss xmm0, [rsp+38h+var_14]
```

Следующей инструкцией вновь сравниваются переменные `var_14` и `var_18`, но на этот раз командой неупорядоченного сравнения. Дополнительные сведения смотри в разделе «Сравнение вещественных чисел» прошлой статьи.

```asm
ucomiss xmm0, [rsp+38h+var_18]
; Переходит на метку, если возведен флаг четности PF
jp short loc_14000114D
; Переходит на метку, если сброшен флаг нуля
; Обрати внимание: переходы ведут на одну и ту же метку
jnz short loc_14000114D
; В нашем случае оба условия не выполняются, и мы выводим строку на экран
lea rcx, aAB_1 ; "a == b\n"
call printf
loc_14000114D: ; CODE XREF: main+5D↑j
; main+5F↑j
movss xmm0, [rsp+38h+var_14]
ucomiss xmm0, [rsp+38h+var_18]
; В рассматриваемом случае при предыдущем сравнении чисел флаг четности не устанавливается
jp short loc_14000115C
; Зато устанавливается флаг нуля ZF = 1, поэтому переходим на эту метку
jz short loc_140001168
loc_14000115C: ; CODE XREF: main+78↑j
lea rcx, aAB_2 ; "a != b\n"
call printf
loc_140001168: ; CODE XREF: main+7A↑j
movss xmm0, [rsp+38h+var_14]
comiss xmm0, [rsp+38h+var_18]
; Сравниваем var_14 и var_18, переход не делаем, так как наши числа равны,
; а он происходит в случае, когда первый операнд меньше второго
jb short loc_140001181
lea rcx, aAB_3 ; "a >= b\n" call printf
loc_140001181: ; CODE XREF: main+93↑j movss xmm0, [rsp+38h+var_18]
comiss xmm0, [rsp+38h+var_14]
; Все то же самое, поэтому пропускаем переход и выводим строку
jb short loc_14000119A
lea rcx, aAB_4 ; "a <= b\n"
call printf
loc_14000119A: ; CODE XREF: main+AC↑j
; Возвращаем ноль
xor eax, eax
add rsp, 38h retn
main endp
```

Изменений в сравнении с программой целочисленных чисел не сказать что много, но они весьма существенны. Главным образом это использование других инструкций, предназначенных для вещественных чисел.

### Embarcadero C++Builder

Бесспорно, рассмотренный пример интересен, но гораздо нагляднее будет изучить код, сгенерированный компилятором C++Builder:

```asm
; int cdecl main(int argc, const char **argv, const char **envp)
public main
main proc near ; DATA XREF: acrtused+29↑o
; Объявлена куча бесполезных переменных
var_30 = dword ptr -30h
var_2C = dword ptr -2Ch
var_28 = dword ptr -28h
var_24 = dword ptr -24h
var_20 = dword ptr -20h
var_1C = dword ptr -1Ch
var_18 = dword ptr -18h
var_14 = dword ptr -14h
var_10 = qword ptr -10h
var_8 = dword ptr -8
var_4 = dword ptr -4
; Манипулирование с кадром стека
push rbp
sub rsp, 50h
lea rbp, [rsp+50h]
; Инициализация переменных, из выполняющих это операций можно однозначно определить тип переменных
movss xmm0, cs:dword_44D000
mov [rbp+var_4], 0
mov [rbp+var_8], ecx
mov [rbp+var_10], rdx
movss [rbp+var_14], xmm0
movss [rbp+var_18], xmm0
movss xmm0, [rbp+var_14]
movss xmm1, [rbp+var_18]
; C++Builder для сравнения вещественных чисел использует команду неупорядоченного сравнения...
ucomiss xmm1, xmm0
; ...на основе чего производит переход — это нам уже знакомо
jbe short loc_401448
lea rcx, aAB ; "a < b\n"
call printf
mov [rbp+var_1C], eax
loc_401448: ; CODE XREF: main+37↑j
movss xmm0, [rbp+var_14]
ucomiss xmm0, [rbp+var_18]
jbe short loc_401462
lea rcx, aAB_0 ; "a > b\n"
call printf
mov [rbp+var_20], eax
loc_401462: ; CODE XREF: main+51↑j
movss xmm0, [rbp+var_14]
ucomiss xmm0, [rbp+var_18]
; Оба нижеследующих перехода ведут к одной метке
jnz short loc_40147E
jp short loc_40147E
lea rcx, aAB_1 ; "a == b\n"
call printf
mov [rbp+var_24], eax
loc_40147E: ; CODE XREF: main+6B↑j
; main+6D↑j movss xmm0, [rbp+var_14]
ucomiss xmm0, [rbp+var_18]
; В этом блоке кода имеем три перехода, два из которых ведут к одной метке
jnz short loc_40148D
jp short loc_40148D
jmp short loc_40149C
; --------------------------------------
loc_40148D: ; CODE XREF: main+87↑j
; main+89↑j
lea rcx, aAB_2 ; "a != b\n" call printf
mov [rbp+var_28], eax
loc_40149C: ; CODE XREF: main+8B↑j
movss xmm0, [rbp+var_14]
ucomiss xmm0, [rbp+var_18]
jb short loc_4014B6
lea rcx, aAB_3 ; "a >= b\n"
call printf
mov [rbp+var_2C], eax
loc_4014B6: ; CODE XREF: main+A5↑j
movss xmm0, [rbp+var_14]
movss xmm1, [rbp+var_18]
ucomiss xmm1, xmm0
jb short loc_4014D4
lea rcx, aAB_4 ; "a <= b\n"
call printf
mov [rbp+var_30], eax
loc_4014D4: ; CODE XREF: main+C3↑j
mov [rbp+var_4], 0
mov eax, [rbp+var_4]
add rsp, 50h
pop rbp retn
main endp
```

Как можно видеть, подготовленный C++Builder код не сильно отличается от такого же, изготовленного Visual C++. Из различий заметен избыток локальных переменных и явный излишек ненужных переходов.

> **INFO**
>
> Нельзя сказать наверняка, какой компилятор генерирует лучший код: Visual C++ или C++Builder. В каких-то примерах C++Builder радует, используя команды условной пересылки вместо условных переходов, а в других — огор- чает, вставляя кучу ненужных локальных перемен- ных и переходов.

### Сложные примеры: использование деревьев для разбора вложенных условий

Теперь, наловчившись на идентификации элементарных условий, перейдем к вещам по-настоящему сложным.

Рассмотрим следующий пример:

```c
#include <stdio.h>
int main()
{
    unsigned int a = 0;
    unsigned int b = 2;

    int c = 1;
    int d = -1;

    if (d)
        printf("TRUE\n");
    else if (((a > b) && (a != 0)) || ((a == c) && (c != 0)))
        printf("OK\n");

    if (c == d)
        printf("+++\n");
}
```

Результат его компиляции должен выглядеть приблизительно так (для большей наглядности в дизассемблерном листинге я изменил имена переменных с ничего не говорящих `var_10`, `var_14` и так далее на прямое соответствие именам переменных из исходного C-кода: `var_c`, `var_b`):

```asm
; int cdecl main(int argc, const char **argv, const char **envp)
main proc near ; CODE XREF:scrt_common_main_seh+107↓p
; DATA XREF: .pdata:0000000140004018↓o
; Четыре локальные переменные, имеющие зеркальное отражение имен переменных из листинга программы на языке высокого уровня
var_a = dword ptr -18h
var_c = dword ptr -14h
var_d = dword ptr -10h
var_b = dword ptr -0Ch
; Резервирование места для локальных переменных
sub rsp, 38h
; Инициализация локальных переменных
mov [rsp+38h+var_a], 0
mov [rsp+38h+var_b], 2
mov [rsp+38h+var_c], 1
mov [rsp+38h+var_d], 0FFFFFFFFh ; -1
; Сравнение значения переменной var_d с нулем
cmp [rsp+38h+var_d], 0
; Если переменная var_d равна нулю, переход к метке loc_140001119, иначе печать строки TRUE
jz short loc_140001119
```

Схематически это можно изобразить так.

![img](./images/21/sch01.png)

```asm
lea rcx, _Format ; "TRUE\n"
call printf
; Вносим этот безусловный переход в наше дерево
jmp short loc_140001147
```

В итоге получается следующее.

![img](./images/21/sch02.png)

```asm
; --------------------------------------
loc_140001119: ; CODE XREF: main+29↑j
; Загружаем в EAX значение переменной var_b
mov eax, [rsp+38h+var_b]
; Сравниваем переменную var_a с переменной var_b
cmp [rsp+38h+var_a], eax
; Если var_a меньше переменной var_b или равна ей, то выполняется переход на метку loc_14000112A
jbe short loc_14000112A
```

Прививаем новое гнездо к нашему дереву, попутно обращая внимание на то, что `var_a` и `var_b` — беззнаковые переменные!

![img](./images/21/sch03.png)

```asm
; Сравниваем значение переменной var_a с нулем
cmp [rsp+38h+var_a], 0
; Переход на метку loc_14000113B, если var_a не равна нулю
jnz short loc_14000113B
```

![img](./images/21/sch04.png)

```asm
loc_14000112A: ; CODE XREF: main+41↑j
```

Смотри: в нашем дереве уже есть метка `loc_14000112A`. Корректируем ее.

![img](./images/21/sch05.png)

```asm
; Загружаем в EAX значение переменной var_c
mov eax, [rsp+38h+var_c]
; Сравниваем значения переменных var_a и var_c
cmp [rsp+38h+var_a], eax
; Переход на метку loc_140001147, если var_a != var_c
jnz short loc_140001147
```

![img](./images/21/sch06.png)

```asm
; Сравнение значения переменной var_c с нулем
cmp [rsp+38h+var_c], 0
; Переход на loc_140001147, если var_c == 0
jz short loc_140001147
```

![img](./images/21/sch07.png)

```asm
loc_14000113B: ; CODE XREF: main+48↑j
; Смотрим: метка loc_14000113B уже есть в дереве, прививаем
lea rcx, aOk ; "OK\n"
call printf
```

![img](./images/21/sch08.png)

```asm
loc_140001147: ; CODE XREF: main+37↑j
; main+52↑j ...
; Ветки loc_140001147 и loc_14000113B смыкаются
```

![img](./images/21/sch09.png)

```asm
; Загружаем в EAX значение переменной var_d
mov eax, [rsp+38h+var_d]
; Сравниваем значение переменной var_c со значением var_d
cmp [rsp+38h+var_c], eax
; Переход, если var_c != var_d
jnz short loc_14000115D
lea rcx, asc_14000225C ; "+++\n"
call printf
```

![img](./images/21/sch10.png)

```asm
loc_14000115D: ; CODE XREF: main+6F↑j
xor eax, eax
add rsp, 38h retn
main endp
```

В итоге вырастает огромное разлапистое дерево, в котором, на первый взгляд, просто невозможно разобраться. Но, как говорится, глаза боятся, а руки делают. Первым делом оптимизируем дерево: избавимся от «перекрученных» ветвей, инвертировав условие в гнезде, и выкинем все метки. Теперь, когда скелет дерева построен, они уже не нужны. Если все сделать правильно, дерево должно выглядеть как на следующем рисунке.

| ![Логическое дерево](./images/21/logic_tree.png)
|:--:|
| *Логическое дерево* |

Сразу же бросается в глаза, что все пути проходят точку Z, сплетающую все ветви воедино. Это значит, что мы имеем дело с двумя самостоятельными деревьями, представленными собственными конструкциями IF. Замечательно! Такой поворот событий весьма упрощает анализ — раз деревья независимые, то и анализироваться они могут независимо! Итак, начинаем с верхнего из них...

От гнезда `var_d != 0` отходят две ветки — правая ведет к `printf("OK")` и далее к завершению конструкции `IF — THEN [ELSE]`, а левая, прежде чем выйти к точке Z, минует целое полчище гнезд. В переводе на русский язык ситуация выглядит так: если переменная `var_d` не равна нулю, то печатаем OK и сваливаем, иначе выполняем дополнительные проверки. Проще говоря: `IF (var_d !=0) THEN printf("OK") ELSE ...` Короче, левая ветка гнезда `(var_d != 0)` есть ветка `ELSE`. Изучим ее.

От гнезда `(var_a <= var_b)` к узлу `printf("OK")` ведут два пути:`!(var_a < = var_b) → !(var_a == 0) и !(var_a != var_c) → !(var_c == 0)`. Где есть альтернатива — там всегда есть `OR`. То есть либо первый путь, либо второй.

В то же время узлы обоих путей последовательно связаны друг с другом, значит, они объединены операцией `AND`. Получается, эта ветка должна выглядеть так: `IF ((var_a > var_b) && (var_0 != 0)) || (var_a == var_c) && (var_c != 0)) printf("OK")`. Прививаем `ELSE` к первому `IF` и получаем: `IF (var_d !=0) THEN printf("OK") ELSE IF (( var_a > var_b) && (var_0 != 0)) || (var_a == var_c) && (var_c != 0)) printf("OK")`.

Ну а разбор второго дерева вообще тривиален: `IF (var_c == var_d) printf("+++")`. Итак, собранные сведения указывают, что исходный текст дизассемблируемой программы выглядел так:

```c
u_int a;
u_int b;
?_int c;
?_int d;

if (d) printf("TRUE"); else
if (((a > b) && (a != 0)) || ((a == c) && (c != 0))) printf("OK\n");
if (c == d) printf("+++\n");
```

Тип переменных `а` и `b` мы определили как `unsigned int`, поскольку результат сравнения анализировался беззнаковой условной командой — `jnb`. А вот тип переменных `c` и `d`, увы, определить так и не удалось. Однако это не убавляет значимости того факта, что мы смогли ретранслировать сложное условие, в котором без деревьев было бы немудрено и запутаться...

## ОПТИМИЗАЦИЯ ВЕТВЛЕНИЙ

Какое коварство: под флагом оптимизации превратить каждую строчку кода в головоломку. Итак, предположим, встретился тебе код следующего содержания. На всякий случай, чтобы избавить тебя от копания в справочниках (хотя покопаться в них лишний раз только на пользу), отмечу, что команда `SETGE` устанавливает выходной операнд в 1, если флаги состояния `SF` и `OF` равны (то есть `SF==OF`). Иначе выходной операнд устанавливается в ноль.

```asm
mov eax, [var_A]
xor ecx, ecx
cmp eax, 0x666
setge cl
dec ecx
and ecx, 0xFFFFFC00
add ecx, 0x300
mov [var_zzz],ecx
```

На первый взгляд кажется, что этот фрагмент заимствован из какого-то хитрого и запутанного защитного механизма. Но нет, перед нами результат компиляции тривиального выражения

```c
if (a<0x666) zzz=0x200 else zzz=0x300
```

которое в неоптимизированном виде выглядит так:

```asm
mov eax,[var_A]
cmp eax, 0x666
jge Label_1
mov ecx, 0x100
jmp Label_2
Label_1:
mov ecx, 0x300
Label_2:
mov [var_zzz], ecx
```

Чем же компилятору не понравился такой вариант? Между прочим, он даже короче! Короче-то он короче, но содержит ветвления, то есть внеплановые изменения нормального хода выполнения программы. А ветвления отрицательно сказываются на производительности хотя бы уже потому, что они приводят к очистке конвейера. Конвейер же в современных процессорах очень длинный, и быстро его не заполнишь... Поэтому избавление от ветвлений путем хитроумных математических вычислений вполне оправданно и горячо приветствуется. Попутно это усложняет анализ программы, защищая ее от всяких подозрительных личностей типа хакеров.

Впрочем, если хорошенько подумать... Начнем пошагово исполнять программу, мысленно комментируя каждую строчку:

```asm
mov eax, [var_A] ; eax == var_A
xor ecx, ecx ; ecx = 0;
cmp eax, 0x666 ; if (eax < 0x666) { SF=1; OF=0; } else { SF=0; OF=0;}
setge cl ; if (eax < 0x666) (т.e. SF==1, OF==0) cl=0; else cl=1;
dec ecx ; if (eax < 0x666) ecx = -1; else ecx = 0;
and ecx, 0xFFFFFC00 ; if (eax < 0x666) (т.e. ecx == -1) ecx = 0xFFFFFC00 (-0x400); else ecx = 0;
add ecx, 0x300 ; if (eax < 0x666) (т.e. ecx = -0x400) ecx = 0x100; else ecx = 0x300;
mov [esp+0x66],ecx
```

Получилось! Мы разобрались с этим алгоритмом и успешно отреверсили его! Теперь видно, что это довольно простой пример (в жизни будут попадаться и более сложные). Но основная идея ясна: если встречается команда `SETxx`, держи нос по ветру — пахнет условными переходами! В вырожденных случаях `SETxx` может быть заменена на `SBB` (вычитание с займом). По этому поводу решим вторую задачу.

```asm
SUB EBX, EAX
SBB ECX, ECX
AND ECX, EBX
ADD EAX, ECX
```

Что этот код делает? Какие-то сложные арифметические действия? Посмотрим...

```asm
SUB EBX, EAX ; if (EBX < EAX) SF = 1; else SF = 0;
SBB ECX, ECX ; if (EBX < EAX) ECX = -1; else ECX=0;
AND ECX, EBX ; if (EBX < EAX) ECX = EBX; else ECX=0;
ADD EAX, ECX ; if (EBX < EAX) EAX = EAX + (EBX - EAX); else EAX = EAX
```

Раскрывая скобки в последнем выражении (мы ведь не забыли, что от `EBX` отняли `EAX`?), получаем: `if (EBX < EAX) EAX=EBX`, то есть это классический алгоритм поиска минимума среди двух знаковых чисел. А вот еще один пример:

```asm
CMP EAX, 1
SBB EAX, EAX
AND ECX, EAX
XOR EAX, -1
AND EAX, EBX
OR EAX, ECX
```

Попробуй решить его самостоятельно и только потом загляни в ответ:

```asm
CMP EAX, 1 ; if (EAX != 0) SF = 0; else SF = 1;
SBB EAX, EAX ; if (EAX != 0) EAX = -1; else EAX = 0;
AND ECX, EAX ; if (EAX != 0) ECX = ECX; else ECX = 0;
XOR EAX, -1 ; if (EAX != 0) EAX = 0; else EAX = -1;
AND EAX, EBX ; if (EAX != 0) EAX = 0; else EAX = EBX;
OR EAX, ECX ; if (EAX != 0) EAX = ECX; else EAX = EBX;
```

Да... после таких упражнений товарищ Буль будет тебе во сне сниться! Но таковы уж издержки цивилизации. К слову сказать, подавляющее большинство компиляторов достаточно лояльно относится к условным переходам и не стремится тотально их изгнать. Так что особо напрягаться при анализе оптимизированного кода не приходится (правда, к ручной оптимизации это не относится — профессиональные разработчики выкидывают переходы в первую очередь).

## ИТОГИ

Сегодня мы еще дальше заплыли в море условных операторов и узнали о них много нового. Главным образом мы закрепили уже пройденный материал на специально подготовленных примерах. Рассмотрев тернарный оператор и сравнив его с обычным условным оператором, мы увидели, что компилятору по большому счету неважно, какой тип условия используется в коде на языке высокого уровня. В любом случае он пытается сделать максимально оптимизированный бинарник, порой прибегая к инвертированию условий или полной их замене даже с отключенной агрессивной оптимизацией.

Центральной темой статьи стало использование деревьев для понимания сути вложенных условий, их многочисленных ходов, выполнение которых зависит от поступающих данных и их анализа. Поначалу может показаться, что использование деревьев только усложняет задачу, но чем запутаннее и навороченнее условия будут тебе встречаться, тем большую благодарность этому методу ты будешь испытывать!

Последней рассмотренной темой стала оптимизация ветвлений, где приведено несколько примеров использования команд условной булевой установки. Более подробно с теоретической точки зрения они были описаны в разделе «У словные команды булевой установки» прошлой статьи.

В следующей публикации мы закончим тему условных операторов, для этого нам останется рассмотреть, каким образом компиляторы интерпретируют switch — case — break, как распознать в дизассемблированном листинге эту конструкцию и какую выгоду могут получить наши программы от ее использования.
