# ЗАТРУДНЯЕМ АНАЛИЗ ПРОГРАММ

Под хакерством обычно понимают изучение дизассемблированного кода приложений с целью их взлома. Сегодня мы займем противоположную сторону баррикад — изучим приемы защиты. Настала пора основательно разобраться в устройстве используемых хакерами инструментов, функциональной поддержке со стороны операционной системы и процессора, чтобы противостоять взломщику, скрыть свои секреты и не стать легкой наживой для злоумышленника!

Три основных этапа взлома защитных механизмов — это локализация кода защиты в коде приложения, анализ алгоритма его работы и собственно сам взлом. Все этапы одинаково важны: если, например, не будет пройден второй из них — за взлом нечего браться.

Можно классифицировать защиты по типу «этапа преткновения». Например, шифры и криптозащиты опираются на третий этап — алгоритм их работы обычно общедоступен, хорошо документирован и в общем случае известен хакеру, но это не сильно облегчает взлом (разве что упрощает написание переборщика). Механизмы с использованием регистрационных номеров, напротив, делают упор на засекречивании алгоритма генерации, а также затруднении его поиска и анализа в коде программы (еще бы, зная алгоритм, можно легко написать кейген).

Однако даже если защита построена с применением криптографических методов, скажем шифрует тело критически важных функций криптостойким методом по непомерно длинному ключу, она может быть «отвязана» от ключа, например копированием дампа программы после расшифровки. Еще проще — распространять программу вместе с ключом (обычная тактика пиратов). Один из способов воспрепятствовать такому беспределу — заложить в ключ зашифрованную привязку к компьютеру или проверять «чистоту» копии через интернет (можно даже и втихомолку — скрыто от пользователя, хотя это считается дурным тоном). Но что помешает хакеру, владеющему лицензионной копией программы, расшифровать ее своим ключом и выкусить все проверки чего бы там ни было?

Таким образом, любая защита должна уметь эффективно препятствовать своему обнаружению, анализу, попутно отравляя жизнь дизассемблеру и отладчику — основным инструментам взломщика. Без этого защита не защита.

В эпоху царствования MS-DOS в основном использовались программы реального режима, которые монопольно распоряжаются процессором, памятью и аппаратурой, в любой момент беспрепятственно переходят в защищенный режим и возвращаются обратно. Отладчики в то время (еще хлипкие, немощные, нежизнеспособные) легко обманывались, срубались и завешивались тривиальными приемами программирования, активно используемыми защитами. Другими словами, дизассемблеры тогда были очень глупыми и впадали в ступор от одного только вида зашифрованного или самомодифицирующегося кода. Словом, настоящий рай для разработчиков защит.

Сегодня все изменилось. Прежде всего, прикладной программе под Windows особо выделываться никто не даст. С защищенным режимом теперь слишком не разгонишься — используй прозаические непривилегированные инструкции, а о разных хитростях не помышляй. Та же небольшая часть защитных приемов, что может функционировать даже в такой ориентированной на пользователя среде, наталкивается на сильно поумневшие отладчики и дизассемблеры.

Аппаратная поддержка отладки в процессорах 386+ в совокупности с виртуальным режимом работы, привилегированными инструкциями и виртуальной памятью позволяет создавать отладчики, которые практически не могут быть обнаружены прикладной программой, и уж тем более она не в состоянии получить над ними контроль.

Существуют и отладчики-эмуляторы, фактически настоящие виртуальные машины, самостоятельно исполняющие код, вместо того чтобы пустить его на «живой» процессор. При этом эмулятор всегда исполняется в режиме супервизора даже по отношению к отлаживаемому коду нулевого кольца. У защиты очень мало шансов обнаружить отладчик или помешать его работе (да и то если эмулятор реализован с ошибками).

Появились и интерактивные дизассемблеры (та же IDA), которые в силу тесного взаимодействия с пользователем (в смысле хакером) могут обходить любые мыслимые и немыслимые ловушки, оставленные разработчиком.

Даже на уровне нулевого кольца в Windows очень трудно что-либо скрыть. Чтобы обеспечить совместимость со всем парком Windows-подобных операционных систем, приходится использовать только документированные возможности. Строить в «окнах» защиту — все равно что пытаться заблудиться в парке. Будь там хоть миллион деревьев, все они геометрически правильно расположены и обильно увешаны табличками «Выход — там».

Таким образом, надежно противостоять изучению программы очень трудно, если вообще возможно. Однако многие приемы против отладчиков и дизассемблеров просто интересны сами по себе и достойны того, чтобы рассмотреть их в этой статье.

## ПРИЕМЫ ПРОТИВ ОТЛАДЧИКОВ

### Немного истории

Раньше всех появился debug.com — пародия, которая отдаленно напоминает отладчик, зато входит в штатную поставку MS-DOS. Сегодня этот инструмент годится разве что для забавы и изучения ассемблера. Впрочем, и тогда от него мало кто был в восторге, и новые отладчики росли как грибы после дождя. Правда, большинство из них недалеко ушло от своего прототипа, отличаясь от оригинала разве что интерфейсом.

Это было золотое время для разработчиков защит. Стоило лишь «запереть» клавиатуру, запретить прерывания, сбросить флаг трассировки, и отладка программы становилась невозможной.

Первые мало-мальски пригодные для взлома отладчики появились только после оснащения компьютеров процессором 80286. В памяти хакеров навсегда останутся AFD PRO, написанный в 1987 году AdTec GmbH, знаменитый Turbo Debugger, созданный годом позже братьями Крисом и Ричем Вильямсами, первый эмулирующий отладчик Сергея Пачковски, написанный, правда, с большим опозданием — в 1991 году. Разработчики защит крякнули, но выдержали — эти отладчики по-прежнему позволяли отлаживаемой программе захватить над собой контроль и очень плохо переносили «извращения» со стеком, экраном, клавиатурой...

Ситуация изменилась с выходом процессора 80386 — резкое усложнение программного обеспечения и, как следствие, огромные сложности с его отладкой диктовали необходимость наличия развитых отладочных средств в самом процессоре. И в 386 они появились! С этого момента разработчикам защит стали наступать на пятки.

Масла в огонь подлила NuMega, выпустившая в конце восьмидесятых годов свой замечательный SoftICE, пользовавшийся у хакеров огромной популярностью, а позже портированный на Windows 9x и Windows NT. Он долгое время оставался бесспорным фаворитом (хотя не без конкуренции). Впрочем, неверно было бы считать, что NuMega — криминальная фирма, а SoftICE исключительно хакерский продукт. Этот отладчик предназначен в первую очередь для разработчиков драйверов и для легальных исследователей операционной системы (не разбираясь во внутренностях ОС, с драйверами особо не разгонишься).

Но так или иначе, SoftICE задал копоти всем защитам и их разработчикам. Пускай он не был полностью невидимым для отлаживаемых программ Stealth-отладчиком, имел ряд ошибок, позволяющих себя обнаружить, и давал защите возможность вырваться из-под контроля, но в умелых руках отладчик справлялся со всеми этими ограничениями и обходил заботливо расставленные «капканы». И с каждой версией SoftICE противостоять ему становилось все труднее и труднее (старые ошибки устранялись быстрее, чем вносились новые).

Постепенно мода на антиотладочные приемы сошла на нет и уж совсем заглохла под победное шествие Windows XP SP3. Поскольку одновременно с ее выходом SoftICE перестал корректно работать, полностью завешивая систему без возможности дальнейшего функционирования. Все указывает на то, что такой же эффект был в Windows Vista, вышедшей годом ранее, но на ней я самолично SoftICE не проверял.

К тому моменту NuMega уже была приобретена компанией Compuware, и SoftICE распространялся как часть пакета для разработки драйверов DriverStudio. Последняя версия SoftICE была выпущена в апреле 2006 года. Service Pack 3 для Windows XP появился двумя годами позднее.

Так как SoftICE использовал недокументированные возможности Windows, с выходом новой версии винды он переставал работать. Как было сказано ранее, после 2006 года SoftICE перестал получать обновления, затем был продан компании Micro Focus и ею же окончательно похоронен. Вероятно, главную роль в смерти SoftICE сыграли деловые разногласия создателей отладчика с Microsoft, а не технические проблемы, как принято считать. Похоже, SoftICE использовал какие-то особые механизмы операционной системы, которые Microsoft посчитала «неудобными».

При срабатывании поставленной точки останова SoftICE завешивал систему, позволяя пользователю (в данном случае хакеру) продолжать взаимодействовать с компьютером через свою прослойку между аппаратным обеспечением и ОС. Поэтому после завешивания системы взломщик мог спокойно возобновить ее работу, а она, в свою очередь, о зависании даже не подозревала. У нее даже отставало системное время на тот период, пока она находилась в «отключке» — в зависшем состоянии.

Свято место пусто не бывает. В качестве отладчика пользовательского уровня популярным стал OllyDbg. Однако сейчас по данным с официального сайта разработка отладчика заморожена.

Предприимчивые китайцы выпустили отладчик уровня ядра под названием Syser, по своим возможностям во многом схожий с испустившим дух SoftICE. Только энтузиазма хватило ненадолго, и уже совсем скоро разработка была прекращена. Сейчас его не найти даже на файловых помойках.

Зато Microsoft с течением времени превратила свой захудалый отладчик WinDbg в действительно мощный и полезный для системных программистов и хакеров инструмент. Грубо говоря, WinDbg представляет собой оболочку для отладки с помощью движка **dbgeng.dll**, который включен непосредственно в операционную систему. WinDbg может использоваться как отладчик либо пользовательского режима, либо режима ядра, но не одновременно.

Вместе с гибелью SoftICE распространилось совершенно нелепое убеждение, что под Windows на прикладном уровне дернуть хвост человеку с отладчиком невозможно. Это вызывает ухмылку профессионалов, эпизодически встраивающих разные ловушки в свои программы — так, больше для разминки (дабы мозги жиром не заплыли), чем для серьезной борьбы с хакерами.

Бороться с хакерами при современном уровне средств анализа приложений несколько наивно — те и от тигра хвост оторвут. Но сегодня, кроме хакеров, серьезную угрозу представляют и вчерашние желторотые пользователи, начитавшиеся справочных руководств о том, «как ломать программы» (благо сейчас они доступны всем и каждому), и теперь только и ищущие, на чем бы испытать свою силушку богатырскую.

### Как работает отладчик

Бороться с отладчиком, не представляя себе, как он работает, было бы по меньшей мере некультурно. Поэтому ниже мы рассмотрим базовые принципы, лежащие в его основе. Это изложение не является всеобъемлющим, тем не менее оно позволяет составить общее представление о вопросе. Технические подробности исчерпывающе изложены в 17-й главе Debug, Branch Prole, Tsc, And Resource Monitoring Features технического руководства Intel 64 and IA-32 Architectures Software Developer’s Manual, Volume 3B: System Programming Guide, Part 2 ([PDF](https://www.intel.de/content/dam/www/public/us/en/documents/manuals/64-ia-32-architectures-software-developer-vol-3b-part-2-manual.pdf)), которое бесплатно распространяет фирма Intel.

Все отладчики можно разделить на две категории: первые используют отладочные средства процессора, а вторые самостоятельно эмулируют процессор, полностью контролируя выполнение «подопытной» программы.

Качественный эмулирующий отладчик отлаживаемый код ни обнаружить, ни обойти не может, но полноценных эмуляторов современных Core-процессоров на сегодняшний день нет, и вряд ли они появятся в обозримом будущем.

Да и есть ли смысл их создавать? Микропроцессоры линейки Core предоставляют в распоряжение разработчика богатейшие отладочные возможности, позволяющие контролировать даже привилегированный код! Они поддерживают пошаговое исполнение программы, отслеживают выполнение инструкции по заданному адресу, контролируют обращения к нужным ячейкам памяти (или портам ввода-вывода), сигнализируют о переключениях задач и так далее.

В общей сложности микропроцессоры серии x86, начиная с модели x386, содержат восемь отладочных 32-битных регистров. В обновленной архитектуре x86-64 количество отладочных регистров не изменилось, однако они увеличились в размере до 64 бит.

Четыре отладочных регистра **DR0 — DR3** хранят линейные адреса четырех контрольных точек, а управляющий регистр **DR7** содержит для каждой из них условие, при выполнении которого процессор генерирует исключение **INT 0x1**, передавая управление отладчику. Всего существует четыре различных условия: прерывание при выполнении команды, прерывание при модификации ячейки памяти, прерывание при чтении или модификации, но не исполнении ячейки памяти и прерывание при обращении к порту ввода-вывода.

Отладочные регистры для линейных адресов точек останова в зависимости от архитектуры процессора имеют размеры 32 или 64 бита.

Регистры **DR4** и **DR5** зарезервированы для использования в будущем. Флаги регистра **DR6** устанавливаются в зависимости от происходящих исключений. Например, первые четыре флага устанавливаются в зависимости от срабатывания соответствующих точек останова в регистрах **DR0 — DR3**. Флаги регистра **DR7** позволяют включать или отключать установленные в регистрах **DR0 — DR3** точки останова, а также определять условия срабатывания прерываний.

| ![Регистры отладки в процессорах на архитектуре IA-32](./images/26/ai_32_debug_reg.png)
|:--:|
| *Регистры отладки в процессорах на архитектуре IA-32* |

| ![Регистры отладки в процессорах на архитектуре AMD64](./images/26/amd64_debug_reg.png)
|:--:|
| *Регистры отладки в процессорах на архитектуре AMD64* |

Если бит трассировки регистра флагов установлен, то после выполнения каждой машинной инструкции автоматически генерируется отладочное исключение **INT 1** и управление передается отладчику. Отлаживаемый код может обнаружить трассировку анализом регистра флагов, поэтому для обеспечения собственной невидимости отладчик должен распознавать команды чтения регистра флагов и эмулировать их выполнение, возвращая нулевое значение флага трассировки.

Следует обратить внимание на одно важное обстоятельство: после выполнения команды, модифицирующей значение регистра SS, отладочное исключение не генерируется! Отладчик должен уметь распознавать такую ситуацию и самостоятельно устанавливать точку останова на следующую инструкцию. В противном случае войти в процедуру, предваренную инструкцией POP SS, автоматический трассировщик не сможет:

```asm
PUSH SS
POP SS
CALL MySecretProc
```

Не все современные отладчики учитывают эту тонкость, и такой прием, несмотря на свою архаичность, может оказаться далеко не бесполезным.

Установкой специального бита можно добиться генерации отладочного исключения при всяком обращении к отладочным регистрам, которое возникает даже в том случае, если их пытается прочесть (модифицировать) привилегированный код. Грамотно спроектированный отладчик может скрыть факт своего присутствия, не позволяя отлаживаемому коду себя обнаружить, какие бы ни были у него привилегии. Правда, если «подопытный» код отлаживает сам себя, задействовав все четыре контрольные точки, отладчик не сможет работать.

Если бит **Т** в **TSS** отлаживаемой задачи установлен, то при каждом переключении на нее будет генерироваться отладочное исключение до выполнения первой команды задачи. Чтобы предотвратить собственное обнаружение, отладчик может отслеживать всякие обращения к **TSS** и возвращать программе подложные данные. Необходимо заметить: Windows NT по соображениям производительности не использует **TSS** (точнее, использует, но всего один) и эта отладочная возможность для нее совершенно бесполезна.

Программная точка останова — единственное, что нельзя замаскировать, не прибегая к написанию полноценного эмулятора процессора. Она представляет собой однобайтовый код **0xCC**, который, если его поместить в начало инструкции, вызывает исключение **INT 0x3** при попытке ее выполнения. Отлаживаемой программе достаточно подсчитать свою контрольную сумму, чтобы выяснить, была ли установлена хоть одна точка останова или нет. Для достижения этой цели она может воспользоваться командами **MOV**, **MOVS**, **LODS**, **POP**, **CMP**, **CMPS** или любыми другими, ни один отладчик не в состоянии их все отследить и эмулировать.

Настоятельно рекомендую использовать программные точки останова в тех и только в тех случаях, когда аппаратных уже не хватает. Как мы увидели выше, в процессорах x86-64 имеется четыре аппаратные точки останова. Много это или мало — зависит от случая и отлаживаемой программы. Однако практически все современные отладчики всегда устанавливают программные, а не аппаратные точки останова. Это обстоятельство может быть с успехом использовано в защитных механизмах, примеры реализаций которых приведены в разделе «Как противостоять трассировке».

### Обработка исключений в реальном и защищенном режимах

Когда возникает отладочное исключение (как, впрочем, и любое другое), процессор заносит в стек регистр флагов, адрес следующей (или текущей — в зависимости от рода исключения) выполняемой инструкции и лишь затем передает управление отладчику.

В реальном режиме флаги с адресом возврата заносятся в стек отлаживаемой программы, поэтому факт отладки обнаружить очень просто — достаточно контролировать целостность содержимого, лежащего выше указателя стека. Или, как вариант, установить указатель на его вершину, тогда добавление новых данных в стек окажется невозможным и отладчик не сможет функционировать.

Иная ситуация складывается при работе в защищенном режиме — обработчик исключения может находиться в своем собственном адресном пространстве и не использовать никаких ресурсов отлаживаемого приложения, в том числе и стека. Грамотно спроектированный отладчик защищенного режима ни обнаружить, ни блокировать принципиально не может даже привилегированный код, исполняющийся в нулевом кольце. Сказанное справедливо для всех версий Windows семейства NT.

### Как хакеры ломают программы?

Вскрыть защитный механизм в общем случае не составляет проблемы. Куда сложнее найти этот самый механизм в коде ломаемого приложения. Сегодня мало кто использует для этой цели автоматическую трассировку — на смену ей пришли аппаратные контрольные точки.

Пусть некая защита запрашивает пароль, затем каким-то образом удостоверяется в его подлинности (например, сравнивает с оригиналом) и в зависимости от результатов проверки передает управление соответствующей ветке программы. Вскрыть такую защиту взломщик может, даже не вникая в алгоритм аутентификации! Он просто введет первый пришедший ему на ум пароль (необязательно совпадающий с правильным), найдет его в памяти, установит контрольную точку на первый символ строки своего пароля, дождется «всплытия» отладчика, отследившего обращение к паролю, выйдет из сравнивающей процедуры и «подправит» условие перехода так, чтобы управление всегда получала нужная ветвь программы.

Время снятия подобных защит измеряется секундами, и обычно такие программы ломаются раньше, чем успевают дойти до легального потребителя. К счастью, этому можно противостоять!

### Как защитить свои программы

Откуда бы ни бралась ключевая информация — из реестра, файла или с клавиатуры, взломщик может практически мгновенно локализовать ее местоположение в памяти и установить контрольную точку. Помешать этому нельзя, но не составит труда подложить хакеру сюрприз — пусть ключевая информация не анализируется сразу же после получения, а передается в качестве аргумента множеству функций, которые что-то с ней делают и затем передают другим функциям, а те, в свою очередь, следующим.

Защитный механизм может быть встроен во что угодно, хоть в процедуру открытия файла или расчета зарплаты. Не стоит делать явных проверок, пусть лучше в случае вызова функции с неверной ключевой информацией она возвратит неправильный результат, но не сигнализирует об ошибке. Взломанная программа на первый взгляд будет исправно работать, и далеко не сразу выяснится, что работает она неправильно (например, выводит на экран одни числа, а на принтер — совсем другие). А чтобы обезопасить легального пользователя от ошибочного ввода пароля, достаточно в одном месте явно проверить его контрольную сумму, которая не дает взломщику никакой информации об истинном значении пароля.

Таким образом, защита как бы «размазывается» по всей программе, буфера с ключевыми данными многократно дублируются, и на отслеживание обращений у взломщика не хватит ни контрольных точек, ни терпения для анализа огромного объема манипулирующего с ними кода. Будет еще лучше после проверки ключевой информации эти же самые буфера использовать для хранения служебных данных, обращаться к которым по возможности максимально часто. Это не позволит взломщику быстро отличить защитный механизм от прочего прикладного кода.

Поскольку большинство взломщиков ставит контрольную точку на начало буфера, имеет смысл поместить в первые четыре байта ключа «заглушку», обращения к которой не происходит вовсе (либо с ней манипулирует имитатор защиты, направляя хакера по ложному пути).

В такой ситуации взломщику ничего не останется, кроме как, затарившись пивом, плотно засесть за кропотливое изучение всего кода программы, прямо или косвенно манипулирующего ключевой информацией (а это многие мегабайты дизассемблерного листинга!). Если критическая часть кода зашифрована, причем в процессе работы программы не расшифровывается полностью (при выходе в каждую функцию она расшифровывается, а при выходе зашифровывается вновь), хакер не сможет получить готовый к дизассемблированию дамп и будет вынужден прибегнуть к трассировке. А вот тут его будет ждать второй сюрприз!

### Как противостоять трассировке

Принципиальная возможность создания подлинно «невидимых» отладчиков остается гипотетической — большинство из них позволяет обнаружить себя даже непривилегированному коду.

Наибольшие нарекания вызывает использование однобайтового кода **0xCC** (инструкция **INT 0x3**) для создания точки останова вместо поручения той же задачи специально для этого предназначенным отладочным регистрам процессора. Так поступают WinDbg, SoftICE, Turbo Debugger, Code Viewer и отладчик, интегрированный в Microsoft Visual Studio. Причем последний неявно использует точки останова при пошаговом прогоне программы, помещая в начало следующей инструкции этот пресловутый байт **0xCC**.

Тривиальная проверка собственной целостности позволяет обнаружить факт установки точек останова, свидетельствующий об отладке. Не стоит использовать конструкции наподобие

```c
if (CalculateMyCRC() != MyValidCRC) {
    printf("Hello, Hacker!\n");
    return;
}
```

Их слишком легко обнаружить и нейтрализовать, подправив условный переход так, чтобы он всегда передавал управление нужной ветке программы. Лучше расшифровывать полученным значением контрольной суммы критические данные или некоторый код.

Простейшая защита, реализованная в демонстрационной программе cipherString, может выглядеть так:

```c
#include <iostream> int main()
{
    // Зашифрованная строка Hello, Free World!
    char s0[] = "\x0C\x21\x28\x28\x2B\x68\x64\x02\x36\x21\x21\x64\x13\x2B\x36\x28\x20\x65\x49\x4E";
    _asm
    {
    BeginCode: // Начало контролируемого кода
        pusha // Сохранение всех регистров общего назначения
        lea ebx, s0 // ebx = &s0[0]
        GetNextChar : // do
        XOR eax, eax // eax = 0;
        LEA esi, BeginCode // esi = &BeginCode
        LEA ecx, EndCode // Вычисление длины...
        SUB ecx, esi // ...контролируемого кода
    HarvestCRC : // do
        LODSB // Загрузка очередного байта в al
        ADD eax, eax // Вычисление контрольной суммы
        LOOP HarvestCRC // until (-cx > 0)
        xor [ebx], ah // Расшифровка очередного символа строки
        s0 inc ebx // Указатель на следующий символ
        cmp[ebx], 0 // until (пока не конец строки)
        jnz GetNextChar // Продолжить расшифровку
        popa // Восстановить все регистры
    EndCode : // Конец контролируемого кода
        NOP // Safe Breakpoint
    }

    std::cout << s0; // Вывод строки на экран return 0;
}
```

Обрати внимание, что данный код не будет компилироваться для платформы x64 компилятором Visual C++, поскольку в нем отсутствует поддержка ключевого слова **_asm**. Тем не менее для x86 он построится без проблем. Это не значит, что для x64 нельзя писать программы на C/C++, содержащие ассемблерные вставки. Однако в таком случае их придется оформлять в виде процедур в отдельных файлах *.asm. В рассматриваемом случае разница между платформами для нас некритична, так как в наши цели не входит дизассемблирование расшифровщика.

При нормальном запуске на экране должна появиться строка «Hello, Free World!».

| ![Расшифрованная строка](./images/26/console_decrypte_str.png)
|:--:|
| *Расшифрованная строка* |

А под отладчиком при наличии хотя бы одной точки останова, установленной в пределах от BeginCode до EndCode, мы увидим следующее.

| ![Установлена точка останова](./images/26/set_breakpoint.png)
|:--:|
| *Установлена точка останова* |

На экране появится бессмысленный мусор.

| ![Вместо стройного приветствия программа вывела бессмысленный набор символов](./images/26/console_no_decrypte_str.png)
|:--:|
| *Вместо стройного приветствия программа вывела бессмысленный набор символов* |

#### Дополнительная защита

Можно усилить защиту, поместив процедуру подсчета контрольной суммы в отдельный поток, занимающийся (для сокрытия своей деятельности) еще чем-нибудь полезным, так, чтобы защитный механизм по возможности не бросался в глаза.

Потоки вообще великая вещь, требующая к себе особого подхода. Человеку очень трудно смириться с тем, что программа может исполняться во множестве мест одновременно. Распространенные отладчики грешат тем, что отлаживают каждый поток по отдельности, но никогда два и более сразу. Приведенный ниже пример cipherString_thread показывает, как это можно использовать для защиты.

```c
#include <stdio.h>
#include <string.h>
#include <process.h>

// Эта функция будет выполняться в отдельном потоке,
// ее назначение незаметно изменять регистр символов в строке,
// содержащей имя пользователя
void My(void\* arg)
{

    int p = 1; // Указатель на шифруемый байт
    // Обрати внимание: шифровка выполняется не с первого байта. Это позволяет обойти контрольную точку, установленную на начало буфера

    // Выполнять до тех пор, пока не встретится перенос строки
    while (((char*)arg)[p] != '\n')
    {
        // Ожидать, пока очередной символ не будет инициализирован
        while (((char*)arg)[p] < 0x20);

        // Инвертировать пятый бит
        // Это приводит к изменению регистра латинских символов на противоположный

        ((char*)arg)[p] ^= 0x20;

        // Указатель на следующий обрабатываемый байт
        p++;
    }
}

int main(int argc, char\* argv[])
{

    char name[100]; // Буфер, содержащий имя пользователя
    buff[100]; // Буфер, содержащий пароль

    // Забивка буфера имени пользователя нулями

    // Некоторые компиляторы это делают за нас, но не все!
    memset(&name[0], 0, 100);

    // Выполнять процедуру My в отдельном потоке

    _beginthread(&My, NULL, (void*)&name[0]);

    // Запрос имени пользователя
    printf("Enter name: ");
    fgets(&name[0], 66, stdin);

    // Запрос пароля
```

Важно отметить: пока пользователь вводит пароль, второй поток получает достаточно квантов времени, чтобы изменить регистр всех символов имени пользователя. Это обстоятельство не так очевидно и не вытекает из беглого анализа программы, особенно при ее исследовании под отладчиком, плохо отслеживающим влияние отдельных компонентов программы друг на друга.

```c
    printf("Enter password: ");
    fgets(&buff[0], 66, stdin);

    // Сравнение имени и пароля c эталонными значениями
    if (!(strcmp(&buff[0], "password\n")
```

Важно отметить: поскольку введенное пользователем имя было преобразовано, фактически происходит сравнение не `strcmp(&name[0], "KPNC\n"), а strcmp(&name[0], "Kpnc\n")`, что далеко не очевидно на первый взгляд.

```c
       || strcmp(&name[0], "KPNC\n")))


        // Правильные имя и пароль
        printf("USER OK\n");
    else
        // Ошибка при вводе имени или пароля
        printf("Wrong user or password!\n");

    return 0;
}
```

На первый взгляд, программа ожидает «услышать» `KPNC:password`. Но так ли это на самом деле?

| ![Неверная пара имени пользователя и пароля](./images/26/console_name_psw_wrong.png)
|:--:|
| *Неверная пара имени пользователя и пароля* |

А вот и нет! Верный ответ — `Kpnc:password`.

| ![Правильная пара имени пользователя и пароля](./images/26/console_name_psw_right.png)
|:--:|
| *Правильная пара имени пользователя и пароля* |

Пока пользователь вводит свой пароль, второй поток обрабатывает буфер, содержащий его имя, и меняет регистр всех символов, кроме первого, на противоположный. Фокус состоит в том, что при пошаговой трассировке одного потока все остальные потоки выполняются независимо от него и могут произвольным образом вклиниваться в работу отлаживаемого потока, например модифицировать его код.

Взять потоки под контроль можно, введя в каждый из них точку останова, но, если потоков окажется больше четырех (а что мешает разработчику защиты их создать?), отладочных регистров на всех не хватит и придется прибегать к использованию опкода 0xCC, который защитному механизму ничего не стоит обнаружить!

#### SEH (структурная обработка исключений)

Дело осложняется тем, что большинство отладчиков очень плохо переносят программы со [структурной обработкой исключений (SEH)](https://xakep.ru/2015/04/17/195-exceptions/)[.](https://xakep.ru/2015/04/17/195-exceptions/) Инструкция, вызывающая обрабатываемое исключение, либо «срывает» отладчик, выходя из-под его контроля, либо передает управление на библиотечный фильтр исключений. А он, прежде чем передать управление прикладному обработчику, вызывает множество своих служебных функций, в которых взломщику немудрено и «утонуть».

Впрочем, по сравнению с ранними отладчиками даже это большой прогресс, так как раньше они жестко держали некоторые прерывания, не позволяя программе самостоятельно обрабатывать, скажем, деление на ноль.

Вот пример защиты TryExcept, построенный на обработке структурных исключений:

```c
#include <iostream>
#include <windows.h>

int main()
{
    // Защищенный блок
    __try
    {
        // Попытка деления на ноль
        int a = 1;
        int b = 1;
        int c = 2 / (a - b);
```

Такая многословность объясняется тем, что большинство компиляторов выдают ошибку, встретив конструкцию наподобие `int a=a/0`;. При выполнении деления отладчики, в том числе WinDbg, теряют контроль над отлаживаемой программой и «слетают». Поэтому сюда может быть вставлен для отвода глаз некий код, который никогда не получит управления. Если значение переменным a и b присваивается не непосредственно, а из результата, возвращенного некими функциями, то при дизассемблировании программы их равенство будет не так очевидно. В результате взломщик может потратить много времени на анализ совершенно бесполезного кода.

```c
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        std::cout << "Error";
```

Этот код получит управление, если возникнет исключение «деление на ноль», но отладчики не распознают такой ситуации. Требуется ручная установка точки останова на первую инструкцию блока `__except`. А чтобы определить адрес блока `__except`, требуется разобраться, каким именно образом реализована поддержка SEH в конкретном компиляторе. Что опять же потребует от хакера неприлично много времени.

```c
    }
}
```

Если попытаться прогнать приведенный пример под WinDbg, достигнув строки `int c=c/(a - b)`, отладчик внезапно «слетит», теряя контроль над отлаживаемым приложением. Хотя оператор `cout` сработает и строка будет выведена, отладчик, потеряв ориентиры, не сделает остановку на этой строке. Теоретически исправить ситуацию можно заблаговременной установкой точки останова на первую команду блока `except`, но попробуй-ка вычислить, где расположен этот блок, не заглядывая в исходный текст, которого у хакера заведомо нет!

| ![WinDbg перед возникновением исключения](./images/26/windbg_before_ex.png)
|:--:|
| *WinDbg перед возникновением исключения* |

| ![WinDbg после возникновения исключения теряет контакт с отлаживаемой программой](./images/26/windbg_after_ex.png)
|:--:|
| *WinDbg после возникновения исключения теряет контакт с отлаживаемой программой* |

В идеальном мире после срабатывания исключения WinDbg должен переходить в область кода `__except`. Посмотрим, как это должно быть. Установим точку останова на адрес начала оператора вывода строки Error — `std::cout << "Error";`. При этом вообразим, что у нас нет исходного кода отлаживаемого приложения:

```console
bp 00007ff7`64ff230d
```

И если после остановки на инструкции `idiv` по F5 продолжить выполнять приложение, отладчик нырнет в блок кода `__except` и остановится на инструкции загрузки параметра для вывода строки оператором `cout`.

| ![WinDbg остановился на поставленной вручную точке останова в секции __except](./images/26/windbg_breakpoint.png)
|:--:|
| *WinDbg остановился на поставленной вручную точке останова в секции __except* |

Прежде чем справиться с такой защитой, взломщику придется основательно изучить реализацию механизма обработки структурных исключений, как на уровне операционной системы, так и на уровне конкретного компилятора. В подавляющем большинстве литературы этот вопрос обходится стороной. И неспроста — реализация SEH действительно очень сложна, громоздка, многословна. Все это приводит к тому, что большинство программистов и технических писателей совершенно не представляют, что находится у нее «под капотом».

Поскольку SEH по-разному реализована в каждом компиляторе, неудивительно, что отладчики отказываются ее поддерживать. Поэтому предложенный вариант защиты очень стоек к взлому и в то же время крайне прост в реализации. А самое важное — он одинаково хорошо работает во всех операционных системах семейства Windows.

### Как противостоять контрольным точкам останова

Контрольные точки, установленные на важнейшие системные функции, — мощное оружие в руках взломщика. Пусть, к примеру, защита пытается открыть ключевой файл. Под Windows существует только один документированный способ это сделать — вызвать функцию **CreateFile** (точнее, **CreateFileA** или **CreateFileW** для ASCII- и UNICODE-имени файла соответственно). Все остальные функции, наподобие **OpenFile**, доставшиеся в наследство от ранних версий Windows, на самом деле представляют собой переходники к **CreateFile**.

Зная об этом, взломщик может заблаговременно установить точку останова на адрес начала этой функции (благо он ему известен) и мгновенно локализовать вызывающий эту функцию защитный код, ну а остальное, как говорится, дело техники.

Но не всякий взломщик знает, что открыть файл можно и другим путем — вызвать функцию **ZwCreateFile** (равно как и **NtCreateFile**), экспортируемую **NTDLL.DLL**, или обратиться напрямую к ядру вызовом прерывания **INT 0x2Eh**. Сказанное справедливо не только для **CreateFile**, но и для всех остальных функций ядра. Причем для этого не нужны никакие привилегии и такой вызов можно осуществить даже из прикладного кода!

Опытного взломщика такой трюк надолго не остановит, но почему бы ему не приготовить один маленький сюрприз, поместив вызов **INT 0x2E** в блок **__try**. Это приведет к тому, что управление получит не ядро системы, а обработчик данного исключения, находящийся за блоком **__try**. Взломщик, не имеющий исходных текстов, не сможет быстро определить, относится ли данный вызов к блоку **__try** или нет. Отсюда он может быть легко введен в заблуждение — достаточно имитировать открытие файла, не выполняя его на самом деле! Кроме того, ничто не мешает использовать прерывание **INT 0x2E** для взаимодействия компонентов своей программы — взломщику будет очень непросто отличить, какой вызов пользовательский, а какой системный.

Хорошо, с ядром все понятно, а как же быть с функциями модулей **USER** и **GDI**? Например, **GetWindowsText**, использующейся для считывания введенной пользователем ключевой информации (как правило, серийного номера или пароля)?

На помощь приходит то обстоятельство, что практически все эти функции начинаются с инструкций

```asm
PUSH EBP
MOV EBP, ESP
```

Эти инструкции могут выполнить прикладной код и самостоятельно, передав управление не на начало функции, а на три байта ниже. Поскольку `PUSH EBP` изменяет стек, приходится прибегать к передаче управления посредством `JMP` вместо `CALL`. Контрольная точка, установленная взломщиком на начало функции, не возымеет никакого действия! Такой трюк может сбить с толку даже опытного хакера, хотя рано или поздно он все равно раскусит обман.

Если есть желание окончательно отравить взломщику жизнь, следует скопировать системную функцию в свой собственный стек и передать на него управление — контрольные точки взломщика «отдыхают»! Между тем надо отметить, что современные версии операционных систем могут запрещать выполнение кода на определенных страницах памяти, например в стеке. Эта функция может быть отключена, но стоит иметь ее в виду. Плюс основная сложность заключается в необходимости распознавать все инструкции с относительными адресными аргументами и их соответствующе корректировать.

Например, двойное слово, стоящее после инструкции `CALL`, представляет собой не адрес перехода, а разность целевого адреса и адреса следующей за `CALL` инструкции. Перенос инструкции `CALL` на новое место потребует коррекции ее аргумента. Впрочем, эта задача не так сложна, как может показаться на первый взгляд (глаза страшатся, а руки делают), и результат оправдывает средства. Во-первых, при каждом запуске функции можно произвольным образом менять ее адрес, во-вторых, проверкой целостности кода легко обнаружить программные точки останова — а аппаратных точек на все вызовы просто не хватит!

Разве не заслуживают награды за свою целеустремленность те индивидуумы, которые такую защиту взломают? Под наградой здесь подразумевается отнюдь не сама взломанная программа, а глубокое чувство удовлетворения от того, что «я это сделал!».

Еще легче противостоять аппаратным точкам останова на память, поскольку их всего четыре и каждая может контролировать не более двойного слова. Соответственно, взломщик может одновременно контролировать не более 16 байт памяти. Если же обращения к содержащим ключевую информацию буферам будут происходить не последовательно байт за байтом от начала до конца, а произвольно и количество самих буферов окажется больше четырех, отследить все операции чтения-записи в них станет невозможно.

Некоторые отладчики поддерживают возможность установки точки останова на диапазон памяти, но ее функциональность вызывает большие сомнения. Единственный способ контролировать целый регион — трассировать исследуемую программу, проверяя, не обращается ли очередная команда к отслеживаемому диапазону, и если да — генерировать исключение.

Во-первых, манипулирующих с памятью команд очень много, и можно придумать самые неожиданные комбинации. Например, установить указатель стека на требуемую ячейку памяти и вызвать **RET** для чтения содержащегося в ней значения. Во-вторых, возникшее при этом исключение может служить хорошим средством избавления от трассировщика (см. раздел «Как противостоять трассировке»).

Таким образом справиться с контрольными точками защитному механизму совсем нетрудно!

#### Точка остановка

Контрольная точка (или «бряк» на хакерском жаргоне) представляет собой однобайтовую команду **0xCC**, генерирующую исключение **0x3** при попытке ее выполнения (в просторечии «дергающее отладочным прерыванием»). Обработчик **INT 0x3** получает управление и может делать с программой абсолютно все, что ему заблагорассудится, но прежде — до вызова прерывания — в стек заносятся текущие регистр флагов, указатель кодового сегмента (регистр **CS**), указатель команд (регистр **IP**), запрещаются прерывания (очищается флаг **IF**) и сбрасывается флаг трассировки — словом, вызов отладочного прерывания не отличается от вызова любого прерывания.

<!-- | ![Состояние стека на момент входа в обработчик прерывания](./images/26/stack_breakpoint.png)
|:--:|
| *Состояние стека на момент входа в обработчик прерывания* | -->

Чтобы узнать, в какой точке программы произошел останов, отладчик извлекает из стека сохраненное значение регистров, не забывая о том, что CS:IP указывают на следующую выполняемую команду.

> **Указатель на команду**
>
> Стоит отметить, что описанный выше вариант адресации команд в процессорах на архитектуре x86_64 работает в режиме совместимости. То есть IP — 16-разрядный регистр, имевший место в самых ранних микропроцессорах линейки 8086. В процессорах марки 80386 регистр был расширен до 32 бит и получил название EIP, при этом доступ к младшей 16-разрядной части остался. В остальном же указатель команд работал по прежнему принципу.
>
> Вместе с выпуском процессоров на расширенной архитектуре AMD64 регистр указателя команд расширился до 64 бит и стал называться RIP. Доступ к обеим младшим частям был оставлен для обратной совместимости. Это стало называться режимом совместимости. Между тем одной из целей, преследуемых при проектировании новой архитектуры, было избавление от сегментов кода, данных (CS, DS) и прочих. Поэтому в 64-разрядные процессоры был введен новый метод адресации RIP-relative (RIP-относительная адресация). В этом режиме сегмент кода не используется и вся арифметика с указателем выполняется относительно значения регистра RIP следующей команды.

Условно точки останова (называемые также контрольными точками) можно разделить на две категории: точки останова, жестко прописанные в программе самим разработчиком, и точки динамические, устанавливаемые отладчиком. Ну, с первыми все ясно: хочешь остановить программу и передать управление отладчику в таком-то месте — пишешь `__asm {int 0x3}`.

Несколько сложнее установить точку в произвольное место программы: сначала отладчик должен сохранить текущее значение ячейки памяти по указанному адресу, затем записать сюда код **0xCC**, а перед выходом из отладочного прерывания вернуть все на место и модифицировать сохраненный в стеке **IP** для перемещения его на начало восстановленной команды (иначе он будет указывать на ее середину).

Какими недостатками обладает механизм точек останова процессора 8086? Первое и самое неприятное состоит в том, что отладчик, устанавливая точку останова, вынужден непосредственно модифицировать код.

Причем WinDbg неявно помещает точку останова в начало каждой следующей команды при трассировке программы по Step Over (клавиша F10)! Разумеется, это искажает контрольную сумму, чем и пользуется защита. Этот эффект мы могли наблюдать, разбирая пример cipherString.

Самое простое решение проблемы — положить кирпич на клавишу F8 (покомандная трассировка — Step Into) и идти пить чай, пока программа будет расшифровываться. Шутка, конечно. А если говорить серьезно, то необходимо вспомнить, в каком веке мы живем, и, отбросив каменные топоры, установить аппаратную точку останова!

## ИТОГИ

Мы изучили варианты «побега из-под отладчика», другими словами — рассмотрели методы противостояния отладочным механизмам, а также узнали, каким образом можно защитить свою программу от постороннего исследования и вмешательства в ее работу извне. Получилось, что на каждый отладочный механизм имеется по меньшей мере одно средство противодействия.

Обладая современными средствами, любую защиту можно взломать, поэтому главная задача разработчика защиты — запутать хакера, ввести его в заблуждение, чтобы жизнь медом не казалась и ценность временного промежутка, необходимого для взлома защиты, стала бы выше ценности результата взлома.
